#include <stdio.h>
#include <string.h>

#include <dsdebug.h>
#include <dsassert.h>
#include <dserror.h>

#include "error.h"
#include "defaults.h"
#include "lowlevel.h"
#include "final.h"

#include "ascii.h"
#include "asciibuffer.h"



static int asciiReader(ASCIIDevice dev);
static int asciiProcessor(ASCIIDevice dev, ASCIIReaderBuf buf);
static void asciiDeviceDestroy(ASCIIDevice dev);
static void asciiReset(ASCIIDevice dev);
static int asciiDebug(ASCIIDevice dev);

void asciiDeviceSetDefaults(ASCIIDevice dev) {
    DeviceSetDefaults((Device)dev);

    dev->dw = NULL;
    dev->channel = NULL;
    dev->channels = 0;
}

Device asciiDeviceCreate(DSConfigContext config, DSConfigReference ref) {
    int err = 0;
    ASCIIDevice dev;
    
    dev = (ASCIIDevice)malloc(sizeof(ASCIIDeviceS));
    if (!dev) return (Device)dev;

    asciiDeviceSetDefaults(dev);

    dev->lasttime = dsTimeCreate(NULL);
    if (!dev->lasttime) err = DS_ERR_MEM;
    
    dev->force_duration = dsDurationCreate(NULL);
    if (!dev->force_duration) err = DS_ERR_MEM;

    dev->lock = dsLockCreateContext(NULL);
    if (dev->lock) err = dsLockInitContext(dev->lock);
    else err = DS_ERR_MEM;
    
    if (err) {
	if (dev->force_duration) dsDurationDestroy(dev->force_duration);
	if (dev->lasttime) dsTimeDestroy(dev->lasttime);
	if (dev->lock) dsLockDestroyContext(dev->lock);
	free(dev);
	return NULL;
    }

    dev->device.version = ASCII_VERSION;
    
//    dev->device.configure = asciiConfigure;
    dev->device.destroy = (DeviceDestroyFunction)asciiDeviceDestroy;
    dev->device.reader = (DeviceReaderFunction)asciiReader;
    dev->device.processor = (DeviceProcessorFunction)asciiProcessor;
    dev->device.reset = (DeviceResetFunction)asciiReset;
    dev->device.debug = (DeviceDebugFunction)asciiDebug;
    
    return (Device)dev;
}

static void asciiDeviceDestroy(ASCIIDevice dev) {
    ds_assert_return(dev);
    
    if (dev->dw) DeviceFreeWriters((Device)dev, dev->dw);
    if (dev->channel) free(dev->channel);
    dsTimeDestroy(dev->lasttime);
    dsDurationDestroy(dev->force_duration);
    dsLockDestroyContext(dev->lock);
    free(dev);
}

static int asciiIsSpace(ASCIIDevice dev, char ch) {
    if (((ch>='0')&&(ch<='9'))||(ch=='e')||(ch=='E')||(ch=='.')||(ch=='-')||(ch=='+')) return 0;
    return 1;
}


static int asciiReader(ASCIIDevice dev) {
	int err = 0;
	int i, start;
	unsigned char doread = 0;
	unsigned char informated = 0;
	uint16_t value;

	char *buf = NULL;
	unsigned int channels;
	double *channel = NULL;
	
    
	Connection con;

	DSDuration slp;
	DSDuration duration;
	DSDuration timeout;
	
	DSTime datatime;
	
	xmlNodePtr cfgnode, chipcfg;

	slp = dsDurationCreate(NULL);
	datatime = dsTimeCreate(NULL);
	duration = dsDurationCreate(NULL);
	timeout = dsDurationCreate(NULL);
	
	if ((!slp)||(!datatime)||(!duration)||(!timeout)) {
	    if (slp) dsDurationDestroy(slp);
	    if (datatime) dsTimeDestroy(datatime);
	    if (duration) dsDurationDestroy(duration);
	    if (timeout) dsDurationDestroy(timeout);
	    
	    Abort();
	    return DS_ERR_MEM;
	}    

	dsDurationImportMicroseconds(slp, ASCII_TIMEOUT_REINIT);
	dsDurationZeroDuration(timeout);
	dsDurationZeroDuration(dev->force_duration);
	

	while (DeviceContinueLoop((Device)dev)) {
	    err = 0;
	    dev->device.status = DEVICE_STATUS_NOT_CONNECTED;
	    dev->reset = 0;

	    dsDebugServer(DS_DEBUG_INFO, "ASCII(%s). Reininitializing!", dev->device.id);
	    
	    con = DeviceInitializeConnection((Device)dev);
	    if (!con) {
		if (!informated) {
		    dsDebugServer(DS_DEBUG_CRITICAL, "ASCII(%s). Error opening communication device!", dev->device.id);
		    informated = 1;
		}
		DeviceCloseConnection((Device)dev);
		dsWait(slp);
		continue;
	    } else informated = 0;

	    cfgnode = dsConfigGetConfigByReference(dev->device.cfg, dev->device.cfgref);
	    if (cfgnode) {
		chipcfg = dsXmlFindSubNode(cfgnode, "Configuration");
		if (chipcfg) {
		    dsXmlGetNodeDuration(chipcfg, "Timeout", timeout);
		    dsXmlGetNodeDuration(chipcfg, "ForceDuration", dev->force_duration);
		}
		dsConfigReturnConfig(dev->device.cfg, cfgnode);
	    }
	    
	    ConnectionReadAll(con);


	    dev->device.sf = 1;
	    dsTimeCurrentTime(datatime);
	    while ((DeviceContinueLoop((Device)dev))&&(!err)&&(!dev->reset)) {
		buf = DeviceReadText((Device)dev, con, NULL);
		
		if (buf) {
		    dsTimeCurrentTime(datatime);
		    
		    dsDebugServer(DS_DEBUG_DEBUG, "ASCII(%s). New iteration", dev->device.id);
		    
		    for (i=0, start=-1, channels=0;buf[i];i++) {
			if (asciiIsSpace(dev, buf[i])) {
			    if (i>(start+1)) channels++;
			    start = i;
			}
		    }
		    if (i>(start+1)) channels++;
		} else channels = 0;
		
		if (channels>0) {
		    channel = (double*)malloc(channels*sizeof(double));
		    if (!channel) {
			err = DS_ERR_MEM;
			Abort();
			break;
		    }

		    for (i=0, start = -1, channels = 0;buf[i];i++) {
			if (asciiIsSpace(dev, buf[i])) {
			    if (i>(start+1)) {
				buf[i] = 0;
				channel[channels++] = atof(buf + start + 1);
			    } 
			    start = i;
			}
		    }
		    if (i>(start+1)) channel[channels++] = atof(buf + start + 1);
		    free(buf); buf = NULL;
		    
		    dsLockWrite(dev->lock);
		    if (dev->channel) free(dev->channel);
		    dev->channels = channels;
		    dev->channel = channel;
		    dsLockReturn(dev->lock);

		    dev->device.status = DEVICE_STATUS_OK;
		    
			// submit
		    asciiBufferAdd(dev, datatime, channels, channel);
		} else {
		    if (buf) {
			free(buf);
			buf = NULL;
		    }
			// checking for timeout
		    dsTimeCalcPassedDuration(duration, datatime);
		    if ((!dsDurationIsZero(timeout))&&(dsDurationCmp(duration, timeout)>0)) err = DS_ERR_TIMEOUT;
		}
		    

		if (err) {
		    if (err == DS_ERR_MEM) {
			dsDebugServer(DS_DEBUG_FATAL, "ASCII(%s). Fatal error: %X(%s)", dev->device.id, err, dsError(err, NULL));
			Abort();
		    } else {
			dsDebugServer(DS_DEBUG_ERROR, "ASCII(%s). Error reading data: %X(%s)", dev->device.id, err, dsError(err, NULL));
		    }
		}
	    }
	    
	    dev->device.status = DEVICE_STATUS_NOT_CONNECTED;
	    DeviceCloseConnection((Device)dev);
	    dsWait(slp);
	}
	
	if (buf) free(buf);
	dsDurationDestroy(slp);
	dsDurationDestroy(duration);
	dsDurationDestroy(timeout);
	dsTimeDestroy(datatime);
	
	return 0;
}

static int asciiProcessor(ASCIIDevice dev, ASCIIReaderBuf buf) {
    int err;
    long duration;
    
    size_t pos;
    size_t idlen;
    char *stcBuffer;
    
    ds_assert_param(dev);
    ds_assert_param(buf);

    if (buf->buf.sf) {
	dsTimeCopy(dev->lasttime, buf->buf.tm);

	if (dev->dw) DeviceFreeWriters((Device)dev, dev->dw);
	dev->dw = DeviceInitWriters((Device)dev, NULL);

	return 0;
    } else {
	if (dsDurationIsZero(dev->force_duration))
	    duration = dsTimeDiffMilliseconds(dev->lasttime, buf->buf.tm);
	else
	    duration = dsDurationGetMilliseconds(dev->force_duration);
	    
	dsTimeCopy(dev->lasttime, buf->buf.tm);
    }

    if (dev->dw) err = DeviceOpenWriters((Device)dev, dev->dw, NULL, buf->buf.tm);
    else err = DS_ERR_MEM;

    if (!err) {
	DeviceWritersSetDurationMilliseconds((Device)dev, dev->dw, duration);
	err = DeviceWrite((Device)dev, dev->dw, DATA_CHANNEL, NULL, 0, buf->channels, NULL, buf->channel);
	DeviceCloseWriters((Device)dev, dev->dw);
    }

    if (err) {
	dsDebugServer(DS_DEBUG_ERROR, "ASCII(%s). Data writeout failed: %lX(%s).", dev->device.id, err, dsError(err, NULL));
    }

    idlen = strlen(dev->device.id) + 1;
    stcBuffer = dsMalloca((2+idlen)*sizeof(uint8_t) + 3*sizeof(uint16_t) + sizeof(time_t) + (2)*sizeof(uint32_t) + (1+buf->channels)*sizeof(double));
    if (stcBuffer) {
        *(char*)stcBuffer = 0xE1;
	pos = 1;			*(uint16_t*)(stcBuffer+pos) = dev->device.type;
	pos += sizeof(uint16_t);	*(uint16_t*)(stcBuffer+pos) = dev->device.version;
	pos += sizeof(uint16_t);	memcpy(stcBuffer+pos, dev->device.id, idlen);
	pos += idlen*sizeof(uint8_t); 	*(time_t*)(stcBuffer+1) = dsTimeGetUnixTime(buf->buf.tm);
	pos+=sizeof(time_t);		*(uint32_t*)(stcBuffer+pos)=duration;
	pos+=sizeof(uint32_t);		*(uint32_t*)(stcBuffer+pos)=buf->channels;
	pos+=sizeof(uint32_t);  	*(uint8_t*)(stcBuffer+pos) = DATA_TYPE_MIXED;
	pos+=sizeof(uint8_t);		*(double*)(stcBuffer+pos) = 1;
	pos+=sizeof(double);		*(uint16_t*)(stcBuffer+pos) = 0;
	pos+=sizeof(uint16_t);
	
	memcpy(stcBuffer+pos, buf->channel, buf->channels*sizeof(double));
	pos+=buf->channels*sizeof(double);

	dsDatagramServerBroadcast(dev->device.dsc, NULL, dev->device.id, stcBuffer, pos, 0, NULL);
	dsFreea(stcBuffer);
    }

    return err;
}

static void asciiReset(ASCIIDevice dev) {
    ds_assert_return(dev);
    dev->reset = 1;
}

static int asciiDebug(ASCIIDevice dev) {
    size_t pos, idlen;
    char *stcBuffer;

    ds_assert_param(dev);

    if (dev->device.status != DEVICE_STATUS_OK) return 0;
    if ((!dev->device.debug_dsc)||(dsDatagramServerIsEmpty(dev->device.debug_dsc))) return 0;

    dsLockWrite(dev->lock);
    
    idlen = strlen(dev->device.id) + 1;
    stcBuffer = dsMalloca((2+idlen)*sizeof(uint8_t) + sizeof(time_t) + 3*sizeof(uint16_t) + sizeof(time_t) + (2)*sizeof(uint32_t) + (1+dev->channels)*sizeof(double));
    if (stcBuffer) {

	*(char*)stcBuffer = 0xC1;
	pos = 1; 			*(uint16_t*)(stcBuffer+pos) = dev->device.type;
	pos += sizeof(uint16_t);	*(uint16_t*)(stcBuffer+pos) = dev->device.version;
	pos += sizeof(uint16_t);	memcpy(stcBuffer+pos, dev->device.id, idlen);
	pos += idlen * sizeof(uint8_t);	*(time_t*)(stcBuffer+pos)=0;
	pos += sizeof(time_t);		*(uint32_t*)(stcBuffer+pos)=0;
	pos+=sizeof(uint32_t);		*(uint32_t*)(stcBuffer+pos)=dev->channels;
	pos+=sizeof(uint32_t);  	*(uint8_t*)(stcBuffer+pos) = DATA_TYPE_MIXED;
	pos+=sizeof(uint8_t);		*(double*)(stcBuffer+pos) = 1;
	pos+=sizeof(double);		*(uint16_t*)(stcBuffer+pos) = 0;
	pos+=sizeof(uint16_t);
	
	memcpy(stcBuffer+pos, dev->channel, dev->channels*sizeof(double));
	pos+=dev->channels*sizeof(double);
	
	dsLockReturn(dev->lock);
	
	dsDatagramServerBroadcast(dev->device.debug_dsc, NULL, dev->device.id, stcBuffer, pos, 0, NULL);
	dsFreea(stcBuffer);
	
    } else dsLockReturn(dev->lock);
    
    return 0;
}
