#include <stdio.h>
#include <string.h>

#include <dsassert.h>
#include <dsdebug.h>
#include <dserror.h>

#include "buffer.h"

Buffer BufferCreate(DSConfigContext config, DSConfigReference ref) {
    int err;
    Buffer buf;
    DSLockContext lock;
    DSConditionContext cond, pcond;
    DSConfigReference cfgref;
    
    ds_assert_null(config);
    ds_assert_null(ref);
    
    cfgref = dsConfigCopyReference(config, ref);
    buf = (Buffer)malloc(sizeof(BufferS));
    lock = dsLockCreateContext(NULL);
    if (lock) {
	err = dsLockInitContext(lock);
	if (err) {
	    dsLockDestroyContext(lock);
	    lock = NULL;
	}
    }
    
    cond = dsConditionCreateContext(NULL);
    if (cond) {
	err = dsConditionInitContext(cond);
	if (err) {
	    dsConditionDestroyContext(cond);
	    cond = NULL;
	}
    }

    pcond = dsConditionCreateContext(NULL);
    if (pcond) {
	err = dsConditionInitContext(pcond);
	if (err) {
	    dsConditionDestroyContext(pcond);
	    pcond = NULL;
	}
    }
    
    if ((!buf)||(!lock)||(!cond)||(!pcond)||(!cfgref)) {
	if (cfgref) dsConfigDestroyReference(config, cfgref);
	if (lock) dsLockDestroyContext(lock);
	if (cond) dsConditionDestroyContext(cond);
	if (pcond) dsConditionDestroyContext(pcond);
	if (buf) free(buf);
	return NULL;
    }
    
    buf->cfg = config;
    buf->cfgref = cfgref;
    
    buf->lock = lock;
    buf->cond = cond;
    buf->pcond = pcond;
    buf->rb = NULL;
    buf->rbr = NULL;
    
    buf->thread = NULL;

    return buf;
}

void BufferDestroy(Buffer buf) {
    ReaderBuf rb;
    
    ds_assert_return(buf);

    while (buf->rbr) {
	rb = buf->rbr;
	buf->rbr = rb->next;
        if (rb->destroy) rb->destroy(rb);
	else free(rb);
    }
    
    if (buf->thread) dsThreadDestroyContext(buf->thread);
    
    dsConfigDestroyReference(buf->cfg, buf->cfgref);
    dsLockDestroyContext(buf->lock);
    dsConditionDestroyContext(buf->cond);
    free(buf);
}

int ReaderBufInit(ReaderBuf buf, Device dev, DSTime tm) {
    ds_assert_param(buf);
    ds_assert_param(dev);
    ds_assert_param(tm);
    
    memset(buf, 0, sizeof(ReaderBufS));
    
    buf->tm = dsTimeCreateCopy(NULL, tm);
    if (!buf->tm) return DS_ERR_MEM;
    
    buf->dev = dev;
    
    return 0;
}

int ReaderBufFree(ReaderBuf buf) {
    ds_assert_param(buf);
    
    if (buf->tm) {
	dsTimeDestroy(buf->tm);
	buf->tm = NULL;
    }
    return 0;
}

int BufferAdd(Buffer buf, ReaderBuf rb) {
    ds_assert_param(buf);
    ds_assert_param(rb);
    ds_assert_param(rb->dev);

    rb->sf = rb->dev->sf;
    rb->dev->sf = 0;

    BufferLock(buf);

    if (buf->rbr) {
	buf->rb->next = (ReaderBuf)rb;
	buf->rb = (ReaderBuf)rb;
    } else {
	buf->rb = (ReaderBuf)rb;
	buf->rbr = (ReaderBuf)rb;
    }

    dsDebugServer(DS_DEBUG_INFO, "Buffer. Device(%s) event queued", rb->dev->id);

    BufferUnlock(buf);
    
    dsConditionBroadcast(buf->cond);

    return 0;
}

int BufferProcess(Buffer buf) {
    ReaderBuf rb;
    int err = 0;
    
    Device dev;
    
    ds_assert_param(buf);

    while (buf->rbr) {
	dsLockWrite(buf->lock);
	rb = buf->rbr;
	buf->rbr = rb->next;
	dsLockReturn(buf->lock);

	dsDebugServer(DS_DEBUG_INFO, "Buffer. Processing %s event", rb->dev->id);
	
	dev = rb->dev;
	if ((dev)&&(dev->processor)) err = dev->processor(dev, rb);
	if (err) dsDebugServer(DS_DEBUG_CRITICAL, "Buffer. Device(%s) event proccesing failed: %lx(%s)", rb->dev->id, err, dsError(err, NULL));
	else dsDebugServer(DS_DEBUG_INFO, "Buffer. Device(%s) event processed", rb->dev->id);
	
	if (rb->destroy) rb->destroy(rb);
	else free(rb);
	
	dsConditionBroadcast(buf->pcond);
    }	
     
    return 0;
}


int BufferWait(Buffer buf, DSDuration timeout) {
    ds_assert_param(buf);
    
    dsConditionPreLock(buf->cond);
    if (buf->rbr) {
	dsConditionUnLock(buf->cond);
	return 0;
    }
    
    if (timeout) return dsConditionTimedWait(buf->cond, timeout);
    else {
	dsConditionWait(buf->cond);
	return 0;
    }
}


static int BufferProcessLoop(Buffer buf) {
    int err;

    xmlKeepBlanksDefault(0); /* DS Fix. Blanks default value is separate for all working threads */
    
    while (dsThreadCheckStatusBitsOff((DSBaseThreadContext)buf->thread,DS_THREAD_TERMINATE_FLAG)) {
	BufferWait(buf, NULL);
	err = BufferProcess(buf);
	if (err) dsDebugServer(DS_DEBUG_ERROR, "Buffer. Proccessing failed: %lx(%s)", err, dsError(err, NULL));
    }

    // if (err) Abort();
        
    return 0;
}

int BufferStartLoop(Buffer buf) {
    int err;
    
    ds_assert_param(buf);
    ds_assert_param(!buf->thread);
    
    buf->thread = dsThreadCreateContext(NULL);
    if (buf->thread) {
	err = dsThreadInitContext(buf->thread);
	if (err) {
	    dsThreadDestroyContext(buf->thread);
	    buf->thread = NULL;
	}
    } else err = DS_ERR_MEM;
    
    if (err) return err;

    err = dsThreadRun(buf->thread, (DSThreadFunction)BufferProcessLoop, (void*)buf);
    if (err) {
	dsThreadDestroyContext(buf->thread);
	buf->thread = NULL;
    }
    
    return err;
}

int BufferStopLoop(Buffer buf) {
    ds_assert_param(buf);
    ds_assert_param(buf->thread);

    dsThreadTerminate((DSBaseThreadContext)buf->thread);
    dsConditionBroadcast(buf->cond);
    dsThreadWaitTerminate(buf->thread);
    
    dsThreadDestroyContext(buf->thread);
    buf->thread = NULL;
    
    BufferProcess(buf);
    dsConditionBroadcast(buf->pcond);
    
    return 0;
}

int BufferWaitDeviceEventsProcessed(Buffer buf, Device dev) {
    ReaderBuf rb;
    
    ds_assert_param(buf);
    
    do {
	dsConditionPreLock(buf->pcond);

	dsLockWrite(buf->lock);
	for (rb = buf->rbr;rb;rb=rb->next)
	    if (rb->dev == dev) break;
	dsLockReturn(buf->lock);
	
	if ((rb)&&(dsThreadCheckStatusBitsOff((DSBaseThreadContext)buf->thread,DS_THREAD_TERMINATE_FLAG))) dsConditionWait(buf->pcond);
	else dsConditionUnLock(buf->pcond);
    } while ((rb)&&(dsThreadCheckStatusBitsOff((DSBaseThreadContext)buf->thread,DS_THREAD_TERMINATE_FLAG)));

    return 0;
}

