#define _CONNECTION_C
#define _DS_INCLUDE_XML_HEADERS

#include <stdio.h>

#include <dsassert.h>
#include <dsdebug.h>

#include "defaults.h"
#include "connection.h"

#include <comconnection.h>


Connection ConnectionCreate(DSConfigContext config, DSConfigReference ref) {
    int err;
    int i;
    const char *tmp;
    char *id = NULL;
    Connection ctx;
    DSConnectionContext con;
    DSLockContext lock;
    DSConfigReference cfgref;
    DSDuration timeout, retry_timeout, slow_write_delay;

    ConnectionOpenFunction open_function;
    
    xmlNodePtr cfg;
    
    cfg = dsConfigGetConfigByReference(config, ref);
    if (!cfg) return NULL;
    
    tmp = dsXmlGetPropString(cfg, "id", NULL);
    if (tmp) err = dsExchangeString(NULL, &id, tmp);
    else err = DS_ERR_CONFIG;

    for (i=0;connection_configuration[i].id;i++) {
	if (!xmlStrcmp(cfg->name, connection_configuration[i].id)) {
    	    open_function = connection_configuration[i].open;
	    break;
	}
    }
    if (!connection_configuration[i].id) err = DS_ERR_CONFIG;

    dsConfigReturnConfig(config, cfg);
    
    if (err) {
	if (id) dsExchangeString(NULL, &id, NULL);
	return NULL;
    }
    
    ctx = (Connection)malloc(sizeof(ConnectionS));
    con = dsConnectionCreateContext(NULL);
    lock = dsLockCreateContext(NULL);
    if (lock) {
	err = dsLockInitContext(lock);
	if (err) {
	    dsLockDestroyContext(lock);
	    lock = NULL;
	}
    }
    cfgref = dsConfigCopyReference(config, ref);
    
    timeout = dsDurationCreate(NULL);
    retry_timeout = dsDurationCreate(NULL);
    slow_write_delay = dsDurationCreate(NULL);
    
    if ((!ctx)||(!con)||(!lock)||(!cfgref)||(!timeout)||(!retry_timeout)||(!slow_write_delay)) {
	if (cfgref) dsConfigDestroyReference(config, cfgref);
	if (id) dsExchangeString(NULL, &id, NULL);
	if (con) dsConnectionDestroyContext(con);
	if (lock) dsLockDestroyContext(lock);
	if (timeout) dsDurationDestroy(timeout);
	if (retry_timeout) dsDurationDestroy(retry_timeout);
	if (slow_write_delay) dsDurationDestroy(slow_write_delay);
	if (ctx) free(ctx);
	return NULL;
    }
    
    ctx->cfg = config;
    ctx->cfgref = cfgref;
    ctx->id = id;
    ctx->lock = lock;    
    ctx->con = con;
    ctx->connect = open_function;
    
    if (open_function == COMConnectionOpen)
	ctx->type = CONNECTION_TYPE_UART;
    else if (open_function == NETConnectionOpen)
	ctx->type = CONNECTION_TYPE_UDP;
    else
	ctx->type = CONNECTION_TYPE_UNKNOWN;

    dsDurationImportMicroseconds(timeout, TIMEOUT_RW);
    dsDurationImportMicroseconds(retry_timeout, RETRY_TIMEOUT);
    dsDurationImportMicroseconds(slow_write_delay, SLOW_WRITE_DELAY);
    
    ctx->timeout = timeout;
    ctx->retry_timeout = retry_timeout;
    ctx->slow_write_delay = slow_write_delay;
    ctx->retries = WRITE_RETRIES;
    ctx->disabled_retries = 0;

    ctx->connected = 0;

    ctx->next = NULL;
    
    return ctx;
}

Connection ConnectionFind(Connection ctx, const char *id) {
    if ((!ctx)||(!id)) return NULL;
    
    if (!strcmp(ctx->id, id)) return ctx;
    return ConnectionFind(ctx->next, id);
}

int ConnectionOpen(Connection ctx) {
    int err;
    DSDuration opendelay;
    
    xmlNodePtr cfg;
    
    ds_assert_param(ctx);
    
    opendelay = dsDurationCreate(NULL);
    if (!opendelay) return DS_ERR_MEM;
    
    dsDurationZeroDuration(opendelay);
    
    if (dsLockTryWrite(ctx->lock)) {
	ConnectionClose(ctx);
	dsLockWrite(ctx->lock);
    }

    err = ctx->connect(ctx);

    if (err) dsLockReturn(ctx->lock);
    else {

	cfg = dsConfigGetConfigByReference(ctx->cfg, ctx->cfgref);
	if (cfg) {
	    dsXmlGetNodeDuration(cfg, "Timeout", ctx->timeout);
	    dsXmlGetNodeDuration(cfg, "RetryTimeout", ctx->retry_timeout);
	    dsXmlGetNodeDuration(cfg, "SlowWriteDelay", ctx->slow_write_delay);
	    dsXmlGetNodeDuration(cfg, "OpenDelay", opendelay);

	    if (ctx->disabled_retries)
		ctx->disabled_retries = dsXmlGetNodeInt(cfg, "Retries", ctx->retries);
	    else 
		ctx->retries = dsXmlGetNodeInt(cfg, "Retries", ctx->retries);
		
	    dsConfigReturnConfig(ctx->cfg, cfg);
	    
	    if (!dsDurationIsZero(opendelay)) dsWait(opendelay);
	}

	dsConnectionSetDefaultFlags(ctx->con, DS_CONNECTION_FLAG_EXACTREAD);
	dsConnectionSetTimeout(ctx->con, ctx->timeout);
	ctx->connected = 1;
    }

    dsDurationDestroy(opendelay);

    return err;
}

int ConnectionClose(Connection ctx) {
    ds_assert_param(ctx);

    dsConnectionDisconnect(ctx->con);
    ctx->connected = 0;
    dsLockReturn(ctx->lock);
    
    return 0;
}

void ConnectionDestroy(Connection ctx) {
    ds_assert_return(ctx);
    
    ConnectionClose(ctx);
    dsConnectionDestroyContext(ctx->con);
    dsDurationDestroy(ctx->timeout);
    dsDurationDestroy(ctx->retry_timeout);
    dsDurationDestroy(ctx->slow_write_delay);
    dsLockDestroyContext(ctx->lock);
    dsConfigDestroyReference(ctx->cfg, ctx->cfgref);
    dsExchangeString(NULL, &ctx->id, NULL);
    free(ctx);
}

int ConnectionSlowWrite(Connection ctx, const void *buf, size_t len) {
    int err;
    unsigned int i;

    if (!ctx->slow_write_delay) return ConnectionWrite(ctx, buf, len);
    
    for (i=0;i<len;i++) {
	err = dsConnectionWrite(ctx->con,buf+i,1,0,NULL);
	if (err) return err;
	dsWaitRT(ctx->slow_write_delay);
    }

    return 0;
}

int ConnectionDisableRetries(Connection ctx) {
    ds_assert_param(ctx);
    
    if (!ctx->disabled_retries) {
	ctx->disabled_retries = ctx->retries;
	ctx->retries = 1;
    }
    
    return 0;
}

int ConnectionSilence(Connection ctx, DSDebugPriority mdl) {
    ds_assert_param(ctx);
    
    ctx->silence = mdl;
    return 0;
}

int ConnectionEnableRetries(Connection ctx) {
    ds_assert_param(ctx);
    
    if (ctx->disabled_retries) {
	ctx->retries = ctx->disabled_retries;
	ctx->disabled_retries = 0;
    }
    
    return 0;
}

int ConnectionReadAll(Connection ctx) {
    int err = 0;
    char tmp[64];
    size_t readed = 0, skipped = 0;

    ds_assert_param(ctx);

    while (!err) {
	skipped += readed;
	err = dsConnectionTryRead(ctx->con, tmp, 64, 0, &readed);
    }

    return skipped;
}

int ConnectionSkipAll(Connection ctx) {
    int err = 0;
    char tmp[64];
    size_t readed = 1, skipped = 0;

    ds_assert_param(ctx);

    while (readed > 0) {
	readed = 0;
	err = dsConnectionRead(ctx->con, tmp, 64, 0, &readed);
	skipped += readed;
    }

    return skipped;
}
