#ifndef _DEVICE_H_TYPE
#define _DEVICE_H_TYPE

enum DeviceTypeT {
    DEVICE_TYPE_UNKNOWN = 0,
    DEVICE_TYPE_NAMT,
    DEVICE_TYPE_NANM,
    DEVICE_TYPE_ASCII,
    DEVICE_TYPE_SEVAN,
    DEVICE_TYPE_NMV
};
typedef enum DeviceTypeT DeviceType;


#ifdef _NEED_DEVICE_NAMES
struct DeviceConfigurationNameT {
    DeviceType type;
    const char *id;
};

typedef struct DeviceConfigurationNameT *DeviceConfigurationName;
typedef struct DeviceConfigurationNameT DeviceConfigurationNameS;

DeviceConfigurationNameS device_configuration_name[] = {
    { DEVICE_TYPE_NAMT, "NAMT" },
    { DEVICE_TYPE_NANM, "NANM" },
    { DEVICE_TYPE_ASCII, "ASCII" },
    { DEVICE_TYPE_SEVAN, "SEVAN" },
    { DEVICE_TYPE_NMV, "NMV" },
    { 0, NULL }
};
#endif /* _NEED_DEVICE_NAMES */

#endif /* _DEVICE_H_TYPE */

#ifndef _DEVICE_H
#define _DEVICE_H

#include <dslock.h>
#include <dscondition.h>
#include <dsconfig.h>
#include <dsthread.h>
#include <dsdatagramserver.h>

typedef struct DeviceT *Device;

#include "commands.h"
#include "buffer.h"
#include "connection/connection.h"
#include "writer/writer.h"


typedef Device (*DeviceCreateFunction)(DSConfigContext config, DSConfigReference ref);
typedef void (*DeviceDestroyFunction)(Device dev);

typedef int (*DeviceConfigureFunction)(Device dev);
typedef int (*DeviceReaderFunction)(Device dev);
typedef int (*DeviceProcessorFunction)(Device dev, ReaderBuf buf);

typedef void (*DeviceResetFunction)(Device dev);
typedef int (*DeviceApplyOperationModeFunction)(Device dev);
typedef int (*DeviceDebugFunction)(Device dev);
typedef char* (*DeviceControlFunction)(Device dev, const char *request);
typedef int (*DeviceSetConfigurationFunction)(Device dev, xmlDocPtr config);
typedef int (*DeviceGetDriverInfoFunction)(Device dev, xmlNodePtr node);

typedef WriterInstance *DeviceWriter;

enum DeviceOperationModeT {
    DEVICE_OPERATION_STANDARD_MODE = 0,
    DEVICE_OPERATION_CONFIGURATION_MODE = 0x10000
};
typedef enum DeviceOperationModeT DeviceOperationMode;

enum DeviceStatusT {
    DEVICE_STATUS_OK = 0,
    DEVICE_STATUS_NOT_CONNECTED,
    DEVICE_STATUS_INITIALIZING,
    DEVICE_STATUS_RESETING
};
typedef enum DeviceStatusT DeviceStatus;


typedef struct DeviceControlRequestT *DeviceControlQueue;

struct DeviceControlRequestT {
    ADASCommand cmd;
    uint32_t reqid;
    char *req;
    DSDatagramServerContext dsc;
    DSConnectionAddress addr;
    
    DeviceControlQueue next;
};
typedef struct DeviceControlRequestT *DeviceControlRequest;
typedef struct DeviceControlRequestT DeviceControlRequestS;



struct DeviceT {
    Device next;
    DSLockContext lock;
    
    char *id;
    DeviceType type;
    unsigned short int version;
    DeviceStatus status;
    
    DSThreadContext reader_thread;
    DSThreadContext control_thread;

    DSConfigContext cfg;
    DSConfigReference cfgref;

    Connection *ccon;		/* Control Connection */    
    Connection *con;		/* Data Connection */
    Writer *wr;
    Buffer buffer;

    unsigned char sf;
    
    DeviceControlQueue cq, cqr;	/* Control Requests Queue (Last, First) */
    DeviceControlRequest cr;	/* Control Request currently processing */
    DSLockContext cqlock;
    DSConditionContext cqcond;
    unsigned int cqsize;

    DeviceOperationMode operation_mode;
    DeviceOperationMode saved_operation_mode;
    DSTime operation_mode_finishing;
    DSDuration multimessage_delay;
    
    Connection curcon, curccon;
    char *text_separator;
    unsigned long max_text_size;
    
    DeviceDestroyFunction destroy;
    DeviceConfigureFunction configure;
    DeviceReaderFunction reader;
    DeviceProcessorFunction processor;
    DeviceResetFunction reset;
    DeviceDebugFunction debug;
    DeviceControlFunction control;
    DeviceSetConfigurationFunction setconfig;
    DeviceGetDriverInfoFunction getdrvinfo;
    DeviceApplyOperationModeFunction applymode;
    
    DSDatagramServerContext dsc;
    DSDatagramServerContext debug_dsc;
};

typedef struct DeviceT DeviceS;


Device DeviceCreate(DSConfigContext config, DSConfigReference ref, Connection con, Writer wr, Buffer buf);
int DeviceSetDSC(Device dev, DSDatagramServerContext dsc);
int DeviceSetDebugDSC(Device dev, DSDatagramServerContext dsc);
int DeviceSetControlMessageDelay(Device dev, DSDuration mmdelay);
void DeviceDestroy(Device dev);
void DeviceSetDefaults(Device dev);

const char *DeviceGetDeviceID(Device dev);
const char *DeviceGetDeviceName(Device dev);

int DeviceLock(Device dev);
void DeviceUnlock(Device dev);

int DeviceStart(Device dev);
void DeviceStop(Device dev);
void DeviceWaitTermination(Device dev);
int DeviceContinueLoop(Device dev);


int DeviceConfigure(Device dev);

void DeviceReset(Device dev);
int DeviceDebug(Device dev);
char *DeviceControl(Device dev, const char *request);
int DeviceQueueControl(Device dev, ADASCommand cmd, uint32_t reqid, const char *req, DSDatagramServerContext dsc, DSConnectionAddress *request_addr);
int DeviceSetConfiguration(Device dev, const char *config);
int DeviceSetOperationMode(Device dev, DeviceOperationMode mode, DSDuration duration);
int DeviceManageOperationMode(Device dev);

Connection DeviceInitializeConnection(Device dev);
Connection DeviceGetConnection(Device dev);
Connection DeviceTryConnection(Device dev);
void DeviceReturnConnection(Device dev);
void DeviceCloseConnection(Device dev);

Connection DeviceInitializeControlConnection(Device dev);
Connection DeviceGetControlConnection(Device dev);
Connection DeviceTryControlConnection(Device dev);
void DeviceReturnControlConnection(Device dev);
void DeviceCloseControlConnection(Device dev);

#define TEXT_READER_MAX_STRING_SIZE		512
#define TEXT_READER_RECORD_SEPARATOR		"\n"
char *DeviceReadText(Device dev, Connection con, const char *separator);


DeviceWriter DeviceInitWriters(Device dev, const char *instid);
void DeviceFreeWriters(Device dev, DeviceWriter dw);
int DeviceWritersReset(Device dev, DeviceWriter dw);
void DeviceWritersStartBlock(Device dev, DeviceWriter dw);
void DeviceWritersFinishBlock(Device dev, DeviceWriter dw);
void DeviceWritersEnableFixedConfigOptimization(Device dev, DeviceWriter dw);
void DeviceWritersEnableFixedLayoutOptimization(Device dev, DeviceWriter dw, const char *suffix);
int DeviceOpenWriters(Device dev, DeviceWriter dw, const char *suffix, DSTime time);
void DeviceCloseWriters(Device dev, DeviceWriter dw);
int DeviceWritersSetQuality(Device dev, DeviceWriter dw, double quality);
int DeviceWritersSetDuration(Device dev, DeviceWriter dw, DSDuration duration);
int DeviceWritersSetDurationMilliseconds(Device dev, DeviceWriter dw, unsigned long duration);
int DeviceWrite(Device dev, DeviceWriter dw, WriterDataType type, WriterDataOptions options, WriterLayerID layerid, WriterDataID n, WriterDataID *id, WriterData data);

#ifdef _DEVICE_C
struct DeviceConfigurationT {
    DeviceType type;
    const char *id;
    DeviceCreateFunction create;
};

typedef struct DeviceConfigurationT *DeviceConfiguration;
typedef struct DeviceConfigurationT DeviceConfigurationS;

#include "namt/namt.h"
#include "nanm/nanm.h"
#include "ascii/ascii.h"
#include "sevan/sevan.h"
#include "nmv/nmv.h"
DeviceConfigurationS device_configuration[] = {
    { DEVICE_TYPE_NAMT, "NAMT", namtDeviceCreate },
    { DEVICE_TYPE_NANM, "NANM", nanmDeviceCreate },
    { DEVICE_TYPE_ASCII, "ASCII", asciiDeviceCreate },
    { DEVICE_TYPE_SEVAN, "SEVAN", sevanDeviceCreate },
    { DEVICE_TYPE_NMV, "NMV", nmvDeviceCreate },
    { 0, NULL, NULL }
};

#endif /* _DEVICE_C */

typedef uint32_t DeviceConfgurationSection;
#define DEVICE_CONFIGURATION_CONFIGURATION 1
#define DEVICE_CONFIGURATION_PROCESSOR 2
#define DEVICE_CONFIGURATION_READOUT 4
#define DEVICE_CONFIGURATION_WRITEOUT 8
#define DEVICE_CONFIGURATION_LAYOUT 16
#define DEVICE_CONFIGURATION_ALL 0xFF
xmlNodePtr DeviceGetConfig(Device dev, DeviceConfgurationSection sections);
xmlNodePtr DeviceGetDriverInfo(Device dev);

#endif /* _DEVICE_H */
