#define _DS_INCLUDE_XML_HEADERS
#define _FINAL_C

#include <stdio.h>
#include <string.h>

#include <dslib.h>
#include <dsconfig.h>
#include <dssystemlock.h>
#include <dssystemservice.h>
#include <dsdebug.h>
#include <dserror.h>

#include <dstime.h>

#include "buffer.h"
#include "connection/connection.h"
#include "writer/writer.h"
#include "device.h"

#include "namt/namt.h"
#include "nanm/nanm.h"
#include "sevan/sevan.h"
#include "ascii/ascii.h"
#include "nmv/nmv.h"


#include "error.h"
#include "defaults.h"
#include "commands.h"

#include "final.h"

static char RunFlag = 1;

static void signal_exit_handler(int signo) { 
    RunFlag = 0; 
}

void Abort() {
    RunFlag = 0;
}

main(int argc, char *argv[]) {
    int err, suberr;

    int i;
    
    int RestartFlag = 0;
    
    DSConfigContext config;    
    DSSystemLockContext syslock;
    Connection con, connections = NULL;
    Writer wr, writers = NULL;
    Device device, devices = NULL;
    Buffer buffer = NULL;
    
    unsigned long refsetsize;
    DSConfigReferenceSet refset;
    DSConfigReference ref;
    xmlNodePtr *nodeset;
    xmlNodePtr node;
    
    unsigned int control_port;
    DSDuration reanouncement_timeout = NULL;
    DSDuration debug_reanouncement_timeout = NULL;
    
    DSDatagramServerContext dsc = NULL;
    DSDatagramServerContext debug_dsc = NULL;
    
    DSTime curtime = NULL;
    DSDuration cstimeout = NULL;
    DSDuration operation_mode_duration = NULL;
    DSDuration multimessage_delay = NULL;
    char datestr[64];
    
    unsigned char dsc_request = ADAS_COMMAND_SUBSCRIBE_BASE;
    unsigned char dsc_debug_request = ADAS_COMMAND_DEBUG_BASE;
    
    char *membuf, *membufptr;
    size_t membufsize;

    size_t readed;
    DSConnectionAddress request_addr;
    size_t request_buffer_size = 16384;
    unsigned char request_buffer[request_buffer_size+1], *request_buffer_pos;
    size_t id_size, prop_size;
    
    unsigned char buffer_thread = DEFAULT_SEPARATE_BUFFER_THREAD;
    
    const char *debug_method;

    if ((argc<1)||(argc>2)) {
	puts("Usage: me [daemon]");
	exit(1);
    }
    
    if ((argc>1)&&(argv[1][0]!='0')) {
	DaemonFlag=1;
	dsSystemServiceDaemonize();
    }

    dsLibInit();    
    xmlKeepBlanksDefault(0);
    dsDebugSetLevel(DEFAULT_DEBUG_LEVEL);

    syslock = dsSystemLockCreateContext(NULL);
    if (syslock) err = dsSystemLockInitContext(syslock, "adas");
    else err = DS_ERR_MEM;

    if (!err) err = dsSystemLockAcquire(syslock);
    if (err) {
	if (syslock) dsSystemLockDestroyContext(syslock);
	if (err == DS_ERR_BUSY) dsDebugServer(DS_DEBUG_CRITICAL, "Application already running!");
	else dsDebugServer(DS_DEBUG_CRITICAL, "Failed to creat lock file (%s)!", dsError(err, NULL));
	dsLibFree();
	return;
    }

    dsSignalINTSetAction(signal_exit_handler);
    dsSignalINTBlock();
    
    curtime = dsTimeCreate(NULL);
    cstimeout = dsDurationCreate(NULL);
    reanouncement_timeout = dsDurationCreate(NULL);
    debug_reanouncement_timeout = dsDurationCreate(NULL);
    operation_mode_duration = dsDurationCreate(NULL);
    multimessage_delay = dsDurationCreate(NULL);
    
    if ((!curtime)||(!cstimeout)||(!reanouncement_timeout)||(!debug_reanouncement_timeout)||(!operation_mode_duration)||(!multimessage_delay)) err = DS_ERR_MEM;
    else {
	dsDurationImportMicroseconds(cstimeout, HILEVEL_SLEEP);
	dsDurationSet(reanouncement_timeout, DEFAULT_CLIENT_TIMEOUT);
	dsDurationSet(debug_reanouncement_timeout, DEFAULT_DEBUGCLIENT_TIMEOUT);
	dsDurationImportSeconds(operation_mode_duration, DEFAULT_CONFIGURATION_MODE_DURATION);
	dsDurationImportMicroseconds(multimessage_delay, DEFAULT_CONTROL_MULTIMESSAGE_DELAY);
    }

RestartPosition:

    config = dsConfigCreateContext(NULL);
    if (config) err = dsConfigInitContext(config, "adas.cfg");
    else err = DS_ERR_MEM;

    if (!err) {
	if (DaemonFlag) err = dsConfigCustomLoad(config, DS_CONFIG_SKIP_LOCAL);
	else err = dsConfigLoad(config);
    }
    
    if (!err) err = dsConfigEnableAutosetDefaults(config);

    if (!err) {
	node = dsConfigGetConfig(config, NULL, "/ADAS/Configuration");
	if (node) {
	    dsDebugSetLevel(dsXmlGetNodeDebugLevel(node, "DebugLevel", DS_DEBUG_WARNING));
	    debug_method = dsXmlGetNodeString(node, "LogFile", LOG_FILE);
	    if ((DaemonFlag)&&(!strcasecmp(debug_method, "STDOUT"))) dsDebugSetMethod(DS_DEBUG_METHOD_FILE, LOG_FILE);
	    else dsDebugSetMethod(DS_DEBUG_METHOD_FILE, debug_method);
	    control_port = dsXmlGetNodeInt(node, "ControlPort", DEFAULT_CONTROL_PORT);
	    dsXmlGetNodeDuration(node, "ReanouncemntTimeout", reanouncement_timeout);
	    dsXmlGetNodeDuration(node, "DebugReanouncemntTimeout", debug_reanouncement_timeout);
	    dsXmlGetNodeDuration(node, "ControlMultiMessageDelay", multimessage_delay);
	    buffer_thread = dsXmlGetNodeBoolean(node, "SeparateBufferThread", buffer_thread);
	    dsConfigReturnConfig(config, node);
	} else err = DS_ERR_CONFIG;
	dsDebugSetLogTimeOption(1);
    }

    buffer = NULL;
    if (!err) {
	node = dsConfigGetConfig(config, NULL, "/ADAS/Buffer");
	if (node) {
	    ref = dsConfigGetConfigReference(config, node);
	    dsConfigReturnConfig(config, node);

	    if (ref) buffer = BufferCreate(config, ref);
	    else DS_ERR_CONFIG;
	} else err = DS_ERR_CONFIG;
    }	
    
    if (!err) {
	dsc = dsDatagramServerCreateContext(NULL);
	if (dsc) {
	    err = dsDatagramServerInitContext(dsc, NULL, control_port);
	    if (!err) {
		dsDatagramServerSetExpirationDuration(dsc, reanouncement_timeout);
		dsDatagramServerSetMaxManagerIterations(dsc, MAX_DATAGRAMSERVER_MANAGER_ITERATIONS);
	    }
	} else err = DS_ERR_MEM;
    } else dsc = NULL;

    if (!err) {
	debug_dsc = dsDatagramServerCreateContext(NULL);
	if (debug_dsc) {
	    err = dsDatagramServerInitSlaveContext(debug_dsc, dsc);
	    if (!err) {
		dsDatagramServerSetExpirationDuration(debug_dsc, debug_reanouncement_timeout);
		dsDatagramServerSetMaxManagerIterations(debug_dsc, MAX_DATAGRAMSERVER_MANAGER_ITERATIONS);
	    }
	} else err = DS_ERR_MEM;
    } else debug_dsc = NULL;

    
    connections = NULL;
    if (!err) {
	nodeset = dsConfigGetConfigs(config, NULL, "/ADAS/Connections/*[@id]");
	if (nodeset) {
	    refset = dsConfigGetConfigsReferences(config, nodeset);
	    dsConfigReturnConfigs(config, nodeset);
	} else refset = NULL;	    

	if (refset) {
	    for (i=0;refset[i];i++) {
		con = ConnectionCreate(config, refset[i]);
		if (con) {
		    con->next = connections;
		    connections = con;
		} else {
		    node = dsConfigGetConfigByReference(config, refset[i]);
		    if (node) {
			dsDebugServer(DS_DEBUG_ERROR, "Connection initialization failed: %s(%s)", dsXmlGetPropString(node, "id", ""), node->name);
			dsConfigReturnConfig(config, node);
		    } else dsDebugServer(DS_DEBUG_ERROR, "Connection initialization failed: position=%u", i);
		}
	    }
	    dsConfigDestroyReferences(config,refset);
	}
    }

    writers = NULL;
    if (!err) {
	nodeset = dsConfigGetConfigs(config, NULL, "/ADAS/Writers/*[@id]");
	if (nodeset) {
	    refset = dsConfigGetConfigsReferences(config, nodeset);
	    dsConfigReturnConfigs(config, nodeset);
	} else refset = NULL;	    

	if (refset) {
	    for (i=0;refset[i];i++) {
		wr = WriterCreate(config, refset[i]);
		if (wr) {
		    wr->next = writers;
		    writers = wr;
		} else {
		    node = dsConfigGetConfigByReference(config, refset[i]);
		    if (node) {
			dsDebugServer(DS_DEBUG_ERROR, "Writer initialization failed: %s(%s)", dsXmlGetPropString(node, "id", ""), node->name);
			dsConfigReturnConfig(config, node);
		    } else dsDebugServer(DS_DEBUG_ERROR, "Writer initialization failed: position=%u", i);
		}
	    }
	    dsConfigDestroyReferences(config, refset);
	}
    }


    devices = NULL;
    if (!err) {
	nodeset = dsConfigGetConfigs(config, NULL, "/ADAS/Devices/*[@id]");
	if (nodeset) {
	    refset = dsConfigGetConfigsReferences(config, nodeset);

	    for (refsetsize=0;refset[refsetsize];refsetsize++) {
		if (dsXmlGetPropBoolean(nodeset[refsetsize], "disabled", 0)) {
		    refset[refsetsize] = NULL;
		}
	    }
	    
	    dsConfigReturnConfigs(config, nodeset);
	} else refset = NULL;	    

	if (refset) {
	    for (i=0;i<refsetsize;i++) {
		if (refset[i]) {
		    device = DeviceCreate(config, refset[i], connections, writers, buffer);
		    if (device) {
			DeviceSetDSC(device, dsc);
			DeviceSetDebugDSC(device, debug_dsc);
			DeviceSetControlMessageDelay(device, multimessage_delay);
		    } else {
			node = dsConfigGetConfigByReference(config, refset[i]);
			if (node) {
			    dsDebugServer(DS_DEBUG_ERROR, "Device initialization failed: %s(%s)", dsXmlGetPropString(node, "id", ""), node->name);
			    dsConfigReturnConfig(config, node);
			} else dsDebugServer(DS_DEBUG_ERROR, "Device initialization failed: position=%u", i);
		    }
		} else device = NULL;
		
		if (device) {
		    device->next = devices;
		    devices = device;
		}
	    }
	    dsConfigDestroyReferences(config, refset);
	}
	
	if (!devices) err = ADAS_ERR_CONFIG_NO_DEVICES;
    }
    

    if ((!err)&&(DaemonFlag)) {
	dsTimeCurrentTime(curtime);
	dsTimeGetDateTime(curtime, datestr);
    	dsDebugServer(DS_DEBUG_WARNING, "Started at: %s", datestr);
    }

    if ((!err)&&(buffer_thread)) {
	err = BufferStartLoop(buffer);
	if (!err) dsDebugServer(DS_DEBUG_INFO, "Buffer. Buffer thread started");
    }

    if (!err) {
	for (device=devices;device;device=device->next) {
	    DeviceStart(device);
	    dsDebugServer(DS_DEBUG_INFO, "Device(%s). Device started", device->id);
	}
    }

    dsSignalINTUnblock();

    while ((RunFlag)&&(!err)) {
	if (buffer_thread) dsWait(cstimeout);
	else {
	    BufferWait(buffer, cstimeout);

	    err = BufferProcess(buffer);
	    if (err) {
		dsDebugServer(DS_DEBUG_ERROR, "Proccessing failed: %lx(%s)", err, dsError(err, NULL));
		err = 0;
	    } 
	}

	for (device = devices; device; device=device->next)
	    DeviceManageOperationMode(device);

	err = dsDatagramServerManageClients(dsc, (char*)&dsc_request, 1);
	if (err == DS_ERR_NOTSUPPORTED) err = dsDatagramServerManageClients(debug_dsc, (char*)&dsc_debug_request, 1);

	if (err == DS_ERR_NOTSUPPORTED) {
	    err = dsDatagramServerRead(dsc, &request_addr, request_buffer, request_buffer_size, 0, &readed);
	    if ((!err)&&(readed)) {
		switch (request_buffer[0]) {
		    case ADAS_COMMAND_GET_INFO:
			// cmd(1),devices(1),version(3: major, minor, micro), status
			// server time: seconds(4), microseconds(4), offset(4)
			// (name_size(1),properties_size(1),device(2),version(2),status(4),name,properties)
			// NAMT v.1-2: Properties: subdevices(1),(detectors(1))
			// NAMT v.3: Properties: hw_type(2), hw_version(2), subdevices(1), (detectors|channels(2))
			// NANM Properties: detectors(1)
			// ASCII Properties: channels(2)
			
			request_buffer[0] = ADAS_COMMAND_GET_INFO;
			request_buffer[1] = 0;
			request_buffer[2] = VERSION_MAJOR;
			request_buffer[3] = VERSION_MINOR;
			request_buffer[4] = VERSION_MICRO;
			request_buffer[5] = 0;
			
			dsTimeCurrentTime(curtime);
			dsTimeGetDateTime(curtime, datestr);
			strcpy(request_buffer + 6, datestr);

			request_buffer_pos = request_buffer + 6 + strlen(datestr) + 1;
			for (device = devices; device; device=device->next) {
			    id_size = strlen(device->id);
			    if (id_size>0xFF) id_size = 0xFF;
			    
			    switch (device->type) {
				case DEVICE_TYPE_NAMT:
				    if (dsLockTryRead(((NAMTDevice)device)->ilock))
					prop_size = 5; /* devices can't be equal to 0 */
				    else 
					prop_size = 5 + 2 * ((NAMTDevice)device)->devices;
				break;
				case DEVICE_TYPE_NANM:
				    prop_size = 1;
				break;
				case DEVICE_TYPE_ASCII:
				    prop_size = sizeof(uint16_t);
				break;
				case DEVICE_TYPE_SEVAN:
				    prop_size = 2 * sizeof(uint16_t);
				break;
				case DEVICE_TYPE_NMV:
				    prop_size = sizeof(uint16_t);
				break;
				default:
				    prop_size = 0;
			    }
			    
			    if ( (request_buffer_pos - request_buffer) + id_size + prop_size + 10 > request_buffer_size) {
				dsDebugServer(DS_DEBUG_ERROR, "Error proccesing client request. Buffer is to small to handle all working devices.");
				break;
			    }
			    
			    *request_buffer_pos = (unsigned char)id_size; request_buffer_pos++;
			    *request_buffer_pos = (unsigned char)prop_size; request_buffer_pos++;
			    *(uint16_t*)request_buffer_pos = htons(device->type); request_buffer_pos += sizeof(uint16_t);
			    *(uint16_t*)request_buffer_pos = htons(device->version); request_buffer_pos += sizeof(uint16_t);
			    *(uint32_t*)request_buffer_pos = htonl(device->status|device->operation_mode); request_buffer_pos += sizeof(uint32_t);
			    memcpy(request_buffer_pos, device->id, id_size); request_buffer_pos += id_size;
			    if (prop_size) {
				switch (device->type) {
				    case DEVICE_TYPE_NAMT:
				    	*(uint16_t*)(request_buffer_pos) = htons(((NAMTDevice)device)->device_type); request_buffer_pos+=2;
					*(uint16_t*)(request_buffer_pos) = htons(((NAMTDevice)device)->device_version); request_buffer_pos+=2;

					if (prop_size > 5) {
					    *request_buffer_pos = (unsigned char)((NAMTDevice)device)->devices; request_buffer_pos++;
					    for (i = 0; i < (unsigned char)((NAMTDevice)device)->devices;i++) {
						*(uint16_t*)(request_buffer_pos) = htons(((NAMTDevice)device)->detectors[i]); 
						request_buffer_pos += 2;
					    }
					    dsLockReturn(((NAMTDevice)device)->ilock);
					} else {
					    *request_buffer_pos = 0; request_buffer_pos++;
					}
				    break;
				    case DEVICE_TYPE_NANM:
					*request_buffer_pos = (unsigned char)((NANMDevice)device)->hosts; request_buffer_pos++;
				    break;
				    case DEVICE_TYPE_ASCII:
					*(uint16_t*)request_buffer_pos = (uint16_t)((ASCIIDevice)device)->channels; request_buffer_pos+=sizeof(uint16_t);
				    break;
				    case DEVICE_TYPE_SEVAN:
					*(uint16_t*)request_buffer_pos = (uint16_t)((SEVANDevice)device)->sensors; request_buffer_pos+=sizeof(uint16_t);
					*(uint16_t*)request_buffer_pos = (uint16_t)((SEVANDevice)device)->channels; request_buffer_pos+=sizeof(uint16_t);
				    break;
				    case DEVICE_TYPE_NMV:
					*(uint16_t*)request_buffer_pos = (uint16_t)((NMVDevice)device)->channels; request_buffer_pos+=sizeof(uint16_t);
				    break;
				    default:
					request_buffer_pos += prop_size;
				}
			    }
			    
			    request_buffer[1]++;
			}
			
			dsDatagramServerWrite(dsc, &request_addr, request_buffer, request_buffer_pos - request_buffer, 0, NULL);
		    break;
		    case ADAS_COMMAND_GET_MCU_INFO:
			// cmd(1), type(2), version(2), status(4), mcu_type(2), mcu_version(2)
			// subdevices(1) [ sensors(2), coincidences(2), spectrums(2) ]
			// flags(4):
			//  LAYOUT: max_sensors(2), max_channels(2), max_rules(2), max_spectrums(2)

			if (readed < request_buffer_size) request_buffer[readed] = 0;
			else request_buffer[request_buffer_size - 1] = 0;

			for (device = devices; device; device=device->next) {
			    if (!strcmp(request_buffer + 1, device->id)) break;
			}
			
			if (device) {
			    request_buffer[0] = ADAS_COMMAND_GET_MCU_INFO;

			    request_buffer_pos = request_buffer + 1;
			    *(uint16_t*)(request_buffer_pos) = htons(device->type); request_buffer_pos+=2;
			    *(uint16_t*)(request_buffer_pos) = htons(device->version); request_buffer_pos+=2;
			    *(uint32_t*)(request_buffer_pos) = htonl(device->status|device->operation_mode); request_buffer_pos+=4;

			    switch (device->type) {
				case DEVICE_TYPE_NANM:
				case DEVICE_TYPE_ASCII:
				case DEVICE_TYPE_SEVAN:
				case DEVICE_TYPE_NMV:
				    *(uint8_t*)(request_buffer_pos) = 1; request_buffer_pos+=1;
				    switch(device->type) {
					case DEVICE_TYPE_NANM:
					    *(uint16_t*)request_buffer_pos = htons(((NANMDevice)device)->hosts); request_buffer_pos+=2;
					    *(uint16_t*)request_buffer_pos = htons(0); request_buffer_pos+=2;
					break;
					case DEVICE_TYPE_ASCII:
					    *(uint16_t*)request_buffer_pos = htons(((ASCIIDevice)device)->channels); request_buffer_pos+=2;
					    *(uint16_t*)request_buffer_pos = htons(0); request_buffer_pos+=2;
					break;
				        case DEVICE_TYPE_SEVAN:
				    	    *(uint16_t*)request_buffer_pos = htons(((SEVANDevice)device)->sensors); request_buffer_pos+=2;
					    *(uint16_t*)request_buffer_pos = htons(((SEVANDevice)device)->channels - ((SEVANDevice)device)->sensors); request_buffer_pos+=2;
					break;
					case DEVICE_TYPE_NMV:
					    *(uint16_t*)request_buffer_pos = htons(((NMVDevice)device)->channels); request_buffer_pos+=2;
					    *(uint16_t*)request_buffer_pos = htons(0); request_buffer_pos+=2;
					break;
				    }
				    *(uint16_t*)request_buffer_pos = htons(0); request_buffer_pos+=2; /* spectrum */

				    *(uint16_t*)(request_buffer_pos) = htons(0); request_buffer_pos+=2;
				    *(uint16_t*)(request_buffer_pos) = htons(0); request_buffer_pos+=2;
				    *(uint32_t*)(request_buffer_pos) = htonl(0); request_buffer_pos+=4;
				break;
				case DEVICE_TYPE_NAMT:
				    if (!dsLockTryRead(((NAMTDevice)device)->ilock)) {
					*(uint16_t*)(request_buffer_pos) = htons(((NAMTDevice)device)->device_type); request_buffer_pos+=2;
					*(uint16_t*)(request_buffer_pos) = htons(((NAMTDevice)device)->device_version); request_buffer_pos+=2;

					if (((NAMTDevice)device)->plain) {
					    *(uint8_t*)(request_buffer_pos) = 1; request_buffer_pos+=1;
					    *(uint16_t*)(request_buffer_pos) = htons(((NAMTDevice)device)->detectors[0]); request_buffer_pos+=2;
					    *(uint16_t*)(request_buffer_pos) = htons(((NAMTDevice)device)->coincidences); request_buffer_pos+=2;
					    if (((NAMTDevice)device)->write_spectrum) {
						*(uint16_t*)(request_buffer_pos) = htons(((NAMTDevice)device)->detectors[0]+((NAMTDevice)device)->spectrums); request_buffer_pos+=2;
					    } else {
						*(uint16_t*)(request_buffer_pos) = htons(0); request_buffer_pos+=2;
					    }
					} else {
					    *(uint8_t*)(request_buffer_pos) = ((NAMTDevice)device)->devices; request_buffer_pos+=1;
					    for (i = 0; i < (unsigned char)((NAMTDevice)device)->devices;i++) {
					        *(uint16_t*)(request_buffer_pos) = htons(2*((NAMTDevice)device)->detectors[i]); request_buffer_pos+=2;
						*(uint16_t*)(request_buffer_pos) = htons(((NAMTDevice)device)->detectors[i] * ((NAMTDevice)device)->detectors[i] + 7); request_buffer_pos+=2;
						
						if (((NAMTDevice)device)->write_spectrum) {
						    *(uint16_t*)(request_buffer_pos) = htons(2*((NAMTDevice)device)->detectors[i]); request_buffer_pos+=2;
						} else {
						    *(uint16_t*)(request_buffer_pos) = htons(0); request_buffer_pos+=2;
						}
					    }
					}

					if (((NAMTDevice)device)->layout) {
					    *(uint32_t*)(request_buffer_pos) = htonl(ADAS_HARDWARE_FLAG_CONFIGURABLE_LAYOUT); request_buffer_pos+=4;
					    *(uint16_t*)(request_buffer_pos) = htons(((NAMTDevice)device)->layout_sensors); request_buffer_pos+=2;
					    *(uint16_t*)(request_buffer_pos) = htons(((NAMTDevice)device)->layout_channels); request_buffer_pos+=2;
					    *(uint16_t*)(request_buffer_pos) = htons(((NAMTDevice)device)->layout_rules); request_buffer_pos+=2;
					    *(uint16_t*)(request_buffer_pos) = htons(((NAMTDevice)device)->layout_spectrums); request_buffer_pos+=2;
					} else {
					    *(uint32_t*)(request_buffer_pos) = htonl(0); request_buffer_pos+=4;
					}
					dsLockReturn(((NAMTDevice)device)->ilock);
					break;
				    } 
				/*break;*/
				default:
				    *(uint8_t*)(request_buffer_pos) = 0; request_buffer_pos+=1;
				    *(uint16_t*)(request_buffer_pos) = htons(0); request_buffer_pos+=2;
				    *(uint16_t*)(request_buffer_pos) = htons(0); request_buffer_pos+=2;
				    *(uint32_t*)(request_buffer_pos) = htonl(0); request_buffer_pos+=4;
			    }
			    dsDatagramServerWrite(dsc, &request_addr, request_buffer, request_buffer_pos - request_buffer, 0, NULL);
			}
		    break;
		    case ADAS_COMMAND_RESET:
			if (readed < request_buffer_size) request_buffer[readed] = 0;
			else request_buffer[request_buffer_size - 1] = 0;

			if (!request_buffer[1]) {
			    RunFlag = 0;
			    RestartFlag = 1;
			    break;
			}

			for (device = devices; device; device=device->next) {
			    if (!strcmp(request_buffer + 1, device->id)) {
				DeviceReset(device);
				break;
			    }
			}
		    break;
		    case ADAS_COMMAND_CONFIGURATION_MODE:
			if (readed < request_buffer_size) request_buffer[readed] = 0;
			else request_buffer[request_buffer_size - 1] = 0;
			
			for (device = devices; device; device=device->next) {
			    if (!strcmp(request_buffer + 1, device->id)) {
				DeviceSetOperationMode(device, DEVICE_OPERATION_CONFIGURATION_MODE, operation_mode_duration);
			    }
			}
		    break;
		    case ADAS_COMMAND_GET_CONFIG:
			if (readed < request_buffer_size) request_buffer[readed] = 0;
			else request_buffer[request_buffer_size - 1] = 0;
			if (request_buffer[1]) {
			    for (device = devices; device; device=device->next)
				if (!strcmp(request_buffer + 1, device->id)) break;

			    if (device) {
				node = DeviceGetConfig(device, DEVICE_CONFIGURATION_ALL);
				if (node) {
				    if (!dsXmlDumpMemory(node, &membuf, &membufsize)) {
					request_buffer[0] = 0x03;
					if (membufsize>=request_buffer_size) membufsize=request_buffer_size-1;
					memcpy(request_buffer+1, membuf, membufsize);
					free(membuf);
					dsDatagramServerWrite(dsc, &request_addr, request_buffer, membufsize+1, 0, NULL);
				    }
				    xmlFreeNode(node);
				}
			    }
			} else {
			    if (!dsConfigDumpMemory(config, &membuf, &membufsize)) {
				request_buffer[0] = ADAS_COMMAND_GET_CONFIG;
				if (membufsize>=request_buffer_size) membufsize=request_buffer_size-1;
				memcpy(request_buffer+1, membuf, membufsize);
				free(membuf);
				dsDatagramServerWrite(dsc, &request_addr, request_buffer, membufsize+1, 0, NULL);
			    }
			}
		    break;
		    case ADAS_COMMAND_CONFIGURE:
			request_buffer[readed] = 0;
			if (strlen(request_buffer+1) + 2 >= readed) break;
			for (device = devices; device; device=device->next) {
			    if (!strcmp(request_buffer + 1, device->id)) {
				suberr = DeviceSetConfiguration(device, request_buffer+1+strlen(request_buffer+1) + 1);
				*((int*)(request_buffer+1))=suberr;
				dsDatagramServerWrite(dsc, &request_addr, request_buffer, sizeof(int)+1, 0, NULL);
				break;
			    }
			}
			
			*((int*)(request_buffer+1))=DS_ERR_NOTFOUND;
			dsDatagramServerWrite(dsc, &request_addr, request_buffer, sizeof(int)+1, 0, NULL);
		    break;
		    case ADAS_COMMAND_CONTROL:
			    // cmd(1), flags(1), id(2), reqid(4), data
			request_buffer[readed] = 0;
			if (strlen(request_buffer+1) + 6 >= readed) break;
			request_buffer_pos = request_buffer + 1 + strlen(request_buffer + 1) + 1;
			for (device = devices; device; device=device->next) {
			    if (!strcmp(request_buffer + 1, device->id)) {
				DeviceQueueControl(device, *request_buffer, *(uint32_t*)(request_buffer_pos), request_buffer_pos + 4, dsc, &request_addr);
				break;
			    }
			}
		    break;
		    case ADAS_COMMAND_SHUTDOWN: /* Shutdown */
		        RunFlag = 0; 
		    break;
		    default:
			dsDebugServer(DS_DEBUG_WARNING, "Unsupported request from the client: %lX", request_buffer[0]);
		}
	    } else {
		dsDebugServer(DS_DEBUG_WARNING, "Error reading client request: %lx(%s)", err, dsError(err, NULL));
	    }
	}

	if (err) {
	    dsDebugServer(DS_DEBUG_WARNING, "Client processing failed: %lx(%s)", err, dsError(err, NULL));
	    err = 0;
	}
	
	dsDatagramServerManage(dsc);
	dsDatagramServerManage(debug_dsc);

	for (device = devices; device; device = device->next) {
	    err = DeviceDebug(device);
	    if (err) {
		if (err != DS_ERR_BUSY) dsDebugServer(DS_DEBUG_NOTICE, "Device(%s). Debugging request failed: %lx(%s)", device->id, err, dsError(err, NULL));
		err = 0;
	    }
	}
    }

    dsSignalINTBlock();

    for (device=devices;device;device=device->next) DeviceStop(device);
    for (device=devices;device;device=device->next) {
	DeviceWaitTermination(device);
	dsDebugServer(DS_DEBUG_INFO, "Device(%s). Device terminated", device->id);
    }
    
    if (buffer_thread) {
	BufferStopLoop(buffer);
	dsDebugServer(DS_DEBUG_INFO, "Buffer. Buffer thread terminated");
    }

    // Wait for termination
    while (devices) {
	device = devices;
	devices = devices->next;
        DeviceDestroy(device);
    }

    while (writers) {
	wr = writers;
	writers = writers->next;
	WriterDestroy(wr);
    }    

    while (connections) {
	con = connections;
	connections = connections->next;
	ConnectionDestroy(con);
    }    

    if (dsc) dsDatagramServerDestroyContext(dsc);
    
    if (buffer) BufferDestroy(buffer);
    
    if (config) dsConfigDestroyContext(config);

    if (err) dsDebugServer(DS_DEBUG_FATAL, "Error: %lx(%s)", err, dsError(err, NULL));
    else if (DaemonFlag) {
	dsTimeCurrentTime(curtime);
	dsTimeGetDateTime(curtime, datestr);
	dsDebugServer(DS_DEBUG_WARNING, "Succesfully exited at: %s", datestr); 
    }
    
    if (RestartFlag) {
	RunFlag = 1;
	RestartFlag = 0;
	goto RestartPosition;
    }

    if (syslock) {
	dsSystemLockReturn(syslock);
	dsSystemLockDestroyContext(syslock);
    }

    if (multimessage_delay) dsDurationDestroy(multimessage_delay);
    if (operation_mode_duration) dsDurationDestroy(operation_mode_duration);
    if (debug_reanouncement_timeout) dsDurationDestroy(debug_reanouncement_timeout);
    if (reanouncement_timeout) dsDurationDestroy(reanouncement_timeout);
    if (cstimeout) dsDurationDestroy(cstimeout);
    if (curtime) dsTimeDestroy(curtime);

    dsLibFree();
}
