#include <stdio.h>

#include <dsassert.h>
#include <dsdebug.h>

#include "../namt/fujimc/Src/commands.h"

#include "lowlevel.h"


#define RETRY_REQUEST(res) \
	if (--retries) { \
	    if (res == COMMAND_RESPONSE_RETRY) { \
		res = ReadByte(ctx); \
		dsDebugServer(DS_DEBUG_DEBUG, "LowLevel (cmd: %lx, retry %u). Microcode retry received, code: %lx", cmd, ctx->retries - retries, res); \
		dsWait(ctx->retry_timeout); \
	    } else { \
		if (res == COMMAND_RESPONSE_ERROR) { \
		    res = ReadByte(ctx); \
		    if (res < COMMAND_ERROR_UART) { \
			ctx->llerr = res;\
			return COMMAND_RESPONSE_ERROR; \
		    } \
		    dsDebugServer((ctx->silence>DS_DEBUG_NOTICE)?ctx->silence:DS_DEBUG_NOTICE, "LowLevel (cmd: %lx, retry %u). Microcode error detected: %lx", cmd, ctx->retries - retries, res); \
		} else if (res == LL_ERR_TIMEOUT) { \
		    dsDebugServer((ctx->silence>DS_DEBUG_NOTICE)?ctx->silence:DS_DEBUG_NOTICE, "LowLevel (cmd: %lx, retry %u). Microcode command timed out", cmd, ctx->retries - retries); \
		} else dsDebugServer((ctx->silence>DS_DEBUG_NOTICE)?ctx->silence:DS_DEBUG_NOTICE, "LowLevel (cmd: %lx, retry %u). Invalid response from microcode: %lx", cmd, ctx->retries - retries, res); \
		dsWait(ctx->retry_timeout); \
		res = ConnectionSkipAll(ctx); \
		if (res) dsDebugServer((ctx->silence>DS_DEBUG_INFO)?ctx->silence:DS_DEBUG_INFO, "LowLevel (cmd: %lx, retry %u). Skipping %u bytes", cmd, ctx->retries - retries, res); \
	    } \
	    goto restart; \
	} \

#define RETURN(err) { \
    if ((err == COMMAND_RESPONSE_ERROR)||(err == COMMAND_RESPONSE_RETRY)) ctx->llerr = ReadByte(ctx); \
    /*if (((unsigned char)err) != cmd) ConnectionSkipAll(ctx);*/ \
    return err; \
}

#define DebugPriority(ctx, retries) \
    ((retries==ctx->retries)?DS_DEBUG_DEBUG:((ctx->silence>DS_DEBUG_DEBUG)?ctx->silence:DS_DEBUG_DEBUG))

static int ConnectionConvertError(int dserr, int deferr) {
    switch (dserr) {
	case DS_ERR_TIMEOUT:
	    return LL_ERR_TIMEOUT;
	case DS_ERR_NOTIMPLEMENTED:
	    return LL_ERR_IMPLEMENTED;
	case DS_ERR_SOCKET_READ:
	    return LL_ERR_READ;
	case DS_ERR_SOCKET_WRITE:
	    return LL_ERR_WRITE;
	case DS_ERR_MEM:
	    return LL_ERR_MEM;
	case DS_ERR_INVALID_ARGUMENT:
	case DS_ERR_INVALID_OPTIONS:
	    return LL_ERR_INIT;
    }
    return deferr;
}

int ReadByte(Connection ctx) {
    int err;
    unsigned char tmp;

    ds_assert_err(ctx, LL_ERR_INIT);

    err = ConnectionRead(ctx,&tmp,1);
    if (err) return ConnectionConvertError(err, LL_ERR_READ);

    return tmp;
}


int ReadWord(Connection ctx) {
    int err;
    unsigned short int tmp;

    ds_assert_err(ctx, LL_ERR_INIT);

    err = ConnectionRead(ctx,(char*)&tmp,2);
    if (err) return ConnectionConvertError(err, LL_ERR_READ);

    return tmp;
}

int WriteByte(Connection ctx, unsigned char tmp) {
    int err;

    ds_assert_err(ctx, LL_ERR_INIT);

    err = ConnectionWrite(ctx,&tmp,1);
    if (err) return ConnectionConvertError(err, LL_ERR_WRITE);

    return 0;
}

int WriteWord(Connection ctx, unsigned short int tmp) {
    int err;

    ds_assert_err(ctx, LL_ERR_INIT);
    
    err = ConnectionWrite(ctx,(char*)&tmp,2);
    if (err) return ConnectionConvertError(err, LL_ERR_WRITE);
    
    return 0;
}

int SendSimple(Connection ctx, unsigned char cmd) {
    int err;
    int retries;

    ds_assert_err(ctx, LL_ERR_INIT);

    retries = ctx->retries;
    
restart:
    dsDebugServer(DebugPriority(ctx, retries), "LowLevel. Send Simple:  %lx, retry %u", cmd, ctx->retries - retries + 1);

    err = WriteByte(ctx, cmd);
    if (err<0) return err;

    err = ReadByte(ctx);
    if (err != cmd) {
	RETRY_REQUEST(err)
	RETURN(err)
    }

    return err;
}

int SendAttribute(Connection ctx, unsigned char cmd, unsigned short int attr) {
    int err;
    int retries;

    ds_assert_err(ctx, LL_ERR_INIT);

    retries = ctx->retries;

restart:
    
    dsDebugServer(DebugPriority(ctx, retries), "LowLevel. Send Attribute:  %lx(%u), retry %u", cmd, attr, ctx->retries - retries + 1);
    
    err = WriteByte(ctx, cmd);
    if (err<0) return err;

    err = WriteWord(ctx, attr);
    if (err<0) RETURN(err);
    
    err = ReadByte(ctx);
    if (err != cmd) {
	RETRY_REQUEST(err)
	RETURN(err)
    }

    return err;
}

int SendData(Connection ctx, unsigned char cmd, unsigned short int isize, const char *ibuff) {
    int err;
    int retries;

    ds_assert_err(ctx, LL_ERR_INIT);

    retries = ctx->retries;

restart:
    
    dsDebugServer(DebugPriority(ctx, retries), "LowLevel. Send Data: %lx, retry %u", cmd, ctx->retries - retries + 1);
    
    err = WriteByte(ctx, cmd);
    if (err<0) return err;

    err = WriteWord(ctx, isize);
    if (err<0) RETURN(err);

    err = ConnectionWrite(ctx,ibuff,isize);
    if (err) RETURN(ConnectionConvertError(err, LL_ERR_WRITE));

    err = ReadByte(ctx);

    if (err != cmd) {
	RETRY_REQUEST(err)
	RETURN(err)
    }

    return err;
}

int RequestAttribute(Connection ctx, unsigned char cmd, unsigned short int *attr) {
    int res, ares;
    int retries;

    ds_assert_err(ctx, LL_ERR_INIT);

    retries = ctx->retries;

restart:    
    dsDebugServer(DebugPriority(ctx, retries), "LowLevel. Request Attribute:  %lx, retry %u", cmd, ctx->retries - retries + 1);

    res=WriteByte(ctx, cmd);
    if (res<0) return res;
    
    res = ReadByte(ctx);
    if (res!=cmd) {
	RETRY_REQUEST(res)
	RETURN(res)
    }
    ares = ReadWord(ctx);
    if (ares<0) RETURN(ares);
    *attr = ares;

    return res;
}

int RequestData(Connection ctx, unsigned char cmd, unsigned short int *osize, char **obuff) {
    char *tmp;
    int res;
    int size;
    int retries;

    ds_assert_err(ctx, LL_ERR_INIT);

    retries = ctx->retries;

        
restart:    
    dsDebugServer(DebugPriority(ctx, retries), "LowLevel. Request Data:  %lx, retry %u", cmd, ctx->retries - retries + 1);

    res=WriteByte(ctx, cmd);
    if (res<0) return res;

    res = ReadByte(ctx);
    if (res!=cmd) {
	RETRY_REQUEST(res)
	RETURN(res);
    }
    size = ReadWord(ctx);
    if (size<0) RETURN(size);

    tmp = (char*)malloc(size);
    if (!tmp) RETURN(LL_ERR_MEM);

    dsDebugServer(DS_DEBUG_DEBUG, "LowLevel. Reading %lu bytes of data", size);
    
    res=ConnectionRead(ctx,tmp,size);
    if (res) {
	if (tmp!=*obuff) free(tmp);
	RETURN(ConnectionConvertError(res, LL_ERR_READ));
    }
    
    *obuff = tmp;
    *osize = size;

    return cmd;    
}

int RequestDataFixed(Connection ctx, unsigned char cmd, unsigned short int osize, char *obuff) {
    int res;
    int size;
    int retries;

    ds_assert_err(ctx, LL_ERR_INIT);

    retries = ctx->retries;
    
restart:    
    dsDebugServer(DebugPriority(ctx, retries), "LowLevel. Request Data Fixed:  %lx, retry %u", cmd, ctx->retries - retries + 1);

    res=WriteByte(ctx, cmd);
    if (res<0) return res;

    res = ReadByte(ctx);
    if (res!=cmd) {
	RETRY_REQUEST(res)
	RETURN(res);
    }
    size = ReadWord(ctx);
    if ((size<0)||((osize)&&(size!=osize))) RETURN(LL_ERR_READ);
    
    dsDebugServer(DS_DEBUG_DEBUG, "LowLevel. Reading %lu bytes of data", size);

    res = ConnectionRead(ctx,obuff,size);
    if (res) RETURN(ConnectionConvertError(res, LL_ERR_READ));

    return cmd;    
}

int ComplexRequestData(Connection ctx, unsigned cmd, unsigned short int attr, unsigned short int *osize, char **obuff) {
    char *tmp;
    int res;
    int size;
    int retries;

    ds_assert_err(ctx, LL_ERR_INIT);

    retries = ctx->retries;
    
restart:    
    dsDebugServer(DebugPriority(ctx, retries), "LowLevel. Complex Request Data:  %lx(%u), retry %u", cmd, attr, ctx->retries - retries + 1);

    res = WriteByte(ctx, cmd);
    if (res<0) return res;
    
    res = WriteWord(ctx, attr);
    if (res<0) RETURN(res);

    res = ReadByte(ctx);
    if (res!=cmd) {
	if (--retries) {
	    if ((res == COMMAND_RESPONSE_RETRY)||(res == LL_ERR_TIMEOUT)) {
		if (res == COMMAND_RESPONSE_RETRY) {
		    res = ReadByte(ctx);
/*		
		    DS: I don't think it is neccesary.
		    while ((res<0)&&(--retries)) res = ReadByte(ctx);
		    if (!retries) return res;
*/
		}
	    } else {
		/* We can get here for the following reason:
		    1. The request gets very long to process for some reason
		    2. Second request (retry) issue. And then we get answer
		    for first retry. This may cause problems...
		*/
		ConnectionSkipAll(ctx);
	    }
	    dsWait(ctx->retry_timeout);
	    goto restart;
	}
	RETURN(res);
    }

    size = ReadWord(ctx);
    if (size<0) RETURN(size);
    
    tmp = (char*)malloc(size?size:1);
    if (!tmp) RETURN(LL_ERR_MEM);

    dsDebugServer(DS_DEBUG_DEBUG, "LowLevel. Reading %lu bytes of data", size);

    if (size) {
	res = ConnectionRead(ctx,tmp,size);
	if (res) {
	    /*if (tmp!=*obuff)*/ free(tmp);
	    RETURN(ConnectionConvertError(res, LL_ERR_READ));
	}
    }
    
    *obuff = tmp;
    *osize = size;

    return cmd;    
}

int ComplexRequestDataFixed(Connection ctx, unsigned cmd, unsigned short int attr, unsigned short int osize, char *obuff) {
    int res;
    int size;
    int retries;

    ds_assert_err(ctx, LL_ERR_INIT);

    retries = ctx->retries;
    
restart:    
    dsDebugServer(DebugPriority(ctx, retries), "LowLevel. Complex Request Data Fixed:  %lx(%u), retry %u", cmd, attr, ctx->retries - retries + 1);

    res = WriteByte(ctx, cmd);
    if (res<0) return res;
    
    res = WriteWord(ctx, attr);
    if (res<0) RETURN(res);

    res = ReadByte(ctx);
    if (res!=cmd) {
	RETRY_REQUEST(res)
	RETURN(res);
    }
    size = ReadWord(ctx);
    if ((size<0)||((osize)&&(size!=osize))) RETURN(LL_ERR_READ);
    
    dsDebugServer(DS_DEBUG_DEBUG, "LowLevel. Reading %lu bytes of data", size);
    
    if (size) {
	res = ConnectionRead(ctx,obuff,size);
	if (res) RETURN(ConnectionConvertError(res, LL_ERR_READ));
    }

    return cmd;    
}


int ComplexRequest(Connection ctx, unsigned char cmd, unsigned short int isize, const char *ibuff, unsigned short int *osize, char **obuff) {
    int err;
    int res;
    int size;
    int retries;
    char *tmp;

    ds_assert_err(ctx, LL_ERR_INIT);

    retries = ctx->retries;

restart:
    
    dsDebugServer(DebugPriority(ctx, retries), "LowLevel. Send Data: %lx, retry %u", cmd, ctx->retries - retries + 1);
    
    err = WriteByte(ctx, cmd);
    if (err<0) return err;

    err = WriteWord(ctx, isize);
    if (err<0) RETURN(err);

    err = ConnectionWrite(ctx,ibuff,isize);
    if (err) RETURN(ConnectionConvertError(err, LL_ERR_WRITE));

    res = ReadByte(ctx);
    if (res!=cmd) {
	RETRY_REQUEST(res)
	RETURN(res);
    }

    size = ReadWord(ctx);
    if (size<0) RETURN(size);
    
    tmp = (char*)malloc(size?size:1);
    if (!tmp) RETURN(LL_ERR_MEM);

    dsDebugServer(DS_DEBUG_DEBUG, "LowLevel. Reading %lu bytes of data", size);

    if (size) {
	res = ConnectionRead(ctx,tmp,size);
	if (res) {
	    /*if (tmp!=*obuff)*/ free(tmp);
	    RETURN(ConnectionConvertError(res, LL_ERR_READ));
	}
    }
    
    *obuff = tmp;
    *osize = size;

    return cmd;    
}

int SendChannelAttribute(Connection ctx, unsigned char cmd, unsigned short int channel, unsigned short int attr) {
    int err;
    int retries;

    ds_assert_err(ctx, LL_ERR_INIT);

    retries = ctx->retries;

restart:
    
    dsDebugServer(DebugPriority(ctx, retries), "LowLevel. Send Channel Attribute:  %lx(%u,%u), retry %u", cmd, channel, attr, ctx->retries - retries + 1);
    
    err = WriteByte(ctx, cmd);
    if (err<0) return err;

    err = WriteWord(ctx, channel);
    if (err<0) RETURN(err);
    
    err = WriteWord(ctx, attr);
    if (err<0) RETURN(err);
    
    err = ReadByte(ctx);

    if (err != cmd) {
	RETRY_REQUEST(err);
	RETURN(err);
    }

    return err;
}

int RequestChannelAttribute(Connection ctx, unsigned char cmd, unsigned short int channel, unsigned short int *attr) {
    int res, ares;
    int retries;

    ds_assert_err(ctx, LL_ERR_INIT);

    retries = ctx->retries;

restart:    
    dsDebugServer(DebugPriority(ctx, retries), "LowLevel. Request Channel Attribute:  %lx(%u), retry %u", cmd, channel, ctx->retries - retries + 1);

    res=WriteByte(ctx, cmd);
    if (res<0) return res;

    res=WriteWord(ctx, channel);
    if (res<0) RETURN(res);
    
    res = ReadByte(ctx);
    if (res!=cmd) {
	RETRY_REQUEST(res)
	RETURN(res);
    }
    ares = ReadWord(ctx);
    if (ares<0) RETURN(ares);

    *attr = ares;

    return res;
}
