#define _DS_INCLUDE_XML_HEADERS

#include <stdio.h>

#include <dsassert.h>
#include <dsdebug.h>

#include <dsconfig.h>

#include "defaults.h"
#include "final.h"

#include "namtreader.h"
#include "namtanalyzer.h"
#include "namt.h"

static int namtConfigure(Device dev);
static void namtDeviceDestroy(NAMTDevice dev);

void namtDeviceSetDefaults(NAMTDevice dev) {
    DeviceSetDefaults((Device)dev);
    namtZero(dev);
    namtAnalyzerZero(dev, &dev->analyzer);

    dev->device.version = NAMT_VERSION;
    dev->protocol_version = 0;
    dev->device_type = 0;
    dev->device_version = 0;
    dev->handled_events = 0;
    dev->layout = 0;

    dsDurationImportMilliseconds(dev->control_readout_required_time, SNT_DEFAULT_CONTROL_READOUT_REQUIRED_TIME);
    dsDurationImportMilliseconds(dev->high_read_pause, SNT_DEFAULT_HIGH_READ_PAUSE);
    dsDurationImportMilliseconds(dev->min_sync_error, MINIMAL_SYNC_ERROR);
    dsDurationImportMilliseconds(dev->max_sync_error, MAXIMAL_SYNC_ERROR);

    dev->recs = 1;
    dev->write_control = SNT_DEFAULT_WRITE_CONTROL;
}

Device namtDeviceCreate(DSConfigContext config, DSConfigReference ref) {
    int err;
    
    NAMTDevice dev;
    
    dev = (NAMTDevice)malloc(sizeof(NAMTDeviceS));
    if (!dev) return (Device)dev;

    dev->datatime = dsTimeCreate(NULL);
    dev->high_read_pause = dsDurationCreate(NULL);
    dev->min_sync_error = dsDurationCreate(NULL);
    dev->max_sync_error = dsDurationCreate(NULL);
    dev->control_readout_required_time = dsDurationCreate(NULL);

    dev->ilock = dsLockCreateContext(NULL);
    err = dsLockInitContext(dev->ilock);
    
    if ((err)||(!dev->datatime)||(!dev->high_read_pause)||(!dev->control_readout_required_time)||(!dev->min_sync_error)||(!dev->max_sync_error)) {
	if (dev->datatime) dsTimeDestroy(dev->datatime);
	if (dev->high_read_pause) dsDurationDestroy(dev->high_read_pause);
	if (dev->min_sync_error) dsDurationDestroy(dev->min_sync_error);
	if (dev->max_sync_error) dsDurationDestroy(dev->max_sync_error);
	if (dev->control_readout_required_time) dsDurationDestroy(dev->control_readout_required_time);
	if (dev->ilock) dsLockDestroyContext(dev->ilock);
	free(dev);
	return NULL;
    }


    namtDeviceSetDefaults(dev);
    
    dev->device.configure = namtConfigure;
    dev->device.destroy = (DeviceDestroyFunction)namtDeviceDestroy;
    dev->device.reader = (DeviceReaderFunction)namtReader;
    dev->device.processor = (DeviceProcessorFunction)namtProcessor;
    dev->device.reset = namtReset;
    dev->device.debug = (DeviceDebugFunction)namtDebug;
    dev->device.control = (DeviceControlFunction)namtControl;
    dev->device.setconfig = (DeviceSetConfigurationFunction)namtSetConfig;
    dev->device.getdrvinfo = (DeviceGetDriverInfoFunction)namtGetDriverInfo;

    return (Device)dev;
}


static void namtDeviceDestroy(NAMTDevice dev) {
    ds_assert_return(dev);
    
    namtAnalyzerFree(dev, &dev->analyzer);
    namtFree((NAMTDevice)dev);
    dsTimeDestroy(((NAMTDevice)dev)->datatime);
    dsDurationDestroy(((NAMTDevice)dev)->high_read_pause);
    dsDurationDestroy(((NAMTDevice)dev)->min_sync_error);
    dsDurationDestroy(((NAMTDevice)dev)->max_sync_error);
    dsDurationDestroy(((NAMTDevice)dev)->control_readout_required_time);
    dsLockDestroyContext(dev->ilock);
    free(dev);
}



static int namtConfigure(Device dev) {
    xmlNodePtr cfgnode, chipcfg, proccfg;

    ds_assert_param(dev);
    
    cfgnode = dsConfigGetConfigByReference(dev->cfg, dev->cfgref);
    if (!cfgnode) return 0;

    chipcfg = dsXmlFindSubNode(cfgnode, "Configuration");
    proccfg = dsXmlFindSubNode(cfgnode, "Processor");

    ((NAMTDevice)dev)->recs = dsXmlGetNodeLong(proccfg, "WriteInterval", ((NAMTDevice)dev)->recs);
    ((NAMTDevice)dev)->write_control = dsXmlGetNodeLong(proccfg, "ControlWriteInterval", ((NAMTDevice)dev)->write_control);

    dsConfigReturnConfig(dev->cfg, cfgnode);
    
    return 0;
}


