#include <stdio.h>
#include <string.h>
#include <math.h>

#include <dsdebug.h>
#include <dsassert.h>

#include "../../snt/commands.h"

#include "namtreader.h"
#include "namtanalyzer.h"

void namtAnalyzerZero(NAMTDevice dev, NAMTAnalyzer ctx) {
    ctx->curtime = NULL;
    ctx->duration = NULL;

    ctx->devices = 0;
    ctx->detectors = NULL;

    ctx->data = NULL;
    ctx->cdata = NULL;
    ctx->cor = NULL;
    ctx->data_skiped = NULL;
    ctx->cdata_skiped = NULL;
    
    ctx->writer = NULL;
}

static int namtAnalyzerInit(NAMTDevice dev, NAMTAnalyzer ctx, unsigned char plain, unsigned char devices, unsigned short *detectors, unsigned short coincidences, unsigned short spectrums) {
    int i, j;
    int err;
    char stmp[16];
    int used_spectrums;
    
    namtAnalyzerFree(dev, ctx);

    ctx->curtime = dsTimeCreate(NULL);
    ctx->duration = dsDurationCreate(NULL);

    ctx->detectors = (unsigned short*)malloc(devices*sizeof(unsigned short));

    ctx->data = (NAMTDetectorDataS*)malloc(devices*sizeof(NAMTDetectorDataS));
    ctx->cdata = (NAMTDetectorCDataS*)malloc(devices*sizeof(NAMTDetectorCDataS));
    ctx->cor = (DSCorrelationContext*)malloc(devices*sizeof(DSCorrelationContext));
    ctx->data_skiped = (unsigned int*)malloc(devices*sizeof(unsigned int));
    ctx->cdata_skiped = (unsigned int*)malloc(devices*sizeof(unsigned int));
    ctx->writer = (DeviceWriter*)malloc(devices*sizeof(DeviceWriter));
    
    if ((!ctx->data)||(!ctx->cdata)||(!ctx->detectors)||(!ctx->cor)||(!ctx->data_skiped)||(!ctx->cdata_skiped)||(!ctx->curtime)||(!ctx->duration)||(!ctx->writer)) {
	namtAnalyzerFree(dev, ctx);
	return DS_ERR_MEM;
    }

    ctx->devices = 0;

    for (i = 0; i < devices; i++) {
        ctx->detectors[i] = detectors[i];

//	ctx->data[i] = (NAMTDetectorData)malloc(sizeof(NAMTDetectorDataS));
//	ctx->cdata[i] = (NAMTDetectorCData)malloc(sizeof(NAMTDetectorCDataS));

	ctx->cor[i] = dsCorrelationCreateContext(NULL);
	if (ctx->cor[i]) err = dsCorrelationInitContext(ctx->cor[i], plain?ctx->detectors[i]:(ctx->detectors[i] * 2));
	else err = DS_ERR_MEM;

	    /* DS: Check it */
	if (devices > 1) {
	    sprintf(stmp, "-%u", i+1);
	    ctx->writer[i] = DeviceInitWriters((Device)dev, stmp);
	} else {
	    ctx->writer[i] = DeviceInitWriters((Device)dev, NULL);
	}

	if (ctx->writer[i]) {
	    DeviceWritersEnableFixedConfigOptimization((Device)dev, ctx->writer[i]);
	    DeviceWritersEnableFixedLayoutOptimization((Device)dev, ctx->writer[i], NULL);
	    DeviceWritersEnableFixedLayoutOptimization((Device)dev, ctx->writer[i], "-events");
	    DeviceWritersEnableFixedLayoutOptimization((Device)dev, ctx->writer[i], "-control");
	    DeviceWritersEnableFixedLayoutOptimization((Device)dev, ctx->writer[i], "-opinfo");

	    if (plain) used_spectrums = detectors[i] + spectrums;
	    else used_spectrums = 2 * dev->analyzer.detectors[i];

	    for (j = 0; j < used_spectrums; j++) {
		sprintf(stmp, "-spectrum%u",j + 1);
		DeviceWritersEnableFixedLayoutOptimization((Device)dev, ctx->writer[i], stmp);
	    }		
	}

	
	ctx->devices++;

	if ((err)||(!ctx->writer[i])/*||(!ctx->data[i])||(!ctx->cdata[i])*/) {
	    namtAnalyzerFree(dev, ctx);
	    return err;
	}
    }
    
    return 0;
}

void namtAnalyzerFree(NAMTDevice dev, NAMTAnalyzer ctx) {
    unsigned char i;
    
    for (i=0;i<ctx->devices;i++) {
	if ((ctx->writer)&&(ctx->writer[i])) DeviceFreeWriters((Device)dev, ctx->writer[i]);
	if ((ctx->cor)&&(ctx->cor[i])) dsCorrelationDestroyContext(ctx->cor[i]);
//	if ((ctx->data)&&(ctx->data[i])) free(ctx->data[i]);
//	if ((ctx->cdata)&&(ctx->cdata[i])) free(ctx->cdata[i]);
    }
    if (ctx->writer) free(ctx->writer);
    
    if (ctx->data) free(ctx->data);
    if (ctx->cdata) free(ctx->cdata);
    if (ctx->cor) free(ctx->cor);
    if (ctx->data_skiped) free(ctx->data_skiped);
    if (ctx->cdata_skiped) free(ctx->cdata_skiped);

    if (ctx->detectors) free(ctx->detectors);

    if (ctx->curtime) dsTimeDestroy(ctx->curtime);
    if (ctx->duration) dsDurationDestroy(ctx->duration);

    namtAnalyzerZero(dev, ctx);    
}

int namtProcessor(NAMTDevice dev, NAMTReaderBuf buf) {
    int i, j;
    unsigned int max;
    int err;

    DSTime event_time = NULL;
    DSDuration spectrum_duration = NULL;
    NAMTEvent event;
    
    double quality1, quality2;
    
    double coef;

    char *stcBuffer;
    size_t stcBufferSize;
    int res, pos, idlen;

    WriterDataID *cids;
    
    char suffix[16];

    int processed_detectors;

    unsigned int sensors;
    unsigned int spectrums, coincidences;
    
    uint32_t u32tmp;
    double dtmp;

    WriterDataOptions dopts;
    
    NAMT_BUFFER_DATA_DECLARE
    
    ds_assert_param(dev);
    ds_assert_param(buf);
    
    if ((buf->sf)||(!buf->devices)) {
	err = namtAnalyzerInit(dev, &dev->analyzer, buf->plain, buf->devices, buf->detectors, buf->coincidences, buf->spectrums);
	if (err) return err;
	
	dev->analyzer.iter = 0;
    }
    

    if (!dev->analyzer.iter) {
	dsTimeCopy(dev->analyzer.curtime, buf->buf.tm);
	dsDurationZeroDuration(dev->analyzer.duration);
	
	for (i=0;i<buf->devices;i++) {
	    dev->analyzer.data_skiped[i] = 0;
	    dev->analyzer.cdata_skiped[i] = 0;
	    
	    memset(dev->analyzer.data+i, 0, sizeof(NAMTDetectorDataS));
	    memset(dev->analyzer.cdata+i, 0, sizeof(NAMTDetectorCDataS));

	    dsCorrelationClear(dev->analyzer.cor[i]);
	}
    }
    
    dsDurationAddMilliseconds(dev->analyzer.duration, buf->rt_time);

    if (buf->spectrum) {
	spectrum_duration = dsDurationCreate(NULL);
	if (spectrum_duration) 
	    dsDurationImportMilliseconds(spectrum_duration, (buf->current_iteration - buf->spectrum_iteration)*buf->rt_time);
	else
	    dsDebugServer(DS_DEBUG_INFO, "NAMT(%s). Memory allocation failed", dev->device.id);
    }

    if (buf->events) {
	event_time = dsTimeCreate(NULL);
	if (!event_time) {
	    dsDebugServer(DS_DEBUG_CRITICAL, "NAMT(%s). Memory allocation failed", dev->device.id);
	    dsDebugServer(DS_DEBUG_INFO, "NAMT(%s). Events writeout failed", dev->device.id);
	}
    }

    dopts = WriterDataOptionsCreate();
    if (!dopts) dsDebugServer(DS_DEBUG_CRITICAL, "NAMT(%s). Memory allocation failed", dev->device.id);

    for (i = 0, processed_detectors = 0; i < buf->devices; i++, processed_detectors += sensors) {
	if (buf->plain) {
	    sensors = dev->analyzer.detectors[i];
	    coincidences = buf->coincidences;
	    spectrums = sensors + buf->spectrums;
	} else {
	    sensors = 2 * dev->analyzer.detectors[i];
	    coincidences = dev->analyzer.detectors[i] * dev->analyzer.detectors[i] + 8;
	    spectrums = sensors;
	}

	err = DeviceOpenWriters((Device)dev, dev->analyzer.writer[i], "-opinfo", buf->buf.tm);
	if (err) {
		dsDebugServer(DS_DEBUG_WARNING, "NAMT(%s). Operational information writeout is failed. Error %lX (%s)", dev->device.id, err, dsError(err, NULL));
	} else {
	    if (dopts) WriterDataOptionsSetType(dopts, "iteration");
	    DeviceWrite((Device)dev, dev->analyzer.writer[i], DATA_CHANNEL|DATA_CHANNEL_UINT32, dopts, 0, 1, NULL, &buf->current_iteration);

	    if (dopts) WriterDataOptionsSetType(dopts, "hwerrors");
	    DeviceWrite((Device)dev, dev->analyzer.writer[i], DATA_CHANNEL|DATA_CHANNEL_UINT16, dopts, 0, 1, NULL, &buf->errors);

	    if (dopts) WriterDataOptionsSetType(dopts, "sync_error");
	    u32tmp = labs(buf->desync);
	    DeviceWrite((Device)dev, dev->analyzer.writer[i], DATA_CHANNEL|DATA_CHANNEL_UINT32, dopts, 0, 1, NULL, &u32tmp);

	    if (dopts) WriterDataOptionsSetType(dopts, "sync_correction");
	    if ((buf->data)&&(buf->data[0])) dtmp = (int)buf->rt_time - (int)buf->data[0]->duration;
	    else dtmp = 0;
	    DeviceWrite((Device)dev, dev->analyzer.writer[i], DATA_CHANNEL, dopts, 0, 1, NULL, &dtmp);

	    DeviceCloseWriters((Device)dev, dev->analyzer.writer[i]);
	}


	cids = (WriterDataID*)dsMalloca(sensors * sizeof(WriterDataID));
	if (cids) {
	    if (buf->plain) {
		for (j=0;j<sensors;j++) cids[j] = j+1;
	    } else {
		for (j=0;j<dev->analyzer.detectors[i];j++) {
		    cids[j] = LAYER_1|(j+1);
		    cids[j+dev->analyzer.detectors[i]] = LAYER_2|(j+1);
		}
	    }
	}

	DeviceWritersStartBlock((Device)dev, dev->analyzer.writer[i]);
	if ((buf->spectrum)&&(buf->spectrum[i])) {
	    for (j = 0; j < spectrums; j++) {
		sprintf(suffix, "-spectrum%u",j + 1);
	        err = DeviceOpenWriters((Device)dev, dev->analyzer.writer[i], suffix, dev->analyzer.curtime /*buf->buf.tm, Changed to make DB programmers happy*/);
		if (err) {
		    if (j<sensors)
			dsDebugServer(DS_DEBUG_WARNING, "NAMT(%s). Spectrum writeout for sensor %u (device %u) is failed. Error %lX (%s)", dev->device.id, j + 1, i + 1, err, dsError(err, NULL));
		    else 
			dsDebugServer(DS_DEBUG_WARNING, "NAMT(%s). Spectrum %u (device %u) writeout is failed. Error %lX (%s)", dev->device.id, j - sensors + 1, i + 1, err, dsError(err, NULL));
		} else {
		    DeviceWritersSetDuration((Device)dev, dev->analyzer.writer[i], spectrum_duration);
		    DeviceWrite((Device)dev, dev->analyzer.writer[i], DATA_CHANNEL|DATA_CHANNEL_UINT32, NULL, 0, 127, NULL, namtBufferSpectrumData(buf, i, j));
		    DeviceCloseWriters((Device)dev, dev->analyzer.writer[i]);
		}
	    }
	}
	DeviceWritersFinishBlock((Device)dev, dev->analyzer.writer[i]);

	if (event_time) {
	    if ((buf->events[i].mmevent)&&(buf->events[i].mmevents)) {
		sprintf(suffix, "-events");
		
		DeviceWritersStartBlock((Device)dev, dev->analyzer.writer[i]);
		for (j = 0; j < buf->events[i].mmevents; j++) {
		    event = namtBufferMMEvent(buf, i, j);
		    
		    dsTimeCopy(event_time, buf->buf.tm);
		    if (event->iter < buf->current_iteration)
			dsTimeSubMilliseconds(event_time, (buf->current_iteration - event->iter)*buf->rt_time);
		    else 
			dsTimeAddMilliseconds(event_time, (event->iter - buf->current_iteration)*buf->rt_time);
		    dsTimeAddMicroseconds(event_time, event->us);
		    
		    err = DeviceOpenWriters((Device)dev, dev->analyzer.writer[i], suffix, event_time);
		    if (err) {
			dsDebugServer(DS_DEBUG_WARNING, "NAMT(%s). Events writeout for device %u is failed. Error %lX (%s)", dev->device.id, i + 1, err, dsError(err, NULL));
		    } else {
			DeviceWrite((Device)dev, dev->analyzer.writer[i], DATA_CHANNEL|DATA_CHANNEL_UINT8, NULL, 0, sensors, cids, event->d);
			DeviceCloseWriters((Device)dev, dev->analyzer.writer[i]);
		    }
		}
		DeviceWritersFinishBlock((Device)dev, dev->analyzer.writer[i]);
	    }
	}


	if (((dev->device_type == NAMTDEVICE_TYPE_SNT)||(dev->device_type == NAMTDEVICE_TYPE_ANALYZER))&&((buf->sf)||((dev->write_control)&&((dev->analyzer.iter%dev->write_control)==0)))) {
	    err = DeviceOpenWriters((Device)dev, dev->analyzer.writer[i], "-control", dev->analyzer.curtime);
	    if (!err) {
		err = DeviceWrite((Device)dev, dev->analyzer.writer[i], DATA_CHANNEL|DATA_CHANNEL_VOLTAGE, NULL, 0, sensors, cids, dev->hv+processed_detectors);
		if (!err) err = DeviceWrite((Device)dev, dev->analyzer.writer[i], DATA_CHANNEL|DATA_CHANNEL_THRESHOLD, NULL, 0, sensors, cids, dev->thresholds+processed_detectors);
		DeviceCloseWriters((Device)dev, dev->analyzer.writer[i]);
	    }
	    
	    if (err) {
		dsDebugServer(DS_DEBUG_WARNING, "NAMT(%s). Control information writeout for device %u is failed. Error %lX (%s)", dev->device.id, i + 1, err, dsError(err, NULL));
		err = 0;
	    }
	}

    	if ((buf->data)&&(buf->data[i])) {

	    if (buf->data[i]->duration != buf->rt_time) {
		coef = 1. * buf->rt_time /  buf->data[i]->duration;
		dsDebugServer(DS_DEBUG_INFO, "NAMT(%s). Processor. Correcting data: %1.6lf", dev->device.id, coef);
	    } else coef = 1;
    

	    namtBufferDataInitData(buf, i);
/*
	    for (j = 0; j < sensors; j++)
	    	dev->analyzer.data[i].plain.sensor[j] += lrint(coef*namtBufferDataNext(buf));
*/

	    
	    if (buf->plain) {
		for (j = 0; j < sensors; j++)
	    	    dev->analyzer.data[i].plain.sensor[j] += lrint(coef*namtBufferDataNext(buf));
	    } else {
		for (j = 0; j < buf->detectors[i]; j++)
	    	    dev->analyzer.data[i].namt.up[j] += lrint(coef*namtBufferDataNext(buf));
		for (j = 0; j < buf->detectors[i]; j++)
	    	    dev->analyzer.data[i].namt.down[j] += lrint(coef*namtBufferDataNext(buf));
	    }

	    dsCorrelationEvent(dev->analyzer.cor[i], buf->data_size==16?DS_CORRELATION_UINT16:DS_CORRELATION_UINT32, ((char*)buf->data[i]) + sizeof(NAMTDataS));

	} else dev->analyzer.data_skiped[i]++;

    	if ((buf->cdata)&&(buf->cdata[i])) {
	    if (buf->cdata[i]->duration != buf->rt_time) {
		coef = 1. * buf->rt_time /  buf->cdata[i]->duration;
		dsDebugServer(DS_DEBUG_INFO, "NAMT(%s). Processor. Correcting coincidence data: %1.6lf", dev->device.id, coef);
	    } else coef = 1;



	    namtBufferDataInitCData(buf, i);
/*
	    for (j = 0; j < coincidences; j++)
		dev->analyzer.cdata[i].plain.coincidences[j] += lrint(coef*namtBufferDataNext(buf));
*/
	    if (buf->plain) {
		for (j = 0; j < buf->coincidences; j++)
		    dev->analyzer.cdata[i].plain.coincidences[j] += lrint(coef*namtBufferDataNext(buf));
	    } else {
		for (j = 0; j < buf->detectors[i] * buf->detectors[i]; j++)
		    dev->analyzer.cdata[i].namt.coincidences[j] += lrint(coef*namtBufferDataNext(buf));

		dev->analyzer.cdata[i].namt.many_many += lrint(coef*namtBufferDataNext(buf));
		dev->analyzer.cdata[i].namt.many_zero += lrint(coef*namtBufferDataNext(buf));
		dev->analyzer.cdata[i].namt.zero_many += lrint(coef*namtBufferDataNext(buf));
		dev->analyzer.cdata[i].namt.zero_one += lrint(coef*namtBufferDataNext(buf));
		dev->analyzer.cdata[i].namt.one_zero += lrint(coef*namtBufferDataNext(buf));
		dev->analyzer.cdata[i].namt.many_one += lrint(coef*namtBufferDataNext(buf));
		dev->analyzer.cdata[i].namt.one_many += lrint(coef*namtBufferDataNext(buf));
		dev->analyzer.cdata[i].namt.total += lrint(coef*namtBufferDataNext(buf));
	    }
	} else dev->analyzer.cdata_skiped[i]++;

	if (cids) dsFreea(cids);
    }
    
    if (spectrum_duration) dsDurationDestroy(spectrum_duration);
    if (event_time) dsTimeDestroy(event_time);
    
    
    if (++dev->analyzer.iter >= buf->recs) {
	dev->analyzer.iter = 0;

	idlen = strlen(dev->device.id) + 1;	
	stcBufferSize = 1 + 2*sizeof(uint16_t) + idlen + sizeof(time_t) + (2*MAX_NAMT_DETECTORS + 3 + 9) * sizeof(uint32_t) + (2 + 2*MAX_NAMT_DETECTORS) * sizeof(double);
	stcBuffer = dsMalloca(stcBufferSize);
	
	for (i = 0; i < buf->devices; i++) {
	    if (buf->plain) {
		sensors = dev->analyzer.detectors[i];
		coincidences = buf->coincidences;
		spectrums = sensors + buf->spectrums;
	    } else {
		sensors = 2 * dev->analyzer.detectors[i];
		coincidences = dev->analyzer.detectors[i] * dev->analyzer.detectors[i] + 8;
		spectrums = sensors;
	    }
	    
	    if (dev->analyzer.data_skiped[i] < (1 + buf->recs / 2)) {
		max = 0;
		
		if (dev->analyzer.data_skiped[i]) {
		    coef = 1. * buf->recs / (buf->recs - dev->analyzer.data_skiped[i]);
/*
		    for (j = 0; j < sensors; j++)
			dev->analyzer.data[i].plain.sensor[j] = lrint(dev->analyzer.data[i].plain.sensor[j] * coef);
*/
		    if (buf->plain) {
			for (j = 0; j < sensors; j++)
			    dev->analyzer.data[i].plain.sensor[j] = lrint(dev->analyzer.data[i].plain.sensor[j] * coef);
		    } else {
			for (j = 0; j < dev->analyzer.detectors[i]; j++) {
			    dev->analyzer.data[i].namt.up[j] = lrint(dev->analyzer.data[i].namt.up[j] * coef);
			    dev->analyzer.data[i].namt.down[j] = lrint(dev->analyzer.data[i].namt.down[j] * coef);
			}
		    }
		
		    if (max < dev->analyzer.data_skiped[i]) max = dev->analyzer.data_skiped[i];
		}
		
		if (dev->analyzer.cdata_skiped[i]) {
		    coef = 1. * buf->recs / (buf->recs - dev->analyzer.data_skiped[i]);

/*
		    for (j = 0; j < coincidences; j++)
			dev->analyzer.cdata[i].plain.coincidences[j] = lrint(dev->analyzer.cdata[i].plain.coincidences[j] * coef);
*/

		    if (buf->plain) {
			for (j = 0; j < coincidences; j++)
			    dev->analyzer.cdata[i].plain.coincidences[j] = lrint(dev->analyzer.cdata[i].plain.coincidences[j] * coef);
		    } else {
			for (j = 0; j < dev->analyzer.detectors[i] * dev->analyzer.detectors[i]; j++) {
			    dev->analyzer.cdata[i].namt.coincidences[j] = lrint(dev->analyzer.cdata[i].namt.coincidences[j] * coef);
			}
		    
			dev->analyzer.cdata[i].namt.many_many = lrint(dev->analyzer.cdata[i].namt.many_many * coef);
			dev->analyzer.cdata[i].namt.many_zero = lrint(dev->analyzer.cdata[i].namt.many_zero * coef);
			dev->analyzer.cdata[i].namt.zero_many = lrint(dev->analyzer.cdata[i].namt.zero_many * coef);
			dev->analyzer.cdata[i].namt.one_many = lrint(dev->analyzer.cdata[i].namt.one_many * coef);
			dev->analyzer.cdata[i].namt.many_one = lrint(dev->analyzer.cdata[i].namt.many_one * coef);
			dev->analyzer.cdata[i].namt.zero_one = lrint(dev->analyzer.cdata[i].namt.zero_one * coef);
	    		dev->analyzer.cdata[i].namt.one_zero = lrint(dev->analyzer.cdata[i].namt.one_zero * coef);
			dev->analyzer.cdata[i].namt.total = lrint(dev->analyzer.cdata[i].namt.total * coef);
		    }
		    
		    if (dev->analyzer.cdata_skiped[i] > max) max = dev->analyzer.cdata_skiped[i];
		}
		
		if (max) dsDebugServer(DS_DEBUG_WARNING, "NAMT(%s). Missing (%lu,%lu) of %lu events on device %u", dev->device.id, dev->analyzer.data_skiped[i],dev->analyzer.cdata_skiped[i],buf->recs, i + 1);
	    }
	    else {
		dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Skiping Device %u Record. %lu of %lu events are missing", dev->device.id , i, dev->analyzer.data_skiped[i], buf->recs);
		continue;
	    }

	    dsCorrelationFinalyze(dev->analyzer.cor[i]);


	    err = DeviceOpenWriters((Device)dev, dev->analyzer.writer[i], NULL, dev->analyzer.curtime);
	    
	    if (!err) {
		cids = (WriterDataID*)dsMalloca(sensors * sizeof(WriterDataID));
		if (cids) {
		    if (buf->plain) {
			for (j=0;j<sensors;j++) cids[j] = j + 1;
		    } else {
			for (j=0;j<dev->analyzer.detectors[i];j++) {
			    cids[j] = LAYER_1|(j+1);
			    cids[j+dev->analyzer.detectors[i]] = LAYER_2|(j+1);
			}
		    }
		}
		
		quality1 = 1. * (buf->recs - dev->analyzer.data_skiped[i]) / buf->recs;
		quality2 = 1. * (buf->recs - dev->analyzer.cdata_skiped[i]) / buf->recs;
		DeviceWritersSetQuality((Device)dev, dev->analyzer.writer[i], (quality1<quality2)?quality1*100:quality2*100);
		DeviceWritersSetDuration((Device)dev, dev->analyzer.writer[i], dev->analyzer.duration);
		if (buf->plain) {
		    err = DeviceWrite((Device)dev, dev->analyzer.writer[i], DATA_DETECTOR_COUNTER, NULL, 0, sensors, NULL, dev->analyzer.data[i].plain.sensor);
		    if (!err) err = DeviceWrite((Device)dev, dev->analyzer.writer[i], DATA_COINCIDENCE|DATA_COINCIDENCE_COINCIDENCE, NULL, 0, coincidences, NULL, dev->analyzer.cdata[i].plain.coincidences);
		} else {
		    err = DeviceWrite((Device)dev, dev->analyzer.writer[i], DATA_DETECTOR_COUNTER, NULL, 1, dev->analyzer.detectors[i], NULL, dev->analyzer.data[i].namt.up);
	    	    if (!err) err = DeviceWrite((Device)dev, dev->analyzer.writer[i], DATA_DETECTOR_COUNTER, NULL, 2, dev->analyzer.detectors[i], NULL, dev->analyzer.data[i].namt.down);
		    if (!err) err = DeviceWrite((Device)dev, dev->analyzer.writer[i], DATA_COINCIDENCE|DATA_COINCIDENCE_PAIR, NULL, 0, dev->analyzer.detectors[i], NULL, dev->analyzer.cdata[i].namt.coincidences);
		    if (!err) err = DeviceWrite((Device)dev, dev->analyzer.writer[i], DATA_COINCIDENCE|DATA_COINCIDENCE_MANY_MANY, NULL, 0, dev->analyzer.detectors[i], NULL, &dev->analyzer.cdata[i].namt.many_many);
		    if (!err) err = DeviceWrite((Device)dev, dev->analyzer.writer[i], DATA_COINCIDENCE|DATA_COINCIDENCE_MANY_ZERO, NULL, 0, dev->analyzer.detectors[i], NULL, &dev->analyzer.cdata[i].namt.many_zero);
		    if (!err) err = DeviceWrite((Device)dev, dev->analyzer.writer[i], DATA_COINCIDENCE|DATA_COINCIDENCE_ZERO_MANY, NULL, 0, dev->analyzer.detectors[i], NULL, &dev->analyzer.cdata[i].namt.zero_many);
		    if (!err) err = DeviceWrite((Device)dev, dev->analyzer.writer[i], DATA_COINCIDENCE|DATA_COINCIDENCE_ZERO_ONE, NULL, 0, dev->analyzer.detectors[i], NULL, &dev->analyzer.cdata[i].namt.zero_one);
		    if (!err) err = DeviceWrite((Device)dev, dev->analyzer.writer[i], DATA_COINCIDENCE|DATA_COINCIDENCE_ONE_ZERO, NULL, 0, dev->analyzer.detectors[i], NULL, &dev->analyzer.cdata[i].namt.one_zero);
		    if (!err) err = DeviceWrite((Device)dev, dev->analyzer.writer[i], DATA_COINCIDENCE|DATA_COINCIDENCE_MANY_ONE, NULL, 0, dev->analyzer.detectors[i], NULL, &dev->analyzer.cdata[i].namt.many_one);
		    if (!err) err = DeviceWrite((Device)dev, dev->analyzer.writer[i], DATA_COINCIDENCE|DATA_COINCIDENCE_ONE_MANY, NULL, 0, dev->analyzer.detectors[i], NULL, &dev->analyzer.cdata[i].namt.one_many);
		    if (!err) err = DeviceWrite((Device)dev, dev->analyzer.writer[i], DATA_COINCIDENCE|DATA_COINCIDENCE_TOTAL, NULL, 0, dev->analyzer.detectors[i], NULL, &dev->analyzer.cdata[i].namt.total);
		}
		if (!err) err = DeviceWrite((Device)dev, dev->analyzer.writer[i], DATA_DISPERSION, NULL, 0, sensors, cids, dsCorrelationGetDispersionArray(dev->analyzer.cor[i]));
		if (!err) err = DeviceWrite((Device)dev, dev->analyzer.writer[i], DATA_CORRELATION, NULL, 0, sensors, cids, dsCorrelationGetCorrelationArray(dev->analyzer.cor[i]));
		DeviceCloseWriters((Device)dev, dev->analyzer.writer[i]);
	    } else cids = NULL;
	    
	    if (cids) dsFreea(cids);

	    if (stcBuffer) {
	        *(char*)stcBuffer = 0xE1 + i;
		pos = 1;		*(uint16_t*)(stcBuffer+pos) = dev->device.type | (buf->device_type<<8);
		pos += sizeof(uint16_t);*(uint16_t*)(stcBuffer+pos) = dev->device.version;
		pos += sizeof(uint16_t);memcpy(stcBuffer+pos, dev->device.id, idlen);
		pos += idlen;		*(time_t*)(stcBuffer+pos) = dsTimeGetUnixTime(dev->analyzer.curtime);
	        pos+=sizeof(time_t);	*(uint32_t*)(stcBuffer+pos)=buf->rt_time * buf->recs;
		pos+=sizeof(uint32_t);	*(uint32_t*)(stcBuffer+pos)=buf->rt_time;
	        pos+=sizeof(uint32_t);	*(uint32_t*)(stcBuffer+pos)=sensors;
		pos+=sizeof(uint32_t);

/*
	        for (j=0;j<sensors;j++,pos+=sizeof(uint32_t))
		    *(uint32_t*)(stcBuffer+pos) = dev->analyzer.data[i].plain.sensor[j];
*/
		if (buf->plain) {
	    	    for (j=0;j<sensors;j++,pos+=sizeof(uint32_t))
			*(uint32_t*)(stcBuffer+pos) = dev->analyzer.data[i].plain.sensor[j];
		} else {
	    	    for (j=0;j<buf->detectors[i];j++,pos+=sizeof(uint32_t))
			*(uint32_t*)(stcBuffer+pos) = dev->analyzer.data[i].namt.up[j];
		    for (j=0;j<buf->detectors[i];j++,pos+=sizeof(uint32_t))
			*(uint32_t*)(stcBuffer+pos) = dev->analyzer.data[i].namt.down[j];
		}

		for (j=0;j<sensors;j++,pos+=sizeof(double))
		    *(double*)(stcBuffer+pos)=dsCorrelationGetDispersion(dev->analyzer.cor[i], j);
		    
		if (!buf->plain) {
		    for (j=0, res=0;j<buf->detectors[i] * buf->detectors[i];j++) 
			res += dev->analyzer.cdata[i].namt.coincidences[j];

		    *(uint32_t*)(stcBuffer+pos) = dev->analyzer.cdata[i].namt.total; pos+=sizeof(uint32_t);
		    *(uint32_t*)(stcBuffer+pos) = res; pos+=sizeof(uint32_t);
		    *(uint32_t*)(stcBuffer+pos) = dev->analyzer.cdata[i].namt.many_many; pos+=sizeof(uint32_t);
		    *(uint32_t*)(stcBuffer+pos) = dev->analyzer.cdata[i].namt.many_zero; pos+=sizeof(uint32_t);
		    *(uint32_t*)(stcBuffer+pos) = dev->analyzer.cdata[i].namt.zero_many; pos+=sizeof(uint32_t);
		    *(uint32_t*)(stcBuffer+pos) = dev->analyzer.cdata[i].namt.zero_one; pos+=sizeof(uint32_t);
		    *(uint32_t*)(stcBuffer+pos) = dev->analyzer.cdata[i].namt.one_zero; pos+=sizeof(uint32_t);
		    *(uint32_t*)(stcBuffer+pos) = dev->analyzer.cdata[i].namt.many_one; pos+=sizeof(uint32_t);
		    *(uint32_t*)(stcBuffer+pos) = dev->analyzer.cdata[i].namt.one_many; pos+=sizeof(uint32_t);
		}
    	
		*(double*)(stcBuffer+pos) = 1. * (buf->recs - dev->analyzer.data_skiped[i]) / buf->recs; pos+=sizeof(double);
		if (!buf->plain) {
		    *(double*)(stcBuffer+pos) = 1. * (buf->recs - dev->analyzer.cdata_skiped[i]) / buf->recs; pos+=sizeof(double);
		}
		
		dsDatagramServerBroadcast(dev->device.dsc, NULL, dev->device.id, stcBuffer, pos, 0, NULL);
	    }
	}
	
	if (stcBuffer) dsFreea(stcBuffer);
    }
	    
    if (dopts) WriterDataOptionsDestroy(dopts);
    
    return 0;
}
