#define _DS_INCLUDE_XML_HEADERS
#define _NAMT_LAYOUT_C
#include <stdio.h>
#include <dsdebug.h>
#include <dsassert.h>

#include "../../snt/commands.h"

#include "defaults.h"
#include "lowlevel.h"
#include "namterror.h"
#include "namtreader.h"
#include "namtlayout.h"


static LayoutRuleSetType namtLayoutGetRuleSetType(const char *type) {
    int i;
    
    ds_assert_default(type, LAYOUT_RULESET_TYPE_UNKNOWN);
    
    for (i=0; layout_ruleset_types[i].id; i++) {
	if (!strcmp(type, layout_ruleset_types[i].id)) return layout_ruleset_types[i].type;
    }
    
    return LAYOUT_RULESET_TYPE_UNKNOWN;
}

static LayoutRuleType namtLayoutGetRuleType(const char *type) {
    int i;
    
    ds_assert_default(type, LAYOUT_RULE_TYPE_UNKNOWN);
    
    for (i=0; layout_rule_types[i].id; i++) {
	if (!strcmp(type, layout_rule_types[i].id)) return layout_rule_types[i].type;
    }
    
    return LAYOUT_RULE_TYPE_UNKNOWN;
}

static LayoutRuleTest namtLayoutGetRuleTest(const char *type) {
    int i;
    
    ds_assert_default(type, LAYOUT_RULE_TEST_UNKNOWN);
    
    for (i=0; layout_rule_tests[i].id; i++) {
	if (!strcmp(type, layout_rule_tests[i].id)) return layout_rule_tests[i].type;
    }
    
    return LAYOUT_RULE_TEST_UNKNOWN;
}


static int namtLayoutSendRuleSet(NAMTDevice dev, Connection con, NAMTRuleSet ruleset, int rules, NAMTRuleS *rule) {
    int i;
    int res, err = 0;
    
    res = SendData(con, COMMAND_REQUEST_LAYOUT_REGISTER_RULESET, sizeof(NAMTRuleSetS), (char*)ruleset);
    if (res != COMMAND_RESPONSE_LAYOUT_REGISTER_RULESET) {
	return namtError(dev, con, COMMAND_REQUEST_LAYOUT_REGISTER_RULESET, res);
    }

    for (i = 0; i < rules; i++) {
	res = SendData(con, COMMAND_REQUEST_LAYOUT_REGISTER_RULE, sizeof(NAMTRuleS), (char*)(rule+i));
	if (res != COMMAND_RESPONSE_LAYOUT_REGISTER_RULE) {
	    err = namtError(dev, con, COMMAND_REQUEST_LAYOUT_REGISTER_RULE, res);
	    break;
	}
    }

    return err;
}

static int namtLayoutSendRuleSpec(NAMTDevice dev, Connection con, NAMTRuleSpec spec) {
    int res;
    
    res = SendData(con, COMMAND_REQUEST_LAYOUT_REGISTER_RULESPEC, sizeof(NAMTRuleSpecS), (char*)spec);
    if (res != COMMAND_RESPONSE_LAYOUT_REGISTER_RULESPEC) {
	return namtError(dev, con, COMMAND_REQUEST_LAYOUT_REGISTER_RULESPEC, res);
    }
    return 0;
}

static int namtLayoutGetRuleSetByName(int rulesets, const char **rsname, const char *name, size_t namelen) {
    int i;
    
    for (i = 0; i < rulesets; i++) {
	if (rsname[i]) {
	    if (namelen) {
		if (!strncmp(rsname[i], name, namelen)) return i + 1;
	    } else {
		if (!strcmp(rsname[i], name)) return i + 1;
	    }
	}
    }
    
    return 0;
}

int namtLayoutProcess(NAMTDevice dev, Connection con, xmlNodePtr layout) {
    int err, res;
    int used_detectors;
    xmlNodePtr *noderulesets, noderuleset;
    xmlNodePtr *nodespectrums, nodespectrum;
    xmlNodePtr *noderules, noderule;
    const char *name, *success;
    const char *pos;

    int i, j, id;    
    int spectrumsensor;
    LayoutRuleSetType type;
    unsigned char sensor[dev->layout_sensors];
    unsigned char store_events, store_counts;
    int rules;
    NAMTRuleS rule[dev->layout_rules];
    int spectrums = 0;
    NAMTRuleSpecS spectrum;
    int coincidences = 0;
    NAMTRuleSetS ruleset;
    
    const char *rsname[dev->layout_channels];
    
    int stmplen = 128;
    char stmp[stmplen+1];
    
    int rulesetidx = 0;

    LayoutRuleType ruletype;
    LayoutRuleTest ruletest;
    int16_t ruletarget;
    int16_t rulethreshold;

    res = SendSimple(con, COMMAND_REQUEST_LAYOUT_RESET);
    if (res != COMMAND_RESPONSE_LAYOUT_RESET) {
	return namtError(dev, con, COMMAND_REQUEST_LAYOUT_RESET, res);
    }

    used_detectors = dsConfigGetNodeInt(dev->device.cfg, layout, "Sensors", dev->layout_sensors);
    if (used_detectors > dev->layout_sensors) {
	    dsDebugServer(DS_DEBUG_CRITICAL, "NAMT(%s). Layout. Device equiped with %hu inputs, layout requires %hu", dev->device.id, dev->layout_sensors, used_detectors);
	    err = DS_ERR_CONFIG;
    } 

    if (used_detectors != dev->layout_sensors) {
	res = SendAttribute(con, COMMAND_REQUEST_LAYOUT_USED_DETECTORS, used_detectors);
	if (res != COMMAND_RESPONSE_LAYOUT_USED_DETECTORS) {
	    return namtError(dev, con, COMMAND_REQUEST_LAYOUT_USED_DETECTORS, res);
	}
    }
    
    noderulesets = dsConfigFindNodes(dev->device.cfg, layout, "Coincidences/RuleSet");
    if (noderulesets) {
	for (i = 0; noderulesets[i]; i++) {
	    err = 0;
	    noderuleset = noderulesets[i];
	    
	    name = dsConfigGetPropString(dev->device.cfg, noderuleset, "name", NULL);
	    if (!name) {
		dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Layout. Invalid RuleSet: name attribute is missing", dev->device.id);
		continue;
	    }
	    
	    pos = dsConfigGetPropString(dev->device.cfg, noderuleset, "type", NULL);
	    if (pos) type = namtLayoutGetRuleSetType(pos);
	    else {
		dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Layout. Invalid RuleSet \"%s\": type attribute is missing", dev->device.id, name);
		continue;
	    }
	    
	    if (type == LAYOUT_RULESET_TYPE_UNKNOWN) {
		dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Layout. Invalid RuleSet \"%s\": Unsupported type \"%s\"", dev->device.id, name, pos);
		continue;
	    }

	    if ((rulesetidx == dev->layout_channels)||((type ==LAYOUT_RULESET_TYPE_ONE)&&((rulesetidx + 3) > dev->layout_channels))) {
		dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Layout. Number of rulesets exceeds maximal allowed value \"%i\"", dev->device.id, dev->layout_channels);
		break;
	    }

	    store_counts = !dsConfigGetPropBoolean(dev->device.cfg, noderuleset, "dummy", 0);
	    if (store_counts) {
	        store_events = dsConfigGetPropBoolean(dev->device.cfg, noderuleset, "events", 0);
	    } else {
		store_events = 0;
	    }

	    switch(type) {
		case LAYOUT_RULESET_TYPE_CUSTOM:
		    noderules = dsConfigFindNodes(dev->device.cfg, noderuleset, "Rule");
		    if (noderules) {
			for (rules = 0, j = 0; noderules[j]; j++) {
			    noderule = noderules[j];

			    if (j == dev->layout_rules) {
				dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Layout. Invalid RuleSet \"%s\": number of rules exceeds maximal allowed value \"%i\"", dev->device.id, name, dev->layout_rules);
				err = DS_ERR_TOOLONG;
				break;
			    }
			    
			    pos = dsConfigGetPropString(dev->device.cfg, noderule, "type", NULL);
			    if (pos) {
				ruletype = namtLayoutGetRuleType(pos);
				if (ruletype == LAYOUT_RULE_TYPE_UNKNOWN) {
				    dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Layout. Invalid Rule in RuleSet \"%s\": Unsupported type \"%s\"", dev->device.id, name, pos);
				    err = DS_ERR_INVALID_OPTIONS;
				    break;
				}
			    } else {
				dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Layout. Invalid Rule in RuleSet \"%s\": type attribute is missing", dev->device.id, name);
				err = DS_ERR_INVALID_OPTIONS;
				break;
			    }
			    
			    switch (ruletype) {
				case LAYOUT_RULE_TYPE_SENSOR:
				    ruletarget = dsConfigGetPropInt(dev->device.cfg, noderule, "target", 0);
				    if ((ruletarget < 1)||(ruletarget > used_detectors)) {
					dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Layout. Invalid Rule in RuleSet \"%s\": sensor \"%i\" is out of range", dev->device.id, name, ruletarget);
					err = DS_ERR_INVALID_OPTIONS;
					break;
				    }
				break;
				case LAYOUT_RULE_TYPE_RULESET:
				    pos = dsConfigGetPropString(dev->device.cfg, noderule, "target", NULL);
				    id = namtLayoutGetRuleSetByName(rulesetidx, rsname, pos, 0);
				    if (id) ruletarget = -id;
				    else {
					dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Layout. Invalid Rule in RuleSet \"%s\": Non existent  RuleSet \"%s\" is specified", dev->device.id, name, pos);
					err = DS_ERR_INVALID_OPTIONS;
				    }
				break;
				default:
				    ;
			    }
			    if (err) break;
			    
			    
			    pos = dsConfigGetPropString(dev->device.cfg, noderule, "test", NULL);
			    if (pos) {
				ruletest = namtLayoutGetRuleTest(pos);
				if (ruletest == LAYOUT_RULE_TEST_UNKNOWN) {
				    dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Layout. Invalid Rule in RuleSet \"%s\": Unsupported test \"%s\"", dev->device.id, name, pos);
				    err = DS_ERR_INVALID_OPTIONS;
				    break;
				}
			    } else {
				dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Layout. Invalid Rule in RuleSet \"%s\": test attribute is missing", dev->device.id, name);
				err = DS_ERR_INVALID_OPTIONS;
				break;
			    }
			    
			    switch (ruletest) {
				case LAYOUT_RULE_TEST_TRUE:
				    rulethreshold = 0;
				break;
				case LAYOUT_RULE_TEST_FALSE:
				    rulethreshold = -127;
				break;
				case LAYOUT_RULE_TEST_ABOVE:
				    rulethreshold = dsConfigGetPropInt(dev->device.cfg, noderule, "threshold", 0);
				    if ((rulethreshold < 1)||(rulethreshold > 127)) {
					err = DS_ERR_INVALID_OPTIONS;
					break;
				    }
				    rulethreshold++;
				break;
				case LAYOUT_RULE_TEST_BELOW:
				    rulethreshold = dsConfigGetPropInt(dev->device.cfg, noderule, "threshold", 0);
				    if ((rulethreshold < 1)||(rulethreshold > 126)) {
					err = DS_ERR_INVALID_OPTIONS;
					break;
				    }
				    rulethreshold = -rulethreshold;
				break;
				default:
				    ;
			    }

			    if (err) {
				dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Layout. Invalid Rule in RuleSet \"%s\": threshold \"%i\" is out of range", dev->device.id, name, rulethreshold);
				break;
			    }
			    
			    rule[rules].target = ruletarget;
			    rule[rules].threshold = rulethreshold;
			    
			    rules++;
			}	
			dsFree(NULL, noderules);
		    } else {
			dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Layout. Invalid RuleSet \"%s\": No rules specified", dev->device.id, name);
			err = DS_ERR_EMPTY;
		    }
		break;
		case LAYOUT_RULESET_TYPE_NONE:
		case LAYOUT_RULESET_TYPE_ONE:
	        case LAYOUT_RULESET_TYPE_SOME:
		case LAYOUT_RULESET_TYPE_ALL:
		case LAYOUT_RULESET_TYPE_SET:
		    for (j=0;j<used_detectors;j++)
			sensor[j] = 0;

		    success = dsConfigGetPropString(dev->device.cfg, noderuleset, "success", NULL);
		    for (pos = success; pos; pos = strchr(pos, ' ')) {
			while (*pos==' ') pos++;
			if (*pos) {
			    id = atoi(pos);
			    if ((id<1)||(id>used_detectors)) {
				dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Layout. Invalid RuleSet \"%s\": refers non existent sensor \"%i\"", dev->device.id, name, id);
				err = DS_ERR_INVALID_OPTIONS;
			    } else sensor[id - 1] = 1;
			}
		    }

		    if (type == LAYOUT_RULESET_TYPE_SET) {
			if (used_detectors == dev->layout_rules) {
			    dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Layout. Invalid RuleSet \"%s\": number of rules exceeds maximal allowed value \"%i\"", dev->device.id, name, dev->layout_rules);
			    err = DS_ERR_TOOLONG;
			} else {
		    	    for (j=0;j<used_detectors;j++) {
				rule[j].target = j + 1;
				if (sensor[j]) rule[j].threshold = 0;
				else rule[j].threshold = -127;
			    }
			    rules = j;
		        }
		    } else {
		      for (rules=0,j=0;j<used_detectors;j++) {
			if (sensor[j]) {
			    if (rules == dev->layout_rules) {
				dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Layout. Invalid RuleSet \"%s\": number of rules exceeds maximal allowed value \"%i\"", dev->device.id, name, dev->layout_rules);
				err = DS_ERR_TOOLONG;
				break;
			    }
			    rule[rules].target = j + 1;
			    rule[rules].threshold = 0;
			    rules++;
			}
		      }
		    }		
		break;
	        case LAYOUT_RULESET_TYPE_OR:
		case LAYOUT_RULESET_TYPE_AND:
		    success = dsConfigGetPropString(dev->device.cfg, noderuleset, "success", NULL);
		    for (rules = 0, pos = success; pos; pos = strchr(pos, ' ')) {
			while (*pos==' ') pos++;

			if (*pos) {
			    if (rules == dev->layout_rules) {
				dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Layout. Invalid RuleSet \"%s\": number of rules exceeds maximal allowed value \"%i\"", dev->device.id, name, dev->layout_rules);
				err = DS_ERR_TOOLONG;
				break;
			    }

			    for (j=0;(pos[j])&&(pos[j]!=' ');j++);

			    id = namtLayoutGetRuleSetByName(rulesetidx, rsname, pos, j);
			    if (id) {
				rule[rules].target = -id;
				rule[rules].threshold = 0;
				rules++;
			    } else {
				strncpy(stmp, pos, (j<stmplen)?j:stmplen);
				dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Layout. Invalid RuleSet \"%s\": refers non existent RuleSet \"%s\"", dev->device.id, name, stmp);
				err = DS_ERR_INVALID_OPTIONS;
			    }
			}
		    }
		break;
		default:
		    ;
	    }
	    
	    if (err) continue;
	    
	    if (!rules) {
		dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Layout. Invalid RuleSet \"%s\": No rules specified", dev->device.id, name);
		continue;
	    }
    
	    ruleset.flags = 0;
	    if (type != LAYOUT_RULESET_TYPE_ONE) {
		if (store_counts) ruleset.flags|=LAYOUT_RULESET_FLAGS_COUNT;
		if (store_events) ruleset.flags|=LAYOUT_RULESET_FLAGS_EVENTS;
	    }
	    
	    switch(type) {
		case LAYOUT_RULESET_TYPE_CUSTOM:
		    pos = dsConfigGetPropString(dev->device.cfg, noderuleset, "success", NULL);
		    if (pos) {
			ruletest = namtLayoutGetRuleTest(pos);
			switch (ruletest) {
			    case LAYOUT_RULE_TEST_ABOVE:
			        rulethreshold = dsConfigGetPropInt(dev->device.cfg, noderuleset, "threshold", 0);
				if ((rulethreshold < 0)||(rulethreshold > 127)) {
				    dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Layout. Invalid Rule in RuleSet \"%s\": threshold \"%i\" is out of range", dev->device.id, name, rulethreshold);
				    err = DS_ERR_INVALID_OPTIONS;
				} else ruleset.success = rulethreshold + 1;
			    break;
			        rulethreshold = dsConfigGetPropInt(dev->device.cfg, noderuleset, "threshold", 0);
				if ((rulethreshold < 1)||(rulethreshold > 126)) {
				    dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Layout. Invalid Rule in RuleSet \"%s\": threshold \"%i\" is out of range", dev->device.id, name, rulethreshold);
				    err = DS_ERR_INVALID_OPTIONS;
				} else ruleset.success = -rulethreshold;
			    break;
			    default:
				dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Layout. Invalid Rule in RuleSet \"%s\": Unsupported test \"%s\"", dev->device.id, name, pos);
				err = DS_ERR_INVALID_OPTIONS;
			}
		    } else {
			dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Layout. Invalid Rule in RuleSet \"%s\": success attribute is missing", dev->device.id, name);
			err = DS_ERR_INVALID_OPTIONS;
		    }
		break;
		case LAYOUT_RULESET_TYPE_NONE:
		    ruleset.success = -1;
		break;
		case LAYOUT_RULESET_TYPE_ONE:
	        case LAYOUT_RULESET_TYPE_SOME:
	        case LAYOUT_RULESET_TYPE_OR:
		    ruleset.success = 1;
		break;
		case LAYOUT_RULESET_TYPE_ALL:
		case LAYOUT_RULESET_TYPE_SET:
		case LAYOUT_RULESET_TYPE_AND:
		    ruleset.success = rules;
		break;
	    }
	    
	    if (err) continue;
	    
	    err = namtLayoutSendRuleSet(dev, con, &ruleset, rules, rule);
	    if (err) break;
	    else rulesetidx++;
	    
	    if (type == LAYOUT_RULESET_TYPE_ONE) {
		rsname[rulesetidx-1] = NULL;

		ruleset.success = -2;

		err = namtLayoutSendRuleSet(dev, con, &ruleset, rules, rule);
		if (err) break;
		else rulesetidx++;

		rsname[rulesetidx-1] = NULL;
		
		ruleset.flags = 0;
		if (store_counts) ruleset.flags|=LAYOUT_RULESET_FLAGS_COUNT;
	        if (store_events) ruleset.flags|=LAYOUT_RULESET_FLAGS_EVENTS;
		ruleset.success = 2;
		rule[0].target = -rulesetidx + 1;
		rule[0].threshold = 0;
		rule[1].target = -rulesetidx;
		rule[1].threshold = 0;

		err = namtLayoutSendRuleSet(dev, con, &ruleset, 2, rule);
		if (err) break;
		else rulesetidx++;
	    }

	    rsname[rulesetidx - 1] = name;
	    if (store_counts) coincidences++;
	}
	    
	dsFree(NULL, noderulesets);
    }

    if (!err) {
	nodespectrums = dsConfigFindNodes(dev->device.cfg, layout, "Spectrums/Spectrum");
	if (nodespectrums) {
	    for (j = 0; nodespectrums[j]; j++) {
		err = 0;
		nodespectrum = nodespectrums[j];

		if (spectrums == dev->layout_spectrums) {
		    dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Layout. Invalid Spectrum: number of spectrums exceeds maximal allowed value \"%i\"", dev->device.id, dev->layout_spectrums);
		    break;
		}

		name = dsConfigGetPropString(dev->device.cfg, nodespectrum, "ruleset", NULL);
		if (name) {
		    id = namtLayoutGetRuleSetByName(rulesetidx, rsname, name, 0);
		    if (!id) {
			dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Layout. Invalid Spectrum: refers non existent RuleSet \"%s\"", dev->device.id, name);
			err = DS_ERR_INVALID_OPTIONS;
		    }
		} else {
		    dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Layout. Invalid Spectrum: ruleset attribute is missing", dev->device.id);
		    err = DS_ERR_INVALID_OPTIONS;
		}

		if (err) continue;
			
		spectrumsensor = dsConfigGetPropInt(dev->device.cfg, nodespectrum, "sensor", 0);
		if ((spectrumsensor > 0)&&(spectrumsensor <= used_detectors))  {
		    spectrum.ruleset = id;
		    spectrum.sensor = spectrumsensor;
			    
		    err = namtLayoutSendRuleSpec(dev, con, &spectrum);
		    if (err) break;
		    else spectrums++;
		} else {
		    dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Layout. Invalid Spectrum: sensor \"%i\" is out of allowed range", dev->device.id, sensor);
		}
	    }
	    dsFree(NULL, nodespectrums);
	}
    }


    if (err) {
	SendSimple(con, COMMAND_REQUEST_LAYOUT_RESET);
    } else {
        err = namtAllocate(dev, 1);
	if (err) {
	    dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Error: %lx (%s)", dev->device.id, err, dsError(err, NULL));
	    return err;
	}
		
	err = namtAllocateNextDevice(dev, used_detectors);
	if (err) {
	    dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Error allocating sub-device: %lx (%s)", dev->device.id, err, dsError(err, NULL));
	    return err;
	}
    }
    
    if (!err) {
	dev->spectrums = spectrums;
	dev->coincidences = coincidences;
    }
    
    return err;
}
