#define _NAMT_READER_C
#define _DS_INCLUDE_XML_HEADERS

#include <stdio.h>

#include <dsdebug.h>
#include <dsassert.h>
#include <dserror.h>

#include <dsdispersion.h>

#include "../../snt/commands.h"

#include "error.h"
#include "defaults.h"
#include "final.h"

#include "lowlevel.h"

#include "namterror.h"
#include "namtreader.h"

//#define CALIBRATION_TIME 10

int namtAllocate(NAMTDevice dev, unsigned char devices) {
    int i;
    
    namtFree(dev);

    dev->detectors = (unsigned short*)malloc(devices*sizeof(unsigned short));

    if (!dev->detectors) {
	namtFree(dev);
	return DS_ERR_MEM;
    }

    dev->devices = 0;
    dev->max_devices = devices;

    return 0;    
}

int namtAllocateNextDevice(NAMTDevice dev, unsigned char detectors) {
    unsigned char i = dev->devices;
    
    ds_assert_param(dev);
    ds_assert_param(dev->devices < dev->max_devices);
    
    dev->detectors[i] = detectors;
    dev->devices++;

    return 0;
}

void namtZero(NAMTDevice dev) {
    dev->devices = 0;
    dev->detectors = NULL;

    dev->current_iteration = 0;
    
    memset(&dev->spectrum, 0, sizeof(SpectrumPropertiesS));
}

void namtFree(NAMTDevice dev) {
    if (dev->detectors) free(dev->detectors);
    namtZero(dev);    
}

int namtEventsInit(NAMTDevice dev, NAMTEvents event, unsigned char subdevice) {
    ds_assert_param(event);

    event->subdevice = subdevice;
    
    event->mmevents = 0;
    event->mmevent = NULL;
    return 0;
}

void namtEventsFree(NAMTEvents event) {
    ds_assert_return(event);
    
    if (event->mmevent) {
	free(event->mmevent);
	event->mmevent = NULL;
    }
}

int namtEventsMM(NAMTDevice dev, NAMTEvents event, size_t data_size, void *data) {
    ds_assert_param(event);

    if (data_size) {    
	if (data_size%namtEventSize(dev, event->subdevice)) return DS_ERR_INVALID_RESPONSE;
        event->mmevents = data_size / namtEventSize(dev, event->subdevice);
	event->mmevent = data;
    } else if (data) free(data);
    
    return 0;
}

static int CalculateStartTime(NAMTDevice dev, unsigned int rt_time, DSTime sync_time, DSTime start_time) {
    DSDuration rts_time;
    
    ds_assert_param(dev);
    ds_assert_param(rt_time);
    ds_assert_param(sync_time);
    ds_assert_param(start_time);
    
    dsTimeCurrentTime(start_time);
    dsTimeAddMilliseconds(start_time, 2*rt_time);
    
    rts_time = dsDurationCreate(NULL);
    if (!rts_time) return DS_ERR_MEM;

    dsDurationImportMilliseconds(rts_time, DaemonFlag?rt_time * dev->recs:rt_time);
    dsDurationCeilSeconds(rts_time, 1);

    dsTimeCeil(start_time, rts_time);
    
    dsTimeCopy(sync_time, start_time);
    dsTimeSubMilliseconds(sync_time, rt_time);
    
    dsDurationDestroy(rts_time);
    return 0;
}

static int ReadControl(NAMTDevice dev, Connection con) {
    int i,j,k;
    int res;
    uint16_t utmp;

    for (i = 0, k = 0; i < dev->devices; i++) {
	for (j=0;j<2*dev->detectors[i];j++,k++) {
	    res = RequestChannelAttribute(con, COMMAND_REQUEST_GET_THRESHOLD, k, &utmp);
	    if (res == COMMAND_RESPONSE_GET_THRESHOLD) dev->thresholds[k] = 1.* utmp / 100;
	    else namtProcessError(dev, con, COMMAND_REQUEST_GET_THRESHOLD, res, DS_DEBUG_DEBUG);
		    
	    res = RequestChannelAttribute(con, COMMAND_REQUEST_GET_HV, k, &utmp);
	    if (res == COMMAND_RESPONSE_GET_HV) dev->hv[k] = utmp;
	    else namtProcessError(dev, con, COMMAND_REQUEST_GET_HV, res, DS_DEBUG_DEBUG);
	}
    }
    
    return 0;
}

static int WaitNextEvent(NAMTDevice dev, Connection con) {
    int err = 0, res;
    struct SyncAttr sync_attr;
    unsigned short int sync_attr_size = sizeof(struct SyncAttr);
    unsigned short int rtm_time;
    unsigned long smin, smax;

    long diff;
    
    DSTime nexttime;
    DSTime ttmp;
    
    nexttime = dsTimeCreate(NULL);
    ttmp = dsTimeCreate(NULL);
    
    if ((!nexttime)||(!ttmp)) {
	if (nexttime) dsTimeDestroy(nexttime);
	if (ttmp) dsTimeDestroy(ttmp);
	return DS_ERR_MEM;
    }


    dsWait(dev->high_read_pause);
    
    if (dev->prepare_sync) {
	res = SendSimple(con, COMMAND_REQUEST_PREPARE_SYNC);
	if (res != COMMAND_RESPONSE_PREPARE_SYNC) {
	    dsTimeDestroy(nexttime);
	    dsTimeDestroy(ttmp);
	    return namtError(dev, con, COMMAND_REQUEST_PREPARE_SYNC, res);
	}
    }

    dsTimeCurrentTime(ttmp);
    res = RequestDataFixed(con, COMMAND_REQUEST_GET_SYNC, sync_attr_size, (char*)&sync_attr);
    dsTimeCurrentTime(nexttime);
    
    if (res != COMMAND_RESPONSE_GET_SYNC) {
	dsTimeDestroy(nexttime);
	dsTimeDestroy(ttmp);
	return namtError(dev, con, COMMAND_REQUEST_GET_SYNC, res);
    }

    diff = dsTimeDiffMilliseconds(ttmp, nexttime);
    dsDebugServer((diff<(MAXIMAL_SYNC_ERROR/4))?DS_DEBUG_DEBUG:DS_DEBUG_NOTICE,"NAMT(%s). Synchronization request processed in %lims", dev->device.id, diff);

    dsTimeDestroy(ttmp);
    
    rtm_time = sync_attr.rtm;

    dsTimeSubMilliseconds(nexttime, dev->rt_time - rtm_time);
    diff = dsTimeDiffMilliseconds(dev->datatime, nexttime);
    dsTimeAddMilliseconds(nexttime, 11 * dev->rt_time / 10);

    dev->desync = diff;
//    dsDebugServer(DS_DEBUG_WARNING, "NAMT(%s). Desynchronization %lims detected", dev->device.id, diff);

    smin = dsDurationGetMilliseconds(dev->min_sync_error);
    if (labs(diff) > smin) {
	smax = dsDurationGetMilliseconds(dev->max_sync_error);
	if (labs(diff) > smax) {
	    dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Desynchronization %lims detected", dev->device.id, diff);
	    dsTimeDestroy(nexttime);
	    return ADAS_ERR_SYNC;
	} else {
	    dsDebugServer((2*labs(diff)>(smin+smax))?DS_DEBUG_WARNING:DS_DEBUG_NOTICE, "NAMT(%s). Desynchronization %lims detected", dev->device.id, diff);
	}
	
	    
	if (labs(dev->diff)<=labs(diff)) {
	    res = SendAttribute(con, COMMAND_REQUEST_SET_SYNC, (unsigned short)dev->diff);
	    if (res != COMMAND_RESPONSE_SET_SYNC) {
	        dsTimeDestroy(nexttime);
		return namtError(dev, con, COMMAND_REQUEST_SET_SYNC, res);
	    }
	}
	dev->diff = diff;
    } else {
	if (diff) dsDebugServer(DS_DEBUG_INFO, "NAMT(%s). Ignorable desynchronization %lims", dev->device.id, diff);
	dev->diff = 0;
    }

    dsDebugServer(DS_DEBUG_INFO, "NAMT(%s). Waiting for next readout %lu ms", dev->device.id, rtm_time);

    if (((dev->device_type == NAMTDEVICE_TYPE_SNT)||(dev->device_type == NAMTDEVICE_TYPE_ANALYZER))&&(dev->read_control)&&((dev->current_iteration - dev->read_control_last_iteration) >= dev->read_control)&&(rtm_time > dsDurationGetMilliseconds(dev->control_readout_required_time))&&(dev->device.operation_mode != DEVICE_OPERATION_CONFIGURATION_MODE)) {
	dsDebugServer(DS_DEBUG_INFO, "NAMT(%s). Reading control information", dev->device.id);

	ConnectionDisableRetries(con);
	ReadControl(dev, con);
	ConnectionEnableRetries(con);

	dev->read_control_last_iteration = dev->current_iteration;
    }

    ConnectionUnlock(con);
    dsDebugServer(DS_DEBUG_INFO, "NAMT(%s). Unlocking device", dev->device.id);
    err = dsWaitUntil(nexttime);
    dsDebugServer(DS_DEBUG_INFO, "NAMT(%s). Locking device", dev->device.id);
    ConnectionLock(con);
    
    dsTimeDestroy(nexttime);
    
    return err;
}

static int ReadEvent(NAMTDevice dev, Connection con) {
    int i,j;
    int res;
    int err, spectrum_err = 0, events_err = 0;

    uint16_t errors = 0;
    
    NAMTData *data, *cdata;
    NAMTSpectrum *spectrum = NULL;

    NAMTEvents events = NULL;
    void *event;
    
    unsigned short int size = 0;
    unsigned short int csize = 0;
    unsigned short int ssize = 0;
    unsigned short int esize = 0;

    data = (NAMTData*)malloc(dev->devices*sizeof(NAMTData));
    cdata = (NAMTData*)malloc(dev->devices*sizeof(NAMTData));
    
    if ((!data)||(!cdata)) {
	if (data) free(data);
	if (cdata) free(cdata);
	return DS_ERR_MEM;
    }
    
restart:

    err = 0;

    for (i = 0; i < dev->devices; i++) {
    	if (err) {
	    data[i] = NULL;
	    res = COMMAND_RESPONSE_GET;
	    size = namtDataSize(dev, i);
	} else {
	    if (i) dsWait(dev->high_read_pause);
	    res = ComplexRequestData(con, COMMAND_REQUEST_GET, i, &size, (char**)data+i);
	}

	if ((res != COMMAND_RESPONSE_GET)||(size != namtDataSize(dev, i))) {
	    if (res == COMMAND_RESPONSE_GET) {
		err = DS_ERR_INVALID_RESPONSE;
		dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Microcode error: Returned %lu bytes of data instead of %lu", dev->device.id, size, namtDataSize(dev, i));
		free(data[i]);
		ConnectionSkipAll(con);
	    } else err = namtError(dev, con, COMMAND_RESPONSE_GET, res);
	    
	    if (err == ADAS_ERR_EXCHANGE_TIME_IS_OVER) {
		data[i] = NULL;
		if (!i) goto restart;
	    } else {
		for (j=0;j<i;j++) {
		    if (data[j]) free(data[j]);
		    if (cdata[j]) free(cdata[j]);
		}
		
		break;
	    }
	}
	
	if (i) {
	    if ((data[i])&&(data[0]->iter != data[i]->iter)) {
		free(data[i]);
		data[i] = NULL;
		err = ADAS_ERR_EXCHANGE_TIME_IS_OVER;
	    }
	} else {
	    if (data[0]->iter == dev->current_iteration) {
		err = WaitNextEvent(dev,con);
		if (err) break;
		goto restart;
	    } 
	    
	    if (dev->current_iteration > data[0]->iter) {
		if ((0xFFFFFFFF - dev->current_iteration)>dev->recs)  {
		    free(data[0]);
		    err = ADAS_ERR_RESTART;
		    break;
		}

		while (dev->current_iteration + 1 < 0xFFFFFFFF) {
		    dsDebugServer(DS_DEBUG_NOTICE, "NAMT(%s). Skipping Iteration: %u", dev->device.id, dev->current_iteration);
		    err = namtBufferAdd(dev, NULL, NULL, NULL, NULL, 0);
		    if (err) break;
		    dsTimeAddMilliseconds(dev->datatime, dev->rt_time);
		    dev->current_iteration++;
		}
		if (err) {
		    free(data[0]);
		    break;
		}
		dev->current_iteration = 0;
	    } else {	
		if ((data[0]->iter - dev->current_iteration)>dev->recs) {
		    free(data[0]);
		    err = ADAS_ERR_RESTART;
		    break;
		}
	    }
	    
	    while (dev->current_iteration + 1 < data[0]->iter) {
		dsDebugServer(DS_DEBUG_NOTICE, "NAMT(%s). Skipping Iteration: %u", dev->device.id, dev->current_iteration);
		err = namtBufferAdd(dev, NULL, NULL, NULL, NULL, 0);
		if (err) break;
		dsTimeAddMilliseconds(dev->datatime, dev->rt_time);
		dev->current_iteration++;
	    }
	    if (err) {
		free(data[0]);
		break;
	    }
	}

	if (err) {
	    cdata[i] = NULL;
	    res = COMMAND_RESPONSE_GET_COINCIDENCE;
	    csize = namtCDataSize(dev, i);
	} else {
	    dsWait(dev->high_read_pause);
	    res = ComplexRequestData(con, COMMAND_REQUEST_GET_COINCIDENCE, i, &csize, (char**)(cdata+i));
	}

	if ((res != COMMAND_RESPONSE_GET_COINCIDENCE)||(csize != namtCDataSize(dev, i))) {
	    if (res == COMMAND_RESPONSE_GET_COINCIDENCE) {
		err = DS_ERR_INVALID_RESPONSE;
		dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Microcode error: Returned %lu bytes of data instead of %lu", dev->device.id, csize, namtCDataSize(dev, i));
		free(cdata[i]);
		ConnectionSkipAll(con);
	    } else err = namtError(dev, con, COMMAND_RESPONSE_GET_COINCIDENCE, res);

	    if (err == ADAS_ERR_EXCHANGE_TIME_IS_OVER) cdata[i] = NULL;
	    else {
		for (j=0;j<i;j++) {
		    if (data[j]) free(data[j]);
		    if (cdata[j]) free(cdata[j]);
		}
		if (data[i]) free(data[i]);
		
		break;
	    }
	}

	if ((cdata[i])&&(data[0]->iter != cdata[i]->iter)) {
	    free(cdata[i]);
	    cdata[i] = NULL;
	    err = ADAS_ERR_EXCHANGE_TIME_IS_OVER;
	}
    }
    
    if ((!err)||(err == ADAS_ERR_EXCHANGE_TIME_IS_OVER)) {
    	dev->current_iteration = data[0]->iter;

	if ((dev->device_type == NAMTDEVICE_TYPE_SNT)||(dev->device_type == NAMTDEVICE_TYPE_ANALYZER)) {
	    if (dev->write_events) {
		events = (NAMTEvents)malloc(dev->devices*sizeof(NAMTEventsS));
		if (events) {
		    for (i = 0; i < dev->devices; i++) {
			namtEventsInit(dev, events + i, i); event = NULL;
			dsWait(dev->high_read_pause);
			res = ComplexRequestData(con, COMMAND_REQUEST_GET_EVENTS, i, &esize, (char**)(&event));
			if ((res != COMMAND_RESPONSE_GET_EVENTS)||(namtEventsMM(dev, events + i, esize, event))) {
			    if (res == COMMAND_RESPONSE_GET_EVENTS) {
				events_err = DS_ERR_INVALID_RESPONSE;
				dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Microcode error: Invalid size (%u) of response message", dev->device.id, esize);
				ConnectionSkipAll(con);
			    } else {
				events_err = namtError(dev, con, COMMAND_RESPONSE_GET_EVENTS, res);
				if (event) free(event);
			    }
			    namtEventsFree(events+i);
			}
		    }
		} else events_err = DS_ERR_MEM;
	    }

	    if ((dev->write_spectrum)&&((dev->current_iteration - dev->write_spectrum_last_iteration) >= dev->write_spectrum)) {
		spectrum = (NAMTSpectrum*)malloc(dev->devices*sizeof(NAMTSpectrum));
		if (spectrum) {
	    	    for (i = 0; i < dev->devices; i++) {
			dsWait(dev->high_read_pause);
			res = ComplexRequestData(con, COMMAND_REQUEST_GET_SPECTRUM, i, &ssize, (char**)(spectrum+i));

			if ((res != COMMAND_RESPONSE_GET_SPECTRUM)||(ssize != namtSpectrumSize(dev, i))) {
			    if (res == COMMAND_RESPONSE_GET_SPECTRUM) {
				spectrum_err = DS_ERR_INVALID_RESPONSE;
				dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Microcode error: Returned %lu bytes of data instead of %lu", dev->device.id, ssize, namtSpectrumSize(dev, i));
				free(spectrum[i]); spectrum[i] = NULL;
				ConnectionSkipAll(con);
			    } else {
				spectrum_err = namtError(dev, con, COMMAND_RESPONSE_GET_SPECTRUM, res);
				spectrum[i] = NULL;
			    }
			}
		    }
		} else spectrum_err = DS_ERR_MEM;
	    }
	}
    }
    
    dsWait(dev->high_read_pause);

    if ((dev->protocol_version>5)&&(RequestAttribute(con, COMMAND_REQUEST_ERRORS, &errors)==COMMAND_RESPONSE_ERRORS)&&(errors)) {
	dsDebugServer(DS_DEBUG_CRITICAL, "NAMT(%s). Microcode Errors: %lx (flags)", dev->device.id, errors);
	namtErrorLogMicrocodeErrors(dev, errors);
    }

    if ((!err)||(err == ADAS_ERR_EXCHANGE_TIME_IS_OVER)) {
	dsDebugServer(DS_DEBUG_INFO, "NAMT(%s). Iteration %u processed successfully", dev->device.id, data[0]->iter);
	err = namtBufferAdd(dev, data, cdata, events, spectrum_err?NULL:spectrum, errors);

	dsTimeAddMilliseconds(dev->datatime, dev->rt_time);

	if (err) {
	    for (i=0;i<dev->devices;i++) {
		if (data[i]) free(data[i]);
		if (cdata[i]) free(cdata[i]);
	    }
	}
	
	if (events) {
	    if (err) {
		for (i=0;i<dev->devices;i++) {
		    namtEventsFree(events+i);
		}
		free(events);
	    }
	}

	if (spectrum) {
	    if ((err)||(spectrum_err)) {
		for (i=0;i<dev->devices;i++) {
		    if (spectrum[i]) free(spectrum[i]);
		}
		free(spectrum);
	    } else {
		dev->write_spectrum_last_iteration = dev->current_iteration;
	
		res = SendSimple(con, COMMAND_REQUEST_SPECTRUM_RESET);
		if (res != COMMAND_RESPONSE_SPECTRUM_RESET) {
		    namtError(dev, con, COMMAND_RESPONSE_SPECTRUM_RESET, res);
		}
	    }
	}
    }

    if (err) {
	free(data);
	free(cdata);
    }

    return err;
}

#define REINITIALIZE \
    BufferWaitDeviceEventsProcessed(dev->device.buffer, (Device)dev); \
    ConnectionUnlock(con); \
    dsLockWrite(dev->ilock); \
    DeviceCloseConnection((Device)dev); \
    dsWait(slp); \
    continue;


int namtReader(NAMTDevice dev) {
	int err = 0;
	int informated = 0;
	int res;
	
	Connection con;

	DSDuration slp;
	
	DSTime curtime;
//	DSDuration rt_time;
	DSDuration trigger_duration;

	DSTime nextread;
	DSTime sync_time;
	DSTime start_time;
	DSDuration first_sleep;

	xmlNodePtr cfgnode, chipcfg, thresholds, layout;	

	unsigned short int info_size;
	Version info = NULL;
	
	unsigned short int layoutinfo_size;
	LayoutInfo layoutinfo = NULL; 

	unsigned long i, j, k, l;
	unsigned short int tmp;
	
#ifdef CALIBRATION_TIME
	DSTime tm1, tm2, tm3;
	DSDuration dur1, dur2;
	unsigned short int tmp2;

	unsigned long round_trip_max;
	DSDispersionContext round_trip_time;
	DSDispersionContext fujitsu_internal_clock;
#endif /* CALIBRATION_TIME */

	unsigned int fujitsu_round_trip_time = FUJITSU_ROUND_TRIP_TIME; /* ms */
	unsigned int sync_coef = 1;

	unsigned char reset_lpc = 0;
	unsigned char voltage = DEFAULT_VOLTAGE, voltage1 = DEFAULT_VOLTAGE, voltage2 = DEFAULT_VOLTAGE;
	unsigned char devices = 1;
	unsigned short int detectors;
	unsigned short int used_detectors;
	unsigned short int device_type, device_version;
	
	unsigned char plain = 0;
	unsigned char devices_are_set = 0;
	
	uint16_t threshold;
	
	slp = dsDurationCreate(NULL);
	trigger_duration = dsDurationCreate(NULL);
	
	curtime = dsTimeCreate(NULL);
	nextread = dsTimeCreate(NULL);
	sync_time = dsTimeCreate(NULL);
	start_time = dsTimeCreate(NULL);
	
	first_sleep = dsDurationCreate(NULL);
	
	if ((!slp)||(!trigger_duration)||(!nextread)||(!curtime)||(!sync_time)||(!first_sleep)) {
	    if (slp) dsDurationDestroy(slp);
	    if (trigger_duration) dsDurationDestroy(trigger_duration);
	    if (curtime) dsTimeDestroy(curtime);
	    
	    if (nextread) dsTimeDestroy(nextread);
	    if (sync_time) dsTimeDestroy(sync_time);
	    if (start_time) dsTimeDestroy(start_time);

	    Abort();
	    return DS_ERR_MEM;
	}
	

	dsDurationImportMicroseconds(slp, NAMT_TIMEOUT_REINIT);
	dsDurationImportMilliseconds(first_sleep, NAMT_DEFAULT_READOUT_DURATION);
	dsDurationImportMicroseconds(trigger_duration, DEFAULT_TRIGGER_DURATION);

	dev->write_spectrum = SNT_DEFAULT_WRITE_SPECTRUM;
	dev->write_events = SNT_DEFAULT_WRITE_EVENTS;
	dev->read_control = SNT_DEFAULT_READ_CONTROL;
	
	dsLockWrite(dev->ilock);
	
	while (DeviceContinueLoop((Device)dev)) {
	    err = 0;

	    dev->reset = 0;
	    dev->protocol_version = 0;
    	    dev->device_type = 0;
	    dev->device_version = 0;
	    dev->handled_events = 0;

	    dev->device.status = DEVICE_STATUS_NOT_CONNECTED;
	    dev->current_iteration = 0;
	    dev->desync = 0;
	    
	    dev->write_spectrum_last_iteration = 0;
	    dev->read_control_last_iteration = 0;
	    
	    for (i=0;i<MAX_NAMT_SENSORS;i++) {
		dev->hv[i] = -1;
		dev->thresholds[i] = -1;
	    }

	    dev->layout = 0;
	    devices_are_set = 0;

		// ReInitialising Fujitsu MCU
	    dsDebugServer(DS_DEBUG_INFO, "NAMT(%s). Reininitializing!", dev->device.id);

	    
		// 0. Reading default parameters
	    cfgnode = dsConfigGetConfigByReference(dev->device.cfg, dev->device.cfgref);
	    if (cfgnode) {
		chipcfg = dsConfigGetSubNode(dev->device.cfg, cfgnode, "Configuration", 0);
		if (chipcfg) {
		    reset_lpc = dsConfigGetNodeBoolean(dev->device.cfg, chipcfg, "ResetLPC", 0);
		    dsConfigGetNodeDuration(dev->device.cfg, chipcfg, "ReadoutInterval", first_sleep);
		    dsConfigGetNodeDuration(dev->device.cfg, chipcfg, "SignificantSyncError", dev->min_sync_error);
		    dsConfigGetNodeDuration(dev->device.cfg, chipcfg, "MaximalSyncError", dev->max_sync_error);
		    dsConfigGetNodeDuration(dev->device.cfg, chipcfg, "HighReadPause", dev->high_read_pause);
		    if (dsConfigGetNodeBoolean(dev->device.cfg, chipcfg, "ConfigurationMode", 0)) {
			dev->device.operation_mode = DEVICE_OPERATION_CONFIGURATION_MODE;
		    }
		}
		dsConfigReturnConfig(dev->device.cfg, cfgnode);
	    }
	    
	    dev->rt_time = dsDurationGetMilliseconds(first_sleep);

		// 1. Configuring serial port
	    con = DeviceInitializeConnection((Device)dev);
	    if (!con) {
		if (informated != 1) {
		    dsDebugServer(DS_DEBUG_CRITICAL, "NAMT(%s). Error opening communication device!", dev->device.id);
		    informated = 1;
		}
		dsWait(slp);
		continue;
	    }

	    if ((reset_lpc)&&(con->type == CONNECTION_TYPE_UART)) {
		ConnectionResetSignal(con, DS_CONNECTION_SIGNAL_RTS);
		ConnectionSetSignal(con, DS_CONNECTION_SIGNAL_DTR);
		ConnectionSetSignal(con, DS_CONNECTION_SIGNAL_BREAK);
		dsWaitMilliseconds(25);
		ConnectionResetSignal(con, DS_CONNECTION_SIGNAL_DTR);
		dsWaitMilliseconds(10);
	    }
	    
		// 2. Skiping all pending data on COM port
	    ConnectionReadAll(con);

		// 3. Geting Version Information
	    if (info) { free(info); info = NULL; }
    
	    res = RequestData(con, COMMAND_REQUEST_INFO, &info_size, (char**)&info);
	    if ((res == COMMAND_RESPONSE_INFO)&&(info_size >= sizeof(VersionS))) {
		informated = 0;
		
		dsDebugServer(DS_DEBUG_WARNING, "NAMT(%s). Protocol Version: %u", dev->device.id, info->version);
		dev->data_size = ((Version1)info)->data_size * 8;
		if ((dev->data_size!=16)&&(dev->data_size!=32)) {
		    dsDebugServer(DS_DEBUG_CRITICAL, "NAMT(%s). Basic Data Size %u not supported!", dev->device.id, dev->data_size);
		    DeviceCloseConnection((Device)dev);
		    dsWait(slp);
		    continue;
		}
		dsDebugServer(DS_DEBUG_NOTICE, "NAMT(%s). Basic Data Size: %u bits", dev->device.id, dev->data_size);

		if ((info->version < PROTOCOL_MINIMAL_VERSION)||(info->version > PROTOCOL_MAXIMAL_VERSION)) {
		    dsDebugServer(DS_DEBUG_CRITICAL, "NAMT(%s). Unsupported version of protocol (microcode)!", dev->device.id);
		    DeviceCloseConnection((Device)dev);
		    dsWait(slp);
		    continue;
		} else if (info->version < 6) {
		    detectors = (unsigned char)((Version1)info)->detectors;
		    device_type = NAMTDEVICE_TYPE_NAMT;
		    device_version = info->version;

		    dsDebugServer(DS_DEBUG_NOTICE, "NAMT(%s). Clock Rate: %u MHZ", dev->device.id, ((Version1)info)->clock);
		    dsDebugServer(DS_DEBUG_NOTICE, "NAMT(%s). R-Bus Divider: %u, Ext-Bus Divider: %u", dev->device.id, ((Version1)info)->r_divider, ((Version1)info)->ext_divider);
		    dsDebugServer(DS_DEBUG_NOTICE, "NAMT(%s). Detectors Supported: %u", dev->device.id, ((Version1)info)->detectors);
		} else {
		    if (info->version == 6) {
			devices = (unsigned char)((Version6)info)->devices;
			device_type = ((Version6)info)->device_type;
			device_version = ((Version6)info)->device_version;
			dev->handled_events = 0;

			dsDebugServer(DS_DEBUG_NOTICE, "NAMT(%s). Clock Rate: %u MHZ", dev->device.id, ((Version6)info)->clock);
		    } else {
			devices = (unsigned char)((Version7)info)->devices;
			device_type = ((Version7)info)->device_type;
			device_version = ((Version7)info)->device_version;
			dev->handled_events = ((Version7)info)->handled_events;

			dsDebugServer(DS_DEBUG_NOTICE, "NAMT(%s). Clock Rate: %u MHZ", dev->device.id, ((Version7)info)->clock);
		    }


		    switch (device_type) {
			case NAMTDEVICE_TYPE_NAMT:
			    dsDebugServer(DS_DEBUG_NOTICE, "NAMT(%s). NAMT Device. Version: %u", dev->device.id, device_version);
			    if ((device_version<NAMT_MINIMAL_VERSION)||(device_version>NAMT_MAXIMAL_VERSION)) device_version = 0;
			    else plain = 0;
			break;
			case NAMTDEVICE_TYPE_SNT:
			    dsDebugServer(DS_DEBUG_NOTICE, "NAMT(%s). SNT Device. Version: %u", dev->device.id, device_version);
			    if ((device_version<SNT_MINIMAL_VERSION)||(device_version>SNT_MAXIMAL_VERSION)) device_version = 0;
			    else plain = 0;
			break;
			case NAMTDEVICE_TYPE_ANALYZER:
			    dsDebugServer(DS_DEBUG_NOTICE, "NAMT(%s). SA Device. Version: %u", dev->device.id, device_version);
			    if ((device_version<ANALYZER_MINIMAL_VERSION)||(device_version>ANALYZER_MAXIMAL_VERSION)) device_version = 0;
			    else plain = 1;
			break;
			default:
			    dsDebugServer(DS_DEBUG_CRITICAL, "NAMT(%s). Unsupported device type: %u!", dev->device.id, device_type);
			    DeviceCloseConnection((Device)dev);
			    dsWait(slp);
			    continue;
		    }
		    
		    if (!device_version) {
			dsDebugServer(DS_DEBUG_CRITICAL, "NAMT(%s). Unsupported version of microcode!", dev->device.id);
			DeviceCloseConnection((Device)dev);
			dsWait(slp);
			continue;
		    }
		    
		    dsDebugServer(DS_DEBUG_NOTICE, "NAMT(%s). Handled Events: %lX", dev->device.id, dev->handled_events);

		    if (info->version == 6) {
			dsDebugServer(DS_DEBUG_NOTICE, "NAMT(%s). Devices Supported: %u", dev->device.id, ((Version6)info)->devices);
			detectors = (unsigned char)((Version6)info)->device[0].detectors;
			for (i=1;i<devices;i++) {
			    if ((unsigned char)((Version6)info)->device[i].detectors != detectors) {
				device_version = 0;
				break;
			    }
			}
		    } else {
			dsDebugServer(DS_DEBUG_NOTICE, "NAMT(%s). Devices Supported: %u", dev->device.id, ((Version7)info)->devices);
			detectors = (unsigned char)((Version7)info)->device[0].detectors;
			for (i=1;i<devices;i++) {
			    if ((unsigned char)((Version7)info)->device[i].detectors != detectors) {
				device_version = 0;
				break;
			    }
			}
		    }

		    if (!device_version) {
			dsDebugServer(DS_DEBUG_CRITICAL, "NAMT(%s). Only devices with equivalent number of detectors are supported yet!", dev->device.id);
			DeviceCloseConnection((Device)dev);
			dsWait(slp);
			continue;
		    }

		    dsDebugServer(DS_DEBUG_NOTICE, "NAMT(%s). Detectors Supported: %u", dev->device.id, detectors);


		    switch (device_type) {
			case NAMTDEVICE_TYPE_ANALYZER:
			    res = RequestData(con, COMMAND_REQUEST_LAYOUT_INFO, &layoutinfo_size, (char**)&layoutinfo);
			    if (res == COMMAND_RESPONSE_LAYOUT_INFO) {
				if (layoutinfo_size >= sizeof(LayoutInfoS)) {
				    if (layoutinfo->flags) {
					if (layoutinfo_size == sizeof(LayoutInfo7S)) {
					    dev->layout = 1;
					    dev->layout_sensors = ((LayoutInfo7)layoutinfo)->sensors;
					    dev->layout_channels = ((LayoutInfo7)layoutinfo)->rulesets;
					    dev->layout_spectrums = ((LayoutInfo7)layoutinfo)->spectrums;
					    dev->layout_rules = ((LayoutInfo7)layoutinfo)->rules;
					    dsDebugServer(DS_DEBUG_NOTICE, "NAMT(%s). Configurable Layout (sensors: %hu, rulesets: %hu, rules: %hu, spectrums: %hu)", dev->device.id, dev->layout_sensors, dev->layout_channels, dev->layout_rules, dev->layout_spectrums);
					} else {
					    dsDebugServer(DS_DEBUG_CRITICAL, "NAMT(%s). Invalid response to LayoutInfo request", dev->device.id);
					    err = ADAS_ERR_BOARD;
					}
				    } else {
					dsDebugServer(DS_DEBUG_NOTICE, "NAMT(%s). Layout configuration is not allowed", dev->device.id);
					dev->layout = 0;
				    }
				} else {
				    dsDebugServer(DS_DEBUG_CRITICAL, "NAMT(%s). Invalid response to LayoutInfo request", dev->device.id);
				    err = ADAS_ERR_BOARD;
				}
				free(layoutinfo);layoutinfo=NULL;
			    } else {
				err = namtError(dev, con, COMMAND_REQUEST_LAYOUT_INFO, res);
			    }

			    if (err) {
				DeviceCloseConnection((Device)dev);
				dsWait(slp);
				continue;
			    }

			break;
			default:
			    ;
		    }

		}
	    } else {
		if ((informated != res)||(dsDebugCheckLevel(DS_DEBUG_NOTICE))) {
		    namtError(dev, con, COMMAND_REQUEST_INFO, res);
		    informated = res;
		}
		DeviceCloseConnection((Device)dev);
		dsWait(slp);
		continue;
	    }
	    
	    dev->device.status = DEVICE_STATUS_INITIALIZING;
		
	    dev->protocol_version = info->version;
    	    dev->device_type = device_type;
	    dev->device_version = device_version;
	    dev->plain = plain;

	    if (plain) {
		dev->coincidences = 0;
		dev->spectrums = 0;
	    }
	    
	    
	    if (device_type == NAMTDEVICE_TYPE_NAMT) {
		if (device_version > 4) dev->prepare_sync = 1;
		else dev->prepare_sync = 0;
	    } else {
		if (device_version > 3) dev->prepare_sync = 1;
		else dev->prepare_sync = 0;
	    }
	    

	    cfgnode = dsConfigGetConfigByReference(dev->device.cfg, dev->device.cfgref);
	    if (cfgnode) {
		chipcfg = dsConfigGetSubNode(dev->device.cfg, cfgnode, "Configuration", 0);
		if (chipcfg) {
		    switch (device_type) {
			case NAMTDEVICE_TYPE_NAMT:
			    dsConfigGetNodeDuration(dev->device.cfg, chipcfg, "TriggerDuration", trigger_duration);
			    voltage = (unsigned char)dsConfigGetNodeLong(dev->device.cfg, chipcfg, "DAVoltage", DEFAULT_VOLTAGE);
			    voltage1 = (unsigned char)dsConfigGetNodeLong(dev->device.cfg, chipcfg, "DAVoltage1", voltage);
			    voltage2 = (unsigned char)dsConfigGetNodeLong(dev->device.cfg, chipcfg, "DAVoltage2", voltage);
			    devices = (unsigned char)dsConfigGetNodeLong(dev->device.cfg, chipcfg, "Devices", devices);
			break;
			case NAMTDEVICE_TYPE_ANALYZER:
			    if (dev->layout) {
				layout = dsConfigGetSubNode(dev->device.cfg, cfgnode, "Layout", 1);
				if (layout) {
				    err = namtLayoutProcess(dev, con, layout);
				    if (err) break;
				    else {
					devices_are_set = 1;
					detectors = dev->detectors[0]; /**/
				    }
				}
			    }
			// Do not break, we consider SNT options as well
			case NAMTDEVICE_TYPE_SNT:
			    dev->write_spectrum = dsConfigGetNodeLong(dev->device.cfg, chipcfg, "SpectrumReadInterval", dev->write_spectrum);
			    dev->read_control = dsConfigGetNodeLong(dev->device.cfg, chipcfg, "ControlReadInterval", dev->read_control);
			    dev->write_events = dsConfigGetNodeBoolean(dev->device.cfg, chipcfg, "ReadEvents", dev->write_events);
			    dsConfigGetNodeDuration(dev->device.cfg, chipcfg, "ControlReadoutRequiredTime", dev->control_readout_required_time);

			    thresholds = dsConfigGetSubNode(dev->device.cfg, chipcfg, "SoftwareThresholds", 1);
			    if (thresholds) {
			      for (k=0,l=0;k<devices;k++) {
				for (i=0;(i<(plain?1:2))&&(!err);i++) {
				    for (j=0;(j<detectors)&&(!err);j++,l++) {
					if (plain) {
					    if (devices>1)
						threshold = dsConfigGetInt(dev->device.cfg, SNT_DEFAULT_SOFTWARE_THRESHOLD, thresholds, "SoftwareThreshold[@device='%s-%u' and @sensor=%u]", dev->device.id, k+1, j+1);
					    else
						threshold = dsConfigGetInt(dev->device.cfg, SNT_DEFAULT_SOFTWARE_THRESHOLD, thresholds, "SoftwareThreshold[@sensor=%u]", j+1);
					} else {
					    if (devices>1)
						threshold = dsConfigGetInt(dev->device.cfg, SNT_DEFAULT_SOFTWARE_THRESHOLD, thresholds, "SoftwareThreshold[@device='%s-%u' and @layer=%u and @sensor=%u]", dev->device.id, k+1, i+1, j+1);
					    else
						threshold = dsConfigGetInt(dev->device.cfg, SNT_DEFAULT_SOFTWARE_THRESHOLD, thresholds, "SoftwareThreshold[@layer=%u and @sensor=%u]", i+1, j+1);
					}
					res = SendChannelAttribute(con, COMMAND_REQUEST_SET_SOFTWARE_THRESHOLD, l, threshold);
					if (res != COMMAND_RESPONSE_SET_SOFTWARE_THRESHOLD) {
					    dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Error setting threshold for device %u layer %u detector %u", dev->device.id, k+1, i+1, j+1);
					    err = namtError(dev, con, COMMAND_REQUEST_SET_THRESHOLD, res);
					}
				    }
				}
			      }
			    }
			break;
		    }
		}
		dsConfigReturnConfig(dev->device.cfg, cfgnode);
	    }
	    
	    if (err) {
		DeviceCloseConnection((Device)dev);
		dsWait(slp);
		continue;
	    }

	    if (!devices_are_set) {
	        err = namtAllocate(dev, devices);
		if (err) {
		    dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Error: %lx (%s)", dev->device.id, err, dsError(err, NULL));
		    DeviceCloseConnection((Device)dev);
		    dsWait(slp);
		    continue;
		}
		
		for (i=0;i<devices;i++) {
		    err = namtAllocateNextDevice(dev, detectors);
		    if (err) break;
		}
	    
		if (err) {
		    dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Error allocating sub-device: %lx (%s)", dev->device.id, err, dsError(err, NULL));
		    DeviceCloseConnection((Device)dev);
		    dsWait(slp);
		    continue;
		}
	    }


		// 4. Setting Default Parameters
	    if (info->version>=2) {
	        res = SendAttribute(con, COMMAND_REQUEST_SET_SYNC_DIVIDER, SYNC_DIVIDER);
		if (res != COMMAND_RESPONSE_SET_SYNC_DIVIDER) {
		    namtError(dev, con, COMMAND_REQUEST_SET_SYNC_DIVIDER, res);
		    DeviceCloseConnection((Device)dev);
		    dsWait(slp);
		    continue;
		}
		dsDebugServer(DS_DEBUG_INFO, "NAMT(%s). Synchronization Divider: %.2f%%", dev->device.id, 100. / SYNC_DIVIDER);
	    }

	    res = SendAttribute(con, COMMAND_REQUEST_SET_COUNTING_DURATION, dev->rt_time);
	    if (res != COMMAND_RESPONSE_SET_COUNTING_DURATION) {
		namtError(dev, con, COMMAND_REQUEST_SET_COUNTING_DURATION, res);
		DeviceCloseConnection((Device)dev);
		dsWait(slp);
		continue;
	    }

	    dsDebugServer(DS_DEBUG_NOTICE, "NAMT(%s). Counting Duration: %ums", dev->device.id, dev->rt_time);
	    dsDebugServer(DS_DEBUG_NOTICE, "NAMT(%s). Readout per Record: %u", dev->device.id, dev->recs);

	    if ((dev->device_type == NAMTDEVICE_TYPE_SNT)||(dev->device_type == NAMTDEVICE_TYPE_ANALYZER)) {
	        err = ReadControl(dev, con);
		if (err) {
		    dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Error reading control information: %lx (%s)", dev->device.id, err, dsError(err, NULL));
		    DeviceCloseConnection((Device)dev);
		    dsWait(slp);
		    continue;
		}
	    } else if (device_type == NAMTDEVICE_TYPE_NAMT) {
		res = SendAttribute(con, COMMAND_REQUEST_SET_TRIGGER_DURATION, (unsigned short)dsDurationGetMilliseconds(trigger_duration));
		if (res != COMMAND_RESPONSE_SET_TRIGGER_DURATION) {
		    namtError(dev, con, COMMAND_REQUEST_SET_TRIGGER_DURATION, res);
		    DeviceCloseConnection((Device)dev);
		    dsWait(slp);
		    continue;
		}
		dsDebugServer(DS_DEBUG_INFO, "NAMT(%s). Trigger Duration: %uus", dev->device.id, dsDurationGetMilliseconds(trigger_duration));

		res = SendAttribute(con, COMMAND_REQUEST_SET_DA_VOLTAGE, voltage1);
	        if (res != COMMAND_RESPONSE_SET_DA_VOLTAGE) {
	    	    namtError(dev, con, COMMAND_REQUEST_SET_DA_VOLTAGE, res);
		    DeviceCloseConnection((Device)dev);
		    dsWait(slp);
		    continue;
		}
	    
	        res = SendAttribute(con, COMMAND_REQUEST_SET_DA_VOLTAGE, 0x100 + voltage2);
		if (res != COMMAND_RESPONSE_SET_DA_VOLTAGE) {
		    namtError(dev, con, COMMAND_REQUEST_SET_DA_VOLTAGE, res);
		    DeviceCloseConnection((Device)dev);
		    dsWait(slp);
		    continue;
		}
		
		dsDebugServer(DS_DEBUG_INFO, "NAMT(%s). D/A Voltages: %u, %u", dev->device.id, voltage1, voltage2);
	    }
	    
		// 5. Calibrating Synchronization Timings
#ifdef CALIBRATION_TIME
	    if (dev->device.mode != DEVICE_OPERATION_CONFIGURATION_MODE) {
		tmp = 0;
		round_trip_max = 0;

		tm1 = dsTimeCreate(NULL);
		tm2 = dsTimeCreate(NULL);
		tm3 = dsTimeCreate(NULL);
		dur1 = dsDurationCreate(NULL);
		dur2 = dsDurationCreate(NULL);
		if ((!tm1)||(!tm2)||(!tm3)||(!dur1)||(!dur2)) err = DS_ERR_MEM;
		
		round_trip_time = dsDispersionCreateContext(NULL);
		if (round_trip_time) err = dsDispersionInitContext(round_trip_time);
		else err = DS_ERR_MEM;
		
		fujitsu_internal_clock = dsDispersionCreateContext(NULL);
		if (fujitsu_internal_clock) err = dsDispersionInitContext(fujitsu_internal_clock);
		else err = DS_ERR_MEM;
		
		if (err) goto calibration_end;
		
		dsTimeCurrentTime(tm3);
		dsDurationImportMilliseconds(dur1, 1000);
		
		res = SendSimple(con, COMMAND_REQUEST_CALIBRATION_START);
		if (res != COMMAND_RESPONSE_CALIBRATION_START) {
		    namtError(dev, con, COMMAND_REQUEST_CALIBRATION_START, res);
		    goto calibration_end;
		}
		
		for (i=0;i<CALIBRATION_TIME;i++) {
		    tmp2 = tmp;

		    dsTimeAdd(tm3, dur1);
		    err = dsWaitUntil(tm3);
		    if (err) break;
		    
		    dsTimeCurrentTime(tm1);
		    res = RequestAttribute(con, COMMAND_REQUEST_CALIBRATION, &tmp);
		    dsTimeCurrentTime(tm2);
	    
		    if (res != COMMAND_RESPONSE_CALIBRATION) {
			err = DS_ERR_INVALID_RESPONSE;
			namtError(dev, con, COMMAND_REQUEST_CALIBRATION, res);
			break;
		    }
		    
		    dsTimeCalcDuration(dur2, tm1, tm2);
		    dsDispersionEvent(round_trip_time, dsDurationGetMilliseconds(dur2));
		    if (dsDurationGetMilliseconds(dur2)>round_trip_max) round_trip_max = dsDurationGetMilliseconds(dur2);
		    if (i) dsDispersionEvent(fujitsu_internal_clock, tmp - tmp2);

		    dsDebugServer(DS_DEBUG_DEBUG, "Roundtrip time: %lu, Clock Precision: %lu", dsDurationGetMilliseconds(dur2), tmp-tmp2);
		    
		    if (!DeviceContinueLoop((Device)dev)) break;
		}
		
		if (err) goto calibration_end;
		
		res = RequestAttribute(con, COMMAND_REQUEST_CALIBRATION_FINISH, &tmp);
		if (res != COMMAND_RESPONSE_CALIBRATION_FINISH) {
		    namtError(dev, con, COMMAND_REQUEST_CALIBRATION_FINISH, res);
		    goto calibration_end;
		}
		
		dsDispersionFinalyze(round_trip_time);
		dsDispersionFinalyze(fujitsu_internal_clock);

		dsDebugServer(DS_DEBUG_INFO, "NAMT(%s). Round trip message time (ms): %lf(%lf). Maximum: %lu ms", dev->device.id, dsDispersionGetMedian(round_trip_time), dsDispersionGetDispersion(round_trip_time), round_trip_max);
		dsDebugServer((dsDispersionGetMedian(fujitsu_internal_clock) - 3 * dsDispersionGetDispersion(fujitsu_internal_clock) < 999 || dsDispersionGetMedian(fujitsu_internal_clock) + 3 * dsDispersionGetDispersion(fujitsu_internal_clock) > 1001) ? DS_DEBUG_ERROR : DS_DEBUG_INFO, "NAMT(%s). Internal clock precision: %lf(%lf)", dev->device.id, dsDispersionGetMedian(fujitsu_internal_clock), dsDispersionGetDispersion(fujitsu_internal_clock));
		
		fujitsu_round_trip_time = dsDispersionGetMedian(round_trip_time);

calibration_end:

		if (tm1) dsTimeDestroy(tm1);
		if (tm2) dsTimeDestroy(tm2);
		if (tm3) dsTimeDestroy(tm3);
		if (dur1) dsDurationDestroy(dur1);
		if (dur2) dsDurationDestroy(dur2);
		if (round_trip_time) dsDispersionDestroyContext(round_trip_time);
		if (fujitsu_internal_clock) dsDispersionDestroyContext(fujitsu_internal_clock);

		if (err) {
		    err = 0;
		    DeviceCloseConnection((Device)dev);
		    dsWait(slp);
		    continue;
		}
	    }
#endif /* CALIBRATION_TIME */

		// 7. Synchronizing

	    dsLockReturn(dev->ilock);
	    
	    if (dev->device.operation_mode == DEVICE_OPERATION_CONFIGURATION_MODE) {
		dsTimeCurrentTime(start_time);
		dsTimeAddMilliseconds(start_time, 100  + sync_coef*NAMT_INITIALIZATION_TIME);
		dsTimeCurrentTime(sync_time);
		dsTimeAddMilliseconds(sync_time, 100);
	    } else CalculateStartTime(dev, dev->rt_time, sync_time, start_time);
	    
	    res = SendSimple(con, COMMAND_REQUEST_PREPARE_SYNC);
	    if (res != COMMAND_RESPONSE_PREPARE_SYNC) {
		if (namtError(dev, con, COMMAND_REQUEST_PREPARE_SYNC, res) == ADAS_ERR_EXCHANGE_TIME_IS_OVER) sync_coef++;
		REINITIALIZE
	    }
	    
	    if ((!DeviceContinueLoop((Device)dev))||(dsWaitUntil(sync_time))) continue;

	    dsTimeCopy(nextread, start_time);
	    dsTimeSubMilliseconds(nextread, fujitsu_round_trip_time);
	    
	    dsTimeCurrentTime(curtime);

	    res = SendAttribute(con, COMMAND_REQUEST_SYNC, dsTimeDiffMilliseconds(curtime, nextread));
	    if (res != COMMAND_RESPONSE_SYNC) {
		namtError(dev, con, COMMAND_REQUEST_SYNC, res);
		REINITIALIZE
	    }
	    dsDebugServer(DS_DEBUG_INFO, "NAMT(%s). Synchronization successeded. Entering main loop!", dev->device.id);

		// 8. Reading Events (Main Loop)
	    dsTimeCopy(nextread, start_time);
	    dsTimeCopy(dev->datatime, start_time);
	    dev->sf = 1;
	    if ((!DeviceContinueLoop((Device)dev))||(dsWaitUntil(start_time))) {
		REINITIALIZE
	    }

	    /*dsDurationCopy(first_sleep, dev->rt_time); already done*/
	    dsDurationMultiply(first_sleep, 1.1);
	    dsTimeAdd(nextread, first_sleep);

	    dev->current_iteration = 0;
	    dev->diff = 0;

	    dev->device.status = DEVICE_STATUS_OK;

	    ConnectionUnlock(con);
	    dsDebugServer(DS_DEBUG_INFO, "NAMT(%s). Unlocking device", dev->device.id);
	    if ((!DeviceContinueLoop((Device)dev))||(dsWaitUntil(nextread))) {
		dsDebugServer(DS_DEBUG_INFO, "NAMT(%s). Locking device", dev->device.id);
		ConnectionLock(con);
		REINITIALIZE
	    }
	    dsDebugServer(DS_DEBUG_INFO, "NAMT(%s). Locking device", dev->device.id);
	    ConnectionLock(con);
	    
	    while ((DeviceContinueLoop((Device)dev))&&(!err)&&(!dev->reset)) {
		if (dev->device.operation_mode == DEVICE_OPERATION_CONFIGURATION_MODE) {
		    dsTimeAddMilliseconds(dev->datatime, dev->rt_time);
		    dev->current_iteration++;
		    err = 0;
		} else err = ReadEvent(dev, con);
		if (err) {
		    if (err == ADAS_ERR_EXCHANGE_TIME_IS_OVER) {
			dsDebugServer(DS_DEBUG_WARNING, "NAMT(%s). We was late to read data", dev->device.id);
			err = 0;
			continue;
		    } else if (err != ADAS_ERR_FATAL) {
			dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Error reading event: %X(%s)", dev->device.id, err, dsError(err, NULL));
			err = 0;
		    } else {
			dsDebugServer(DS_DEBUG_FATAL, "NAMT(%s). Fatal error during event readout", dev->device.id);
		    }
		    break;
		}
		
		err = WaitNextEvent(dev, con);
		if (err) {
		    if (err != ADAS_ERR_FATAL) {
			dsDebugServer(DS_DEBUG_ERROR, "NAMT(%s). Error waiting event: %X(%s)", dev->device.id, err, dsError(err, NULL));
			err = 0;
			DeviceCloseConnection((Device)dev);
			dsWait(slp);
		    } else {
			dsDebugServer(DS_DEBUG_FATAL, "NAMT(%s). Fatal error during event wait", dev->device.id);
		    }
		    break;
		}
	    }
	    
	    if ((!err)&&(DeviceContinueLoop((Device)dev))) {
		if (dev->reset) {
		    if ((!reset_lpc)||(con->type != CONNECTION_TYPE_UART)) {
			res = SendSimple(con, COMMAND_REQUEST_RESET);
			if (res != COMMAND_REQUEST_RESET) {
			    namtError((NAMTDevice)dev, con, COMMAND_REQUEST_RESET, res);
			}
		    }
		}
		REINITIALIZE
	    }
	}
	
	DeviceCloseConnection((Device)dev);

	BufferWaitDeviceEventsProcessed(dev->device.buffer, (Device)dev);
	dsLockReturn(dev->ilock);

        if (info) free(info);

	dsDurationDestroy(slp);
	dsDurationDestroy(trigger_duration);
	dsTimeDestroy(curtime);

	dsTimeDestroy(nextread);
	dsTimeDestroy(sync_time);
	dsTimeDestroy(start_time);

	if (err) Abort();
	return 0;
}

void namtReset(Device dev) {
    ds_assert_return(dev);

    dev->status = DEVICE_STATUS_RESETING;
    ((NAMTDevice)dev)->reset = 1;
}


static int namtGetCurrentData(NAMTDevice dev, unsigned char device, char *data) {
    int res;

    Connection con;

    con = DeviceTryConnection((Device)dev);
    if (!con) return DS_ERR_BUSY;

    ConnectionSilence(con, DS_DEBUG_DEBUG);
    
    res = ComplexRequestDataFixed(con, COMMAND_REQUEST_GET_CURRENT, device, 0, data);
    if (res != COMMAND_RESPONSE_GET_CURRENT) {
	namtProcessError(dev, con, COMMAND_REQUEST_GET_CURRENT, res, DS_DEBUG_DEBUG);
	DeviceReturnConnection((Device)dev);
	return DS_ERR_INVALID_RESPONSE;
    }

    ConnectionSilence(con, DS_DEBUG_NOTHING);
    
    DeviceReturnConnection((Device)dev);

    return 0;
}


int namtGetCurrentCoincidenceData(NAMTDevice dev, unsigned char device, char *cdata) {
    int res;

    Connection con;
    
    con = DeviceTryConnection((Device)dev);
    if (!con) return DS_ERR_BUSY;

    ConnectionSilence(con, DS_DEBUG_DEBUG);
    
    res = ComplexRequestDataFixed(con, COMMAND_REQUEST_GET_CURRENT_COINCIDENCE, device, 0, cdata);
    if (res != COMMAND_RESPONSE_GET_CURRENT_COINCIDENCE) {
	namtProcessError(dev, con, COMMAND_REQUEST_GET_CURRENT_COINCIDENCE, res, DS_DEBUG_DEBUG);
	DeviceReturnConnection((Device)dev);
	return DS_ERR_INVALID_RESPONSE;
    }

    ConnectionSilence(con, DS_DEBUG_NOTHING);
    
    DeviceReturnConnection((Device)dev);
    
    return 0;
}


int namtDebug(NAMTDevice dev) {
    int err = 0;
    DSTime curtime;
    DSDuration duration;
    unsigned long ms;
    
    unsigned char data_size, detectors;
    NAMTData data, cdata;
    char *stcBuffer;
    
    unsigned long i, j, res;
    int pos, total_pos;
    
    unsigned short sensors;
    
    size_t idlen;
    
    NAMT_POINTER_DATA_DECLARE
    
    ds_assert_param(dev);

    if (dev->device.status != DEVICE_STATUS_OK) return 0;

    if ((!dev->device.debug_dsc)||(dsDatagramServerIsEmpty(dev->device.debug_dsc))) return 0;

    err = dsLockTryRead(dev->ilock);
    if (err) return DS_ERR_BUSY;
    
    curtime = dsTimeCreate(NULL);
    duration = dsDurationCreate(NULL);

    data_size = dev->data_size;
    
    if (dev->plain) {
	data = (NAMTData)dsMalloca(namtDetectorPackedDataSize(32, MAX_NAMT_SENSORS));
	cdata = NULL;
    }
    else {
	data = (NAMTData)dsMalloca(namtDetectorPackedDataSize(32, MAX_NAMT_DETECTORS));
	cdata = (NAMTData)dsMalloca(namtDetectorPackedCDataSize(32, MAX_NAMT_DETECTORS));
    }

    idlen = strlen(dev->device.id) + 1;
    stcBuffer = (char*)dsMalloca(3*sizeof(uint8_t) + 2*sizeof(uint16_t) + (MAX_NAMT_SENSORS + 4 + 9)*sizeof(uint32_t) + sizeof(double) + idlen);

    if ((!curtime)||(!duration)||(!data)||((!dev->plain)&&(!cdata))) {
	dsLockReturn(dev->ilock);
	if (curtime) dsTimeDestroy(curtime);
	if (duration) dsDurationDestroy(duration);
	if (data) dsFreea(data);
	if (cdata) dsFreea(cdata);
	if (stcBuffer) dsFreea(stcBuffer);
	return DS_ERR_MEM;
    }
    
    dsTimeCurrentTime(curtime);
    dsTimeCalcDuration(duration, dev->datatime, curtime);
    
    ms = dsDurationGetMilliseconds(duration);
    if (ms >= dev->rt_time) ms -= dev->rt_time;
    
    if ((ms>NAMT_CURRENT_DATA_MINIMUM_DURATION)&&(ms>(dev->rt_time/10)+NAMT_DATA_ACQUISITION_TIME)&&(ms<dev->rt_time-NAMT_CURRENT_DATA_ACQUISITION_TIME)) {
	dsDatagramServerBroadcastStart(dev->device.debug_dsc, NULL);
	for (i = 0; i < dev->devices; i++) {
	    err = namtGetCurrentData(dev, (unsigned char)i, (char*)data);
	    if ((cdata)&&(!err)) err = namtGetCurrentCoincidenceData(dev, (unsigned char)i, (char*)cdata);
	    if (err) break;
	    
	    if (cdata) {
		detectors = dev->detectors[i];
		sensors = 2 * detectors;
	    } else sensors = dev->detectors[i];

	    namtPointerDataInit(data);
    	    *(unsigned char*)stcBuffer=0xC1+(unsigned char)i;
	    pos = 1;			*(uint16_t*)(stcBuffer+pos) = dev->device.type | (dev->device_type<<8);
	    pos += sizeof(uint16_t); 	*(uint16_t*)(stcBuffer+pos) = dev->device.version;
	    pos += sizeof(uint16_t);	memcpy(stcBuffer+pos, dev->device.id, idlen);
	    pos += idlen;	        *(time_t*)(stcBuffer+pos) = time(NULL);
	    pos += sizeof(time_t);	*(uint32_t*)(stcBuffer+pos) = data->duration;
	    pos += sizeof(uint32_t);	*(uint32_t*)(stcBuffer+pos) = sensors;
	    pos += sizeof(uint32_t);	*(uint8_t*)(stcBuffer+pos) = DATA_TYPE_UINT32;
	    pos += sizeof(uint8_t);	*(double*)(stcBuffer+pos) = 1;
	    pos += sizeof(double);	*(uint16_t*)(stcBuffer+pos) = 0;
	    pos += sizeof(uint16_t);
 	    
	    for (j=0;j<sensors;j++,pos+=sizeof(uint32_t)) {
		*(uint32_t*)(stcBuffer+pos)=namtPointerDataNext(data_size);
	    }

	    if (cdata) {
		namtPointerDataInit(cdata);
		for (j=0, res=0;j<(unsigned int)detectors*(unsigned int)detectors;j++) 
		    res += namtPointerDataNext(data_size);

		*(uint32_t*)(stcBuffer+pos)=cdata->duration;pos+=sizeof(uint32_t);
	        *(uint32_t*)(stcBuffer+pos)=9;		pos += sizeof(uint32_t);	
		*(uint8_t*)(stcBuffer+pos)=DATA_TYPE_UINT32;pos += sizeof(uint8_t);

		total_pos = pos; pos+=sizeof(uint32_t);
		*(uint32_t*)(stcBuffer+pos) = res; pos+=sizeof(uint32_t);
		*(uint32_t*)(stcBuffer+pos) = namtPointerDataNext(data_size); pos+=sizeof(uint32_t);
		*(uint32_t*)(stcBuffer+pos) = namtPointerDataNext(data_size); pos+=sizeof(uint32_t);
		*(uint32_t*)(stcBuffer+pos) = namtPointerDataNext(data_size); pos+=sizeof(uint32_t);
		*(uint32_t*)(stcBuffer+pos) = namtPointerDataNext(data_size); pos+=sizeof(uint32_t);
		*(uint32_t*)(stcBuffer+pos) = namtPointerDataNext(data_size); pos+=sizeof(uint32_t);
		*(uint32_t*)(stcBuffer+pos) = namtPointerDataNext(data_size); pos+=sizeof(uint32_t);
		*(uint32_t*)(stcBuffer+pos) = namtPointerDataNext(data_size); pos+=sizeof(uint32_t);
		*(uint32_t*)(stcBuffer+total_pos) = namtPointerDataNext(data_size); 
	    }

	    dsDatagramServerBroadcastIteration(dev->device.debug_dsc, NULL, dev->device.id, stcBuffer, pos, 0, NULL);
	}

	if (err) 
	    dsDatagramServerBroadcastCancel(dev->device.debug_dsc, NULL);
	else 
	    dsDatagramServerBroadcastDone(dev->device.debug_dsc, NULL);
    }
    
    dsLockReturn(dev->ilock);
    
    dsTimeDestroy(curtime);
    dsDurationDestroy(duration);
    dsFreea(data);
    if (cdata) dsFreea(cdata);
    dsFreea(stcBuffer);

    return err;
}

char* namtControl(NAMTDevice dev, const char *request) {
    int err;
    int i, j, k, l, sd, resp;
    char *res = NULL;
    unsigned short int tmp, channel, max_size = 1024, pos;
    unsigned int uval, maxuval, mmaxuval, center;
    unsigned int maxnz, mmaxnz, minnz, mminnz;

    uint16_t size;
    NAMTSpectrum spectrum = NULL;
    uint16_t *spectrum16 = NULL;
    uint32_t *spectrum32 = NULL;
    
    xmlDocPtr doc;
    xmlNodePtr node, snode, vnode;
    char stmp[32] = "";
    
    unsigned char reply = 1;
    unsigned char need_connection = 0;
    unsigned char get_data = 0;

    char *answer = NULL;
    uint16_t answer_size;
    
    unsigned int sensors;
    

    Connection con;
    
    if ((!request)||(!*request)||(!strcasecmp(request,"help"))) {
	err = dsExchangeString(NULL, &answer, 
	"Command List\n"
	" T(R|W)[chan] [value] 	- Read/Write Thresholds (all channels or specified)\n"
	" V(R|W)[chan] [value]	- Read/Write HV (all channels or specified)\n"
	" DVR[chan]		- Get Cached HV (all channels or specified)\n"
	" R			- Reset LADC\n"
	" L			- Swtich On/Off Logarithmic Spectrum\n"
	" A[chan]		- Select Sensor for Spectrum Out\n"
	" C[value]		- Specify Spectrum Cursor\n"
	" :[command]		- Pass thru command to RS485 network\n"
	" @[command]		- Perform command but do not reply\n"
	" \n"
	" GetSpectrum[#]	- Get Spectrums (belonging to specified subdevice)\n"
	);
	if (err) return NULL;
	return answer;
    }

    if (dev->device.status != DEVICE_STATUS_OK) return 0;

    err = dsLockTryRead(dev->ilock);
    if (err) return NULL;
    
    if (dev->device.status != DEVICE_STATUS_OK) {
	dsLockReturn(dev->ilock);
	return NULL;
    }

    res = (char*)malloc(max_size);
    if (res) res[0] = 0;
    else {
	dsLockReturn(dev->ilock);
	return NULL;
    }


    if (request) {
	if (request[0]=='@') {
	    reply = 0;
	    request++;
	}

	if ((request[0]!='d')&&(request[0]!='D')) need_connection = 1;
	else if ((request[0]!='g')&&(request[0]!='G')) get_data = 1;
    }	
    
    if (need_connection) {
	con = DeviceTryConnection((Device)dev);
	if (!con) {
    	    dsLockReturn(dev->ilock);
	    free(res);
	    return NULL;
	}
    
        ConnectionSilence(con, DS_DEBUG_DEBUG);
    }

    for (sensors=0,i=0;i<dev->devices;i++)
	sensors += dev->detectors[i];
    if (!dev->plain) sensors *= 2;


    if ((request)&&((dev->device_type == NAMTDEVICE_TYPE_SNT)||(dev->device_type == NAMTDEVICE_TYPE_ANALYZER))) {
	if (request[0]==':') {
	    resp = ComplexRequest(con, COMMAND_REQUEST_CONTROL, strlen(request+1), request+1, &answer_size, &answer);
	    if (resp == COMMAND_RESPONSE_CONTROL) {
		if (answer_size < max_size) {
		    memcpy(res, answer, answer_size);
		    res[answer_size] = 0;
		} else {
		    sprintf(res, "ERROR. Response from MCU is too big");
		}
	    } else {
		namtProcessError((NAMTDevice)dev, con, COMMAND_RESPONSE_CONTROL, resp, DS_DEBUG_DEBUG);
		sprintf(res, "ERROR. Invalid response from MCU received");
	    }
	} else if ((request[0]=='T')||(request[0]=='t')) {
	    if ((request[1]=='R')||(request[1]=='r')) {
		channel = atoi(request+2);
		if (channel) {
		    if (channel<=sensors) {
			resp = RequestChannelAttribute(con, COMMAND_REQUEST_GET_THRESHOLD, channel-1, &tmp);
			if (resp == COMMAND_REQUEST_GET_THRESHOLD) {
			    sprintf(res, "Thr%hu = %hu", channel, tmp);
			} else {
			    namtProcessError((NAMTDevice)dev, con, COMMAND_REQUEST_GET_THRESHOLD, resp, DS_DEBUG_DEBUG);
			    sprintf(res, "ERROR. Invalid response from MCU received");
			}
		    } else {
			sprintf(res, "ERROR. Invalid channel: %u", channel);
		    }
		} else {
		    pos = 0;
		    for (channel=0;channel<sensors;channel++) {
			resp = RequestChannelAttribute(con, COMMAND_REQUEST_GET_THRESHOLD, channel, &tmp);
			if (resp == COMMAND_REQUEST_GET_THRESHOLD) {
			    sprintf(res+pos, "%sThr%hu = %hu", channel?", ":"", channel+1, tmp);
			} else {
			    namtProcessError((NAMTDevice)dev, con, COMMAND_REQUEST_GET_THRESHOLD, resp, DS_DEBUG_DEBUG);
			    sprintf(res+pos, "%sThr%hu = -1", channel?", ":"", channel+1);
			}
			pos += strlen(res+pos);
		    }
		}
	    } else if ((request[1]=='W')||(request[1]=='w')) {
		channel = atoi(request+2);
		
		if (channel) {
		    if (channel<=sensors) {
			for (i=2;((request[i])&&((request[i]<'0')||(request[i]>'9')));i++);
			for (;((request[i]>='0')&&(request[i]<='9'));i++);
			for (;((request[i])&&((request[i]<'0')||(request[i]>'9')));i++);
			tmp = atoi(request+i);

			resp = SendChannelAttribute(con, COMMAND_REQUEST_SET_THRESHOLD, channel-1, tmp);
			if (resp == COMMAND_REQUEST_SET_THRESHOLD) {
			    sprintf(res, "Done (Thr%hu = %hu)", channel, tmp);
			} else {
			    namtProcessError((NAMTDevice)dev, con, COMMAND_REQUEST_SET_THRESHOLD, resp, DS_DEBUG_DEBUG);
			    sprintf(res, "ERROR. Invalid response from MCU received");
			}
		    } else {
			sprintf(res, "ERROR. Invalid channel: %hu", channel);
		    }
		} else {
		    pos = 0;
		    for (i = 2;((request[i])&&((request[i]<'0')||(request[i]>'9')));i++);
		    for (channel=0;channel<sensors;channel++) {
			for (;((request[i]>='0')&&(request[i]<='9'));i++);
			for (;((request[i])&&((request[i]<'0')||(request[i]>'9')));i++);
			tmp = atoi(request+i);

			resp = SendChannelAttribute(con, COMMAND_REQUEST_SET_THRESHOLD, channel-1, tmp);
			if (resp != COMMAND_REQUEST_SET_THRESHOLD) {
			    namtProcessError((NAMTDevice)dev, con, COMMAND_REQUEST_SET_THRESHOLD, resp, DS_DEBUG_DEBUG);
			    pos++;
			}
		    }
		    if (pos)
			sprintf(res, "Failed to set %hu thresholds", pos);
		    else
			sprintf(res, "Done");
		}
	    } else {
		sprintf(res, "ERROR. Invalid threshold request");
	    }
	} else if (!strncasecmp(request, "dvr", 3)) {
	    channel = atoi(request+3);
	    if (channel) {
		if (channel <= sensors) {
		    tmp = dev->hv[channel-1];
		    if (tmp == (unsigned short)-1) sprintf(res, "e%hu = -1", channel);
		    else sprintf(res, "e%hu = %hu", channel, tmp);
		} else sprintf(res, "ERROR. Invalid channel: %u", channel);
	    } else {
		pos = 0;
		for (channel=0;channel<sensors;channel++) {
		    tmp = dev->hv[channel];
		    if (tmp == (unsigned short)-1) sprintf(res+pos, "%se%hu = -1", channel?", ":"", channel+1);
		    else sprintf(res+pos, "%se%hu = %hu", channel?", ":"", channel+1, tmp);
		    pos += strlen(res+pos);
		}
	    }
	} else if (((request[0]=='H')||(request[0]=='h'))||(((request[0]=='V')||(request[0]=='v'))&&((request[1]=='R')||(request[1]=='r')))) {
		channel = atoi(((request[0]=='H')||(request[0]=='h'))?(request+1):(request+2));
		if (channel) {
		    if (channel<=sensors) {
			resp = RequestChannelAttribute(con, COMMAND_REQUEST_GET_HV, channel-1, &tmp);
			if (resp == COMMAND_REQUEST_GET_HV) {
			    if (tmp > 0) dev->hv[channel-1] = tmp;
			    sprintf(res, "e%hu = %hu", channel, tmp);
			} else {
			    namtProcessError((NAMTDevice)dev, con, COMMAND_REQUEST_GET_HV, resp, DS_DEBUG_DEBUG);
			    sprintf(res, "ERROR. Invalid response from MCU received");
			}
		    } else {
			sprintf(res, "ERROR. Invalid channel: %u", channel);
		    }
		} else {
		    pos = 0;
		    for (channel=0;channel<sensors;channel++) {
			resp = RequestChannelAttribute(con, COMMAND_REQUEST_GET_HV, channel, &tmp);
			if (resp == COMMAND_REQUEST_GET_HV) {
			    if (tmp > 0) dev->hv[channel] = tmp;
			    sprintf(res+pos, "%se%hu = %hu", channel?", ":"", channel+1, tmp);
			} else {
			    namtProcessError((NAMTDevice)dev, con, COMMAND_REQUEST_GET_HV, resp, DS_DEBUG_DEBUG);
			    sprintf(res+pos, "%se%hu = -1", channel?", ":"", channel+1);
			}
			pos += strlen(res+pos);
		    }
		}
	} else if ((request[0]=='V')||(request[0]=='v')) {
		i = ((request[0]=='W')||(request[0]=='w'))?2:1;
		channel = atoi(request+i);
		
		if (channel) {
		    if (channel<=sensors) {
			for (;((request[i])&&((request[i]<'0')||(request[i]>'9')));i++);
			for (;((request[i]>='0')&&(request[i]<='9'));i++);
			for (;((request[i])&&((request[i]<'0')||(request[i]>'9')));i++);
			tmp = atoi(request+i);

			resp = SendChannelAttribute(con, COMMAND_REQUEST_SET_HV, channel-1, tmp);
			if (resp == COMMAND_REQUEST_SET_HV) {
			    sprintf(res, "Done (e%hu = %hu)", channel, tmp);
			} else {
			    namtProcessError((NAMTDevice)dev, con, COMMAND_REQUEST_SET_HV, resp, DS_DEBUG_DEBUG);
			    sprintf(res, "ERROR. Invalid response from MCU received");
			}
		    } else {
			sprintf(res, "ERROR. Invalid channel: %hu", channel);
		    }
		} else {
		    pos = 0;
		    for (;((request[i])&&((request[i]<'0')||(request[i]>'9')));i++);
		    for (channel=0;channel<sensors;channel++) {
			for (;((request[i]>='0')&&(request[i]<='9'));i++);
			for (;((request[i])&&((request[i]<'0')||(request[i]>'9')));i++);
			tmp = atoi(request+i);

			resp = SendChannelAttribute(con, COMMAND_REQUEST_SET_HV, channel-1, tmp);
			if (resp != COMMAND_REQUEST_SET_HV) {
			    namtProcessError((NAMTDevice)dev, con, COMMAND_REQUEST_SET_HV, resp, DS_DEBUG_DEBUG);
			    pos++;
			}
		    }
		    if (pos)
			sprintf(res, "Failed to set %hu HVs", pos);
		    else
			sprintf(res, "Done");
		}
	} else if ((request[0]=='R')||(request[0]=='r')) {
	    resp = SendSimple(con, COMMAND_REQUEST_RESET_FPGA);
	    if (resp == COMMAND_REQUEST_RESET_FPGA) {
		sprintf(res, "Done");
	    } else {
		namtProcessError((NAMTDevice)dev, con, COMMAND_REQUEST_RESET_FPGA, resp, DS_DEBUG_DEBUG);
		sprintf(res, "ERROR. Invalid response from MCU received");
	    }
	} else if ((request[0]=='G')||(request[0]=='g')) {
	    if (!strncasecmp(request, "GetSpectrum",11)) {
		channel = atoi(request + 11);
		if (channel > dev->devices) {
		    sprintf(res, "ERROR. Invalid sub-device specified");
		} else {
		    doc = xmlNewDoc("1.0");
		    if (doc) node = xmlNewChild((xmlNodePtr)doc, NULL, "Spectrums", NULL);
		    else node = NULL;
		    
		    if (node) {
			dsXmlSetPropString(node, "device", dev->device.id);
	    		for (sd = 0, i = 0; i < dev->devices; sd+=sensors, i++) {
			    if (dev->plain) sensors = dev->detectors[i];
			    else sensors = 2*dev->detectors[i];
			    
			    if ((!channel)||((i+1)==channel)) {
				resp = ComplexRequestData(con, COMMAND_REQUEST_GET_SPECTRUM, i, &size, (char**)&spectrum);
				if (resp != COMMAND_RESPONSE_GET_SPECTRUM) {
				    namtProcessError((NAMTDevice)dev, con, COMMAND_REQUEST_GET_SPECTRUM, resp, DS_DEBUG_DEBUG);
				    sprintf(res, "ERROR. Invalid response from MCU received");
				    break;
				} else if (size != namtSpectrumSize(dev, i)) {
				    free(spectrum);
				    sprintf(res, "ERROR. Invalid response from MCU received");
				    ConnectionSkipAll(con);
				    break;
				}
				
				if (dev->data_size == 32) {
				    spectrum32 = (uint32_t*)spectrum;
				} else if (dev->data_size == 16) {
				    spectrum16 = (uint16_t*)spectrum;
				}

				center = 0; mmaxuval = 0; mmaxnz = 0; mminnz = 128;
				for (j = 0, k= 0; j < sensors; j++) {
				    snode = xmlNewChild(node, NULL, "Graph", NULL);
				    if (snode) {
					dsXmlSetPropInt(snode, "sid", sd + j + 1);
					dsXmlSetPropInt(snode, "items", 127);
					
					maxuval = 0; maxnz = 0; minnz = 0;
					for (l = 1; l < 128; l++, k++) {
					    if (spectrum32) uval = spectrum32[k];
					    else if (spectrum16) uval = spectrum16[k];
					    else uval = 0;

//					    uval = j * 4 + rand() / (RAND_MAX/63);

					    
					    if (uval) {
						if (!minnz) minnz = l;
						maxnz = l;
						if (uval > maxuval) {
						    center = l;
						    maxuval = uval;
						}
					    }
					    

					    sprintf(stmp, "%lu", uval);
					    vnode = xmlNewChild(snode, NULL, "Value", stmp);
					    if (vnode) {
						dsXmlSetPropInt(vnode, "id", l);
					    }
					}
					dsXmlSetPropInt(snode, "max", maxuval);
					dsXmlSetPropInt(snode, "center", center);
					dsXmlSetPropInt(snode, "min_nonzero", minnz);
					dsXmlSetPropInt(snode, "max_nonzero", maxnz);
					if (maxuval > mmaxuval) mmaxuval = maxuval;
					if (maxnz > mmaxnz) mmaxnz = maxnz;
					if (minnz < mminnz) mminnz = minnz;
				    }
				}
				
				free(spectrum);
			    }
			}
			dsXmlSetPropInt(node, "max", mmaxuval);
			dsXmlSetPropInt(node, "min_nonzero", mminnz);
			dsXmlSetPropInt(node, "max_nonzero", mmaxnz);
		    } else sprintf(res, "ERROR. Internal XML Failure");
		    
		    if (doc) {
			if (!*res) {
			    free(res);
			    xmlDocDumpMemory(doc, (xmlChar**)&res, &i);
			}
			xmlFreeDoc(doc);
		    } else sprintf(res, "ERROR. Internal XML Failure");
		} 
	    } else {
		sprintf(res, "ERROR. Invalid request");
	    }
	} else if ((request[0]=='Z')||(request[0]=='z')) {
	    resp = SendSimple(con, COMMAND_REQUEST_SPECTRUM_RESET);
	    if (resp == COMMAND_RESPONSE_SPECTRUM_RESET) {
		sprintf(res, "Done");
	    } else {
		namtProcessError((NAMTDevice)dev, con, COMMAND_REQUEST_SPECTRUM_RESET, resp, DS_DEBUG_DEBUG);
		sprintf(res, "ERROR. Invalid response from MCU received");
	    }
	} else if ((request[0]=='L')||(request[0]=='l')) {
	    if (dev->spectrum.flags&SPECTRUM_LOGARITHM_FLAG)
		dev->spectrum.flags&=~SPECTRUM_LOGARITHM_FLAG;
	    else 
		dev->spectrum.flags|=SPECTRUM_LOGARITHM_FLAG;
	    
	    resp = SendData(con, COMMAND_REQUEST_SPECTRUM_CONFIGURE_OUTPUT, sizeof(SpectrumPropertiesS), (char*)&dev->spectrum);
	    if (resp == COMMAND_RESPONSE_SPECTRUM_CONFIGURE_OUTPUT) {
		sprintf(res, "Spectrum is configured. Sensor %u, Cursor %u %s", dev->spectrum.sensor + 1, dev->spectrum.cursor, (dev->spectrum.flags&SPECTRUM_LOGARITHM_FLAG)?" Logarithmic":"");
	    } else {
		namtProcessError((NAMTDevice)dev, con, COMMAND_REQUEST_SPECTRUM_CONFIGURE_OUTPUT, resp, DS_DEBUG_DEBUG);
		sprintf(res, "ERROR. Invalid response from MCU received");
	    }
	} else if ((request[0]=='A')||(request[0]=='a')) {
	    tmp = atoi(request+1);
	    if ((tmp>sensors)||(tmp<=0)) {
		sprintf(res, "ERROR. Invalid sensor specified");
	    } else {
	        dev->spectrum.sensor = tmp - 1;
	    
		resp = SendData(con, COMMAND_REQUEST_SPECTRUM_CONFIGURE_OUTPUT, sizeof(SpectrumPropertiesS), (char*)&dev->spectrum);
		if (resp == COMMAND_RESPONSE_SPECTRUM_CONFIGURE_OUTPUT) {
		    sprintf(res, "Spectrum is configured. Sensor %u, Cursor %u %s", dev->spectrum.sensor + 1, dev->spectrum.cursor, (dev->spectrum.flags&SPECTRUM_LOGARITHM_FLAG)?" Logarithmic":"");
		} else {
		    namtProcessError((NAMTDevice)dev, con, COMMAND_REQUEST_SPECTRUM_CONFIGURE_OUTPUT, resp, DS_DEBUG_DEBUG);
		    sprintf(res, "ERROR. Invalid response from MCU received");
		}
	    }
	} else if ((request[0]=='C')||(request[0]=='c')) {
	    tmp = atoi(request+1);
	    
	    if (tmp>127) tmp = 127;
	    dev->spectrum.cursor = tmp;
	    
	    resp = SendData(con, COMMAND_REQUEST_SPECTRUM_CONFIGURE_OUTPUT, sizeof(SpectrumPropertiesS), (char*)&dev->spectrum);
	    if (resp == COMMAND_RESPONSE_SPECTRUM_CONFIGURE_OUTPUT) {
		sprintf(res, "Spectrum is configured. Sensor %u, Cursor %u %s", dev->spectrum.sensor + 1, dev->spectrum.cursor, (dev->spectrum.flags&SPECTRUM_LOGARITHM_FLAG)?" Logarithmic":"");
	    } else {
		namtProcessError((NAMTDevice)dev, con, COMMAND_REQUEST_SPECTRUM_CONFIGURE_OUTPUT, resp, DS_DEBUG_ERROR);
		sprintf(res, "ERROR. Invalid response from MCU received");
	    }
	} else {
	    sprintf(res, "ERROR. Invalid request");
	}
    }

    if (need_connection) {
	ConnectionSilence(con, DS_DEBUG_NOTHING);
	
	DeviceReturnConnection((Device)dev);
    }
    dsLockReturn(dev->ilock);

    if (reply) return res;

    free(res);
    return NULL;
}

int namtSetConfig(NAMTDevice dev, xmlDocPtr config) {
    int err, resp;
    int i, j;
    const char *stmp;
    const char *device;
    unsigned short int subdev, layer, sensor, channel, fchan, utmp;

    xmlNodeSetPtr nodes;
    xmlNodePtr node, rnode;

    Connection con;
    
    if (dev->device.status != DEVICE_STATUS_OK) return DS_ERR_NOTINITIALIZED;

    rnode = xmlDocGetRootElement(config);
    if (!rnode) return DS_ERR_INVALID_ARGUMENT;

    err = dsLockTryRead(dev->ilock);
    if (err) return err;
    
    if (dev->device.status != DEVICE_STATUS_OK) {
	dsLockReturn(dev->ilock);
	return err;
    }

    con = DeviceGetConnection((Device)dev);
    if (!con) {
    	dsLockReturn(dev->ilock);
	return DS_ERR_CONNECTION;
    }

    nodes = dsXmlFindSubNodes(rnode, "Voltages", "Voltage");
    if (nodes) {
	for (i=0;i<nodes->nodeNr;i++) {
	    node = nodes->nodeTab[i];

	    device = dsXmlGetPropString(node, "device", NULL);
	    if (!device) continue;

	    stmp = strrchr(device, '-');
	    if (stmp) subdev = atoi(stmp+1) - 1;
	    else subdev = 0;

	    if (subdev>=dev->devices) continue;
	    for (fchan=0,j=0;j<subdev;j++) fchan += 2*dev->detectors[j];
	    
	    layer = dsXmlGetPropInt(node, "layer", 0) - 1;
	    sensor = dsXmlGetPropInt(node, "sensor", 0) - 1;
	    if (sensor>=dev->detectors[subdev]) continue;
	    if (layer>1) continue;
	    
	    channel = fchan + layer * dev->detectors[subdev] + sensor;
	    
	    utmp = dsXmlGetNodeInt(node, NULL, 0);
	    if ((utmp<900)||(utmp>3000)) continue;
	    
	    resp = SendChannelAttribute(con, COMMAND_REQUEST_SET_HV, channel, utmp);
	    if (resp != COMMAND_REQUEST_SET_HV) namtError((NAMTDevice)dev, con, COMMAND_REQUEST_SET_HV, resp);
	}
	dsXmlFreeSubNodes(nodes);
    }

    nodes = dsXmlFindSubNodes(rnode, "Thresholds", "Threshold");
    if (nodes) {
	for (i=0;i<nodes->nodeNr;i++) {
	    node = nodes->nodeTab[i];

	    device = dsXmlGetPropString(node, "device", NULL);
	    if (!device) continue;

	    stmp = strrchr(device, '-');
	    if (stmp) subdev = atoi(stmp+1) - 1;
	    else subdev = 0;

	    if (subdev>=dev->devices) continue;
	    for (fchan=0,j=0;j<subdev;j++) fchan += 2*dev->detectors[j];
	    
	    layer = dsXmlGetPropInt(node, "layer", 0) - 1;
	    sensor = dsXmlGetPropInt(node, "sensor", 0) - 1;
	    if (sensor>=dev->detectors[subdev]) continue;
	    if (layer>1) continue;
	    
	    channel = fchan + layer * dev->detectors[subdev] + sensor;
	    
	    utmp = (int)(100*dsXmlGetNodeDouble(node, NULL, 0));
	    if ((utmp<1)||(utmp>255)) continue;

	    resp = SendChannelAttribute(con, COMMAND_REQUEST_SET_THRESHOLD, channel, utmp);
	    if (resp != COMMAND_REQUEST_SET_THRESHOLD) namtError((NAMTDevice)dev, con, COMMAND_REQUEST_SET_THRESHOLD, resp);
	}

	dsXmlFreeSubNodes(nodes);
    }

    DeviceReturnConnection((Device)dev);
    dsLockReturn(dev->ilock);
        
    return 0;
}

int namtGetDriverInfo(NAMTDevice dev, xmlNodePtr node) {
    char stmp[32];

    ds_assert_param(dev);
    ds_assert_param(node);

    sprintf(stmp, "%u", dev->protocol_version);
    xmlNewChild(node, NULL, "ProtocolVersion", stmp);
    
    switch (dev->device_type) {
	case NAMTDEVICE_TYPE_NAMT:
	    xmlNewChild(node, NULL, "MicrocodeType", "NAMT");
	break;
	case NAMTDEVICE_TYPE_SNT:
	    xmlNewChild(node, NULL, "MicrocodeType", "SNT");
	break;
	case NAMTDEVICE_TYPE_ANALYZER:
	    xmlNewChild(node, NULL, "MicrocodeType", "SA");
	break;
	default:
	    ;
    }

    sprintf(stmp, "%u", dev->device_version);
    xmlNewChild(node, NULL, "MicrocodeVersion", stmp);

    sprintf(stmp, "%lX", dev->handled_events);
    xmlNewChild(node, NULL, "HandledEventTypes", stmp);

    return 0;
}
