#include <stdio.h>
#include <string.h>

#include <dsdebug.h>
#include <dsassert.h>
#include <dserror.h>

#include "error.h"
#include "defaults.h"
#include "lowlevel.h"
#include "final.h"

#include "nanm.h"
#include "nanmbuffer.h"



static int nanmReader(NANMDevice dev);
static int nanmProcessor(NANMDevice dev, NANMReaderBuf buf);
static void nanmDeviceDestroy(NANMDevice dev);
static void nanmReset(NANMDevice dev);


void nanmDeviceSetDefaults(NANMDevice dev) {
    DeviceSetDefaults((Device)dev);

    dev->dw = NULL;
//    dev->triggerdw = NULL;
}

Device nanmDeviceCreate(DSConfigContext config, DSConfigReference ref) {
    NANMDevice dev;
    
    dev = (NANMDevice)malloc(sizeof(NANMDeviceS));
    if (!dev) return (Device)dev;

    nanmDeviceSetDefaults(dev);

    dev->datatime = dsTimeCreate(NULL);
    if (!dev->datatime) {
	free(dev);
	return NULL;
    }

    dev->device.version = NANM_VERSION;
    
//    dev->device.configure = nanmConfigure;
    dev->device.destroy = (DeviceDestroyFunction)nanmDeviceDestroy;
    dev->device.reader = (DeviceReaderFunction)nanmReader;
    dev->device.processor = (DeviceProcessorFunction)nanmProcessor;
    dev->device.reset = (DeviceResetFunction)nanmReset;
    
    dev->hosts = 0;
    
    return (Device)dev;
}

static void nanmDeviceDestroy(NANMDevice dev) {
    ds_assert_return(dev);
    
    if (dev->dw) DeviceFreeWriters((Device)dev, dev->dw);
//    if (dev->triggerdw) DeviceFreeWriters((Device)dev, dev->triggerdw);
    dsTimeDestroy(((NANMDevice)dev)->datatime);
    free(dev);
}

#define CRC16(crc, data) \
        crc  = (unsigned char)(crc >> 8) | (crc << 8); \
        crc ^= data; \
        crc ^= (unsigned char)(crc & 0xff) >> 4; \
        crc ^= (crc << 8) << 4; \
        crc ^= ((crc & 0xff) << 4) << 1;

static int nanmRequest(NANMDevice dev, Connection con, NANMId id, NANMCommand cmd, unsigned size, unsigned char *data, unsigned int rsize, unsigned char *rdata, nanmTriggerData td) {
    int err;
    int nak_retries, retries;
    unsigned int i;
    uint16_t crc = 0, rcrc = 0;
    unsigned char header[6], rheader[7];
    unsigned int triggers;
    unsigned char *rbuf = NULL;
    unsigned char rbufsize;
    unsigned char hid;
    uint16_t trigger_time;

    retries = con->retries;
    nak_retries = con->retries;

    header[0] = NANM_BYTE_SD;
    header[1] = NANM_BYTE_SOH;
    header[2] = id;
    header[3] = NANM_MASTER_ID;
    header[4] = data?size+1:1; 		/* command + data */
    header[5] = cmd;

retry:
    crc = 0; rcrc = 0;
    
    if ((retries == con->retries)&&(nak_retries == con->retries))
	dsDebugServer(DS_DEBUG_DEBUG, "NANM(%s). Request: %02X %02X %02X %02X %02X %02X", dev->device.id, header[0], header[1], header[2], header[3], header[4], header[5]);
    else 
	dsDebugServer(DS_DEBUG_INFO, "NANM(%s). Request(retry: %u): %02X %02X %02X %02X %02X %02X", dev->device.id, 2*con->retries - retries - nak_retries + 1, header[0], header[1], header[2], header[3], header[4], header[5]);
    
	/* DS. Simple command */
    if (/*cmd == NANM_COMMAND_REQUEST*/0) {
	err = ConnectionSlowWrite(con, &id, 1);
    } else {
	for (i=2;i<6;i++) {
	    CRC16(crc, header[i]);
	}
    
	err = ConnectionSlowWrite(con, header, 6);
	if (err) return err;
    
	if (size) {
	    ds_assert_param(data);
	
	    for (i=0;i<size;i++) {
		CRC16(crc, data[i]);
	    }
	
	    err = ConnectionSlowWrite(con, data, size);
	    if (err) return err;
	}
	
	err = ConnectionSlowWrite(con, &crc, 2);
	if (err) return err;
    
	switch (cmd) {
	    case NANM_COMMAND_RESET:
		return;
	}
    }

    err = ConnectionRead(con, rheader, 7);
    if (err) {
        if ((err == DS_ERR_TIMEOUT)&&((cmd == NANM_COMMAND_SET_TIMENODE)||(cmd == NANM_COMMAND_RESET))) return 0;
    }

    if (!err) {
	dsDebugServer(DS_DEBUG_DEBUG, "NANM(%s). Response: %02X %02X %02X %02X %02X %02X %02X", dev->device.id, rheader[0], rheader[1], rheader[2], rheader[3], rheader[4], rheader[5], rheader[6]);
	if ((rheader[0] != NANM_BYTE_SD)||(rheader[1] != NANM_BYTE_ACK)||(rheader[2] != NANM_MASTER_ID)||(rheader[3] != id)||(rheader[4] == 0)||(rheader[5] != cmd)) {
	    if ((rheader[0] == NANM_BYTE_SD)&&(rheader[1] == NANM_BYTE_NAK)&&(rheader[2] == NANM_MASTER_ID)&&(rheader[3] == id)) {
		ConnectionRead(con, (char*)&err, 1);
		if (--nak_retries) goto retry;
	    }
	    err = ADAS_ERR_BOARD;
	} else if (cmd == NANM_COMMAND_REQUEST) {
	    for (i=0;i<dev->hosts;i++)
	        if (dev->host[i] == id) break;
		
	    if (i<dev->hosts) {
		hid = i;

		rbufsize = rheader[4] - rsize - 2;
		triggers = rbufsize/3;

		if (hid) {
		    if (td->triggers != triggers) {
			if (td->triggers == (triggers - 1)) {
			    dsDebugServer(DS_DEBUG_WARNING, "NANM(%s). Trigger number mistmatch. ID%lu: %lu, ID%lu: %lu", dev->device.id, dev->host[0], td->triggers, id, triggers);
			    triggers--;
			} else if (td->triggers == (triggers + 1)) {
			    dsDebugServer(DS_DEBUG_WARNING, "NANM(%s). Trigger number mistmatch. ID%lu: %lu, ID%lu: %lu", dev->device.id, dev->host[0], td->triggers, id, triggers);
			    td->triggers--;
			} else {
			    dsDebugServer(DS_DEBUG_ERROR, "NANM(%s). Trigger number mistmatch. ID%lu: %lu, ID%lu: %lu", dev->device.id, dev->host[0], td->triggers, id, triggers);
			    err = ADAS_ERR_BOARD;
			}
		    }
		} else {
		    err = nanmBufferTriggerInit(td, triggers);
		}
	    } else err = ADAS_ERR_BOARD;
	 } else if (rheader[4] != rsize + 2) err = ADAS_ERR_BOARD;
    }

    if (!err) {
	for (i=2;i<7;i++) {
	    CRC16(rcrc, rheader[i]);
	}

	if (rsize) {
	    ds_assert_param(rdata);
	    err = ConnectionRead(con, rdata, rsize);
	    
	    if (cmd == NANM_COMMAND_REQUEST) {
		rbuf = (unsigned char*)dsMalloca(rbufsize*sizeof(unsigned char));
		if (rbuf) err = ConnectionRead(con, rbuf, rbufsize);
		else err = DS_ERR_MEM;
	    }
	    
	    if (!err) {
		for (i=0;i<rsize;i++) {
		    CRC16(rcrc, rdata[i]);
		}
		if (cmd == NANM_COMMAND_REQUEST) {
		    for (i=0;i<rbufsize;i++) {
			CRC16(rcrc, rbuf[i]);
		    }
		}
		err = ConnectionRead(con, (char*)&crc, 2);
	    }

	} else {
	    err = ConnectionRead(con, (char*)&crc, 2);
	}
    }
    
    if ((err)||(rcrc != crc)) {
	if (--retries) {    
	    dsWait(con->retry_timeout);
	    ConnectionReadAll(con);
	    goto retry;
	} else if (!err) err = ADAS_ERR_CRC;
    } else if (cmd == NANM_COMMAND_REQUEST) {
	for (i=0;i<triggers;i++) {
	    td->trigger_count[i][hid] = rbuf[3*i+2];
	    if (!hid) {
		trigger_time = rbuf[3*i] + ((unsigned int)(rbuf[3*i+1])<<8);
		td->trigger_time[i] = trigger_time;
	    }
	    dsDebugServer(DS_DEBUG_DEBUG, "NANM(%s). Triggers ::: %i(%i): %2lx %2lx: %2lx", dev->device.id, i, id, rbuf[3*i+1], rbuf[3*i], rbuf[3*i+2]);
	}
    }

    if ((cmd == NANM_COMMAND_REQUEST)&&(rbuf)) dsFreea(rbuf);
    
    return err;
}

static int nanmWaitBroadcast(NANMDevice dev, Connection con) {
    int err;
    unsigned char header[8];

    err = ConnectionRead(con, header, 8);
    if (err) return err;

    dsDebugServer(DS_DEBUG_DEBUG, "NANM(%s). Broadcast Response: %02X %02X %02X %02X %02X %02X", dev->device.id, header[0], header[1], header[2], header[3], header[4], header[5]);
    if ((header[0] != NANM_BYTE_SD)||(header[1] != NANM_BYTE_SOH)||(header[2] != NANM_BROADCAST_ID)||(header[3] != dev->host[0])||(header[4] != 1)||(header[5] != NANM_COMMAND_BROADCAST)) return ADAS_ERR_BOARD;
    return 0;    
}

static unsigned char bc_header[8] = { NANM_BYTE_SD, NANM_BYTE_SOH, NANM_BROADCAST_ID, NANM_MASTER_ID, 1, NANM_COMMAND_BROADCAST, 0xAB, 0x2A };
static char simple_broadcast = NANM_BYTE_SIMPLE_BROADCAST;

static int nanmSendBroadcast(NANMDevice dev, Connection con) {
    int i;
    
    dsDebugServer(DS_DEBUG_DEBUG, "NANM(%s). BROADCAST", dev->device.id);

	/* DS. Simple command */
//    return ConnectionSlowWrite(con, &simple_broadcast, 1);
    return ConnectionSlowWrite(con, bc_header, 8);
}

static int nanmInitialize(NANMDevice dev, Connection con) {
    int err;
    int i;
    unsigned int hosts = 0;
    DSDuration delay;
    
    delay = dsDurationCreate(NULL);
    if (!delay) return DS_ERR_MEM;
    
    dsDurationImportMilliseconds(delay, NANM_RESET_DELAY);
    
    
    dsDebugServer(DS_DEBUG_INFO, "NANM(%s). Initializing hosts", dev->device.id);
    for (i = 1; i<(dev->max_host+1);i++) {
	if (!DeviceContinueLoop((Device)dev)) break;
	nanmRequest(dev, con, i, NANM_COMMAND_RESET, 0, NULL, 0, NULL, NULL);
    }
    
    dsWaitRT(delay);
    
    for (i = 1; i<(dev->max_host+1);i++) {
	if (!DeviceContinueLoop((Device)dev)) break;
	err = nanmRequest(dev, con, i, NANM_COMMAND_INITIALIZE, 0, NULL, 0, NULL, NULL);
	if (err) dsDebugServer(DS_DEBUG_DEBUG, "NANM(%s). No device detected at %02Xh (Error: %s)", dev->device.id, i, dsError(err, NULL));
	else {
	    dsDebugServer(DS_DEBUG_INFO, "NANM(%s). Device detected at %02Xh", dev->device.id, i);
	    dev->host[hosts++] = i;
	}
    }
    dsDebugServer(DS_DEBUG_WARNING, "NANM(%s). Initialization completed. %u hosts found.", dev->device.id, hosts);
    if (!hosts) return DS_ERR_NOTFOUND;

    dev->hosts = hosts;
    
    dsDurationDestroy(delay);

    return 0;
}

static int CalculateStartTime(NANMDevice dev, DSDuration readout, DSTime start_time) {
    DSDuration rts_time;
    
    ds_assert_param(dev);
    ds_assert_param(readout);
    ds_assert_param(start_time);
    
    dsTimeCurrentTime(start_time);
    dsTimeAddMilliseconds(start_time, 100);
    
    rts_time = dsDurationCreate(NULL);
    if (!rts_time) return DS_ERR_MEM;

    dsDurationCeilSeconds(readout, 1);
    dsTimeCeil(start_time, readout);
    
    return 0;
}

static int nanmReader(NANMDevice dev) {
	int err = 0;
	int i;
	unsigned char doread = 0;
	unsigned char informated = 0;
	uint16_t value;
    
	Connection con;

	DSDuration slp;
	DSDuration readout;
	DSDuration brd_delay;
	
	DSTime waittime;
	
	xmlNodePtr cfgnode, chipcfg;
	NANMReaderBuf buf;

	slp = dsDurationCreate(NULL);
	readout = dsDurationCreate(NULL);
	brd_delay = dsDurationCreate(NULL);
	waittime = dsTimeCreate(NULL);
	
	if ((!slp)||(!readout)||(!waittime)||(!brd_delay)) {
	    if (slp) dsDurationDestroy(slp);
	    if (readout) dsDurationDestroy(readout);
	    if (brd_delay) dsDurationDestroy(brd_delay);
	    if (waittime) dsTimeDestroy(waittime);
	    
	    Abort();
	    return DS_ERR_MEM;
	}    

	dsDurationImportMicroseconds(slp, NANM_TIMEOUT_REINIT);
	dsDurationImportMilliseconds(readout, NANM_DEFAULT_READOUT_DURATION);
	dsDurationImportMilliseconds(brd_delay, 3);
	

	while (DeviceContinueLoop((Device)dev)) {
	    dev->device.status = DEVICE_STATUS_NOT_CONNECTED;
	    err = 0;
	    doread = 0;
	    dev->reset = 0;
	    dev->hosts = 0;

	    dsDebugServer(DS_DEBUG_INFO, "NANM(%s). Reininitializing!", dev->device.id);
	    
	    con = DeviceInitializeConnection((Device)dev);
	    if (!con) {
		if (!informated) {
		    dsDebugServer(DS_DEBUG_CRITICAL, "NANM(%s). Error opening communication device!", dev->device.id);
		    informated = 1;
		}
		DeviceCloseConnection((Device)dev);
		dsWait(slp);
		continue;
	    } else informated = 0;

	    cfgnode = dsConfigGetConfigByReference(dev->device.cfg, dev->device.cfgref);
	    chipcfg = dsXmlFindSubNode(cfgnode, "Configuration");
	    dsXmlGetNodeDuration(chipcfg, "ReadoutInterval", readout);
	    dev->max_host = dsXmlGetNodeInt(chipcfg, "MaxID", NANM_LAST_SLAVE_ID);
	    dsConfigReturnConfig(dev->device.cfg, cfgnode);

	    ConnectionReadAll(con);

	    err = nanmInitialize(dev, con);
	    if (err) {
		DeviceCloseConnection((Device)dev);
		dsWait(slp);
		continue;
	    }
	    
	    dev->device.status = DEVICE_STATUS_OK;

	    dev->rt_time = dsDurationGetMilliseconds(readout);

	    CalculateStartTime(dev, readout, dev->datatime);


	    dev->device.sf = 1;
	    while ((DeviceContinueLoop((Device)dev))&&(!err)&&(!dev->reset)) {
		if (doread) {
		    buf = nanmBufferCreate(dev, dev->hosts, dev->host);
		    if (!buf) {
			err = DS_ERR_MEM;
			break;
		    }
		    nanmBufferSet(buf, dev->rt_time);
		}
		
		
		dsWaitRTUntil(dev->datatime);
		
		err = nanmSendBroadcast(dev, con);
		if (err) {
		    dsDebugServer(DS_DEBUG_ERROR, "NANM(%s). Time synchronization failed (%s).", dev->device.id, dsError(err, NULL));
		    break;
		}
		dsDebugServer(DS_DEBUG_DEBUG, "NANM(%s). New iteration", dev->device.id);
		
		dsWait(slp);

		if (doread) {
		    for (i=0;i<dev->hosts;i++) {
			err = nanmRequest(dev, con, dev->host[i], NANM_COMMAND_REQUEST, 0, NULL, 2, (unsigned char*)&value, &buf->trigger);
			if (err) {
			    dsDebugServer(DS_DEBUG_ERROR, "NANM(%s). Data Request failed (%s).", dev->device.id, dsError(err, NULL));
			    break;
			}
			nanmBufferAdd(buf, i, value);
		    }
		    if (!err) err = nanmBufferRegister(dev, buf);
		    if (err) {
			nanmBufferDestroy(buf);
			//break;
		    }
		} else doread = 1;
		

		dsTimeAddMilliseconds(dev->datatime, dev->rt_time);
	    }

			    
	    if (err) {
		if (err == DS_ERR_MEM) {
		    Abort();
		    break;
		}
		dsWait(slp);
	    }
	    
	    DeviceCloseConnection((Device)dev);
	}
	
	dsDurationDestroy(slp);
	dsDurationDestroy(readout);
	dsDurationDestroy(brd_delay);
	dsTimeDestroy(waittime);
	
	return 0;
}

static int nanmProcessor(NANMDevice dev, NANMReaderBuf buf) {
    unsigned int i;
    DSTime trigger_time;
    char *stcBuffer;
    int err, trigger_err;
    int pos;
    size_t idlen;
    
    
    ds_assert_param(dev);
    ds_assert_param(buf);

    if (buf->buf.sf) {
	if (dev->dw) DeviceFreeWriters((Device)dev, dev->dw);
	dev->dw = DeviceInitWriters((Device)dev, NULL);
//	if (dev->triggerdw) DeviceFreeWriters((Device)dev, dev->triggerdw);
//	dev->triggerdw = DeviceInitWriters((Device)dev, "-triggers");
    }
    
    if (dev->dw) err = DeviceOpenWriters((Device)dev, dev->dw, NULL, buf->buf.tm);
    else err = DS_ERR_MEM;

    if (!err) {
	DeviceWritersSetDurationMilliseconds((Device)dev, dev->dw, buf->rt_time);
	err = DeviceWrite((Device)dev, dev->dw, DATA_DETECTOR_COUNTER, NULL, 0, buf->ids, buf->id, buf->value);
	DeviceCloseWriters((Device)dev, dev->dw);
    }
    
    if (err) {
	dsDebugServer(DS_DEBUG_ERROR, "NANM(%s). Data writeout failed: %lx(%s).", dev->device.id, err, dsError(err, NULL));
    }

    idlen = strlen(dev->device.id) + 1;
    stcBuffer = dsMalloca((2+idlen) * sizeof(uint8_t) + 3 * sizeof(uint16_t) + sizeof(time_t) + (2+buf->ids)*sizeof(uint32_t) + sizeof(double));
    if (stcBuffer) {
        *(char*)stcBuffer = 0xE1;
	pos = 1; 			*(uint16_t*)(stcBuffer+pos) = dev->device.type;
	pos += sizeof(uint16_t);	*(uint16_t*)(stcBuffer+pos) = dev->device.version;
	pos += sizeof(uint16_t);	memcpy(stcBuffer+pos, dev->device.id, idlen);
	pos += idlen*sizeof(uint8_t);	*(time_t*)(stcBuffer+pos) = dsTimeGetUnixTime(buf->buf.tm);
	pos+=sizeof(time_t);		*(uint32_t*)(stcBuffer+pos)=buf->rt_time;
	pos+=sizeof(uint32_t);		*(uint32_t*)(stcBuffer+pos)=buf->ids;
	pos+=sizeof(uint32_t);  	*(uint8_t*)(stcBuffer+pos) = DATA_TYPE_UINT32;
	pos+=sizeof(uint8_t);		*(double*)(stcBuffer+pos) = 1;
	pos+=sizeof(double);		*(uint16_t*)(stcBuffer+pos) = 0;
	pos+=sizeof(uint16_t);
	
	memcpy(stcBuffer+pos, buf->value, buf->ids*sizeof(uint32_t));

	dsDatagramServerBroadcast(dev->device.dsc, NULL, dev->device.id, stcBuffer, pos, 0, NULL);
	dsFreea(stcBuffer);
    }


    if (buf->trigger.triggers) {
	trigger_time = dsTimeCreate(NULL);
	if (trigger_time) {
	    for (i=0;i<buf->trigger.triggers;i++) {
		dsTimeCopy(trigger_time, buf->buf.tm);
		dsTimeAddMilliseconds(trigger_time, buf->trigger.trigger_time[i]);
	
		trigger_err = DeviceOpenWriters((Device)dev, dev->dw, "-trigger", trigger_time);
		if (trigger_err) err = trigger_err;
		else {
		    trigger_err = DeviceWrite((Device)dev, dev->dw, DATA_DETECTOR_COUNTER, NULL, 0, buf->ids, buf->id, buf->trigger.trigger_count[i]);
		    if (trigger_err) err = trigger_err;
		    DeviceCloseWriters((Device)dev, dev->dw);
		}
	    }
	    dsTimeDestroy(trigger_time);
	} else err = DS_ERR_MEM;
    }

    if (err) {
	dsDebugServer(DS_DEBUG_ERROR, "NANM(%s). Triggers writeout failed: %lx(%s).", dev->device.id, err, dsError(err, NULL));
    }

    return 0;
}

static void nanmReset(NANMDevice dev) {
    ds_assert_return(dev);
    dev->reset = 1;
}
