#include <stdio.h>
#include <string.h>

#include <dsdebug.h>
#include <dsassert.h>
#include <dserror.h>

#include "error.h"
#include "defaults.h"
#include "lowlevel.h"
#include "final.h"

#include "nmv.h"
#include "nmvbuffer.h"

#ifndef _WIN32 
# include "nmvfunctions.h"
#endif /* _WIN32 */


#define GateTimeOut 		196

static int nmvReader(NMVDevice dev);
static int nmvProcessor(NMVDevice dev, NMVReaderBuf buf);
static void nmvDeviceDestroy(NMVDevice dev);
static void nmvReset(NMVDevice dev);
static int nmvDebug(NMVDevice dev);

void nmvDeviceSetDefaults(NMVDevice dev) {
    DeviceSetDefaults((Device)dev);

    dev->dw = NULL;
    dev->channel = NULL;
    dev->channels = 0;
}

Device nmvDeviceCreate(DSConfigContext config, DSConfigReference ref) {
    int err = 0;
    NMVDevice dev;

#ifdef _WIN32 
    return NULL;
#endif /* _WIN32 */
 
    dev = (NMVDevice)malloc(sizeof(NMVDeviceS));
    if (!dev) return (Device)dev;

    nmvDeviceSetDefaults(dev);

    dev->lasttime = dsTimeCreate(NULL);
    if (!dev->lasttime) err = DS_ERR_MEM;
    
    dev->force_duration = dsDurationCreate(NULL);
    if (!dev->force_duration) err = DS_ERR_MEM;

    dev->lock = dsLockCreateContext(NULL);
    if (dev->lock) err = dsLockInitContext(dev->lock);
    else err = DS_ERR_MEM;
    
    if (err) {
	if (dev->force_duration) dsDurationDestroy(dev->force_duration);
	if (dev->lasttime) dsTimeDestroy(dev->lasttime);
	if (dev->lock) dsLockDestroyContext(dev->lock);
	free(dev);
	return NULL;
    }

    dev->device.version = NMV_VERSION;
    
//    dev->device.configure = nmvConfigure;
    dev->device.destroy = (DeviceDestroyFunction)nmvDeviceDestroy;
    dev->device.reader = (DeviceReaderFunction)nmvReader;
    dev->device.processor = (DeviceProcessorFunction)nmvProcessor;
    dev->device.reset = (DeviceResetFunction)nmvReset;
    dev->device.debug = (DeviceDebugFunction)nmvDebug;
    
    return (Device)dev;
}

static void nmvDeviceDestroy(NMVDevice dev) {
    ds_assert_return(dev);
    
    if (dev->dw) DeviceFreeWriters((Device)dev, dev->dw);
    if (dev->channel) free(dev->channel);
    dsTimeDestroy(dev->lasttime);
    dsDurationDestroy(dev->force_duration);
    dsLockDestroyContext(dev->lock);
    free(dev);
}


static int nmvReader(NMVDevice dev) {
	int err = 0;
	int i, start;
	unsigned char doread = 0;
	unsigned char informated = 0;
	uint16_t value;

	unsigned int ioport;
	unsigned int channels;
	uint16_t *channel = NULL;
	
    
	DSDuration slp;
	DSDuration duration;
	
	DSTime datatime;
	DSTime nextread;
	
	DSTimeS curtime;

	unsigned long tint;
	unsigned short int Gate;
	double kg;
	
	xmlNodePtr cfgnode, chipcfg;

	slp = dsDurationCreate(NULL);
	datatime = dsTimeCreate(NULL);
	nextread = dsTimeCreate(NULL);
	duration = dsDurationCreate(NULL);
	
	if ((!slp)||(!datatime)||(!nextread)||(!duration)) {
	    if (slp) dsDurationDestroy(slp);
	    if (datatime) dsTimeDestroy(datatime);
	    if (duration) dsDurationDestroy(duration);
	    if (nextread) dsTimeDestroy(nextread);
	    
	    Abort();
	    return DS_ERR_MEM;
	}    

	dsDurationImportMicroseconds(slp, NMV_TIMEOUT_REINIT);
	dsDurationZeroDuration(dev->force_duration);

	while (DeviceContinueLoop((Device)dev)) {
	    err = 0;
	    dev->device.status = DEVICE_STATUS_NOT_CONNECTED;
	    dev->reset = 0;

	    dsDebugServer(DS_DEBUG_INFO, "NMV(%s). Reininitializing!", dev->device.id);
	    
	    cfgnode = dsConfigGetConfigByReference(dev->device.cfg, dev->device.cfgref);
	    if (cfgnode) {
		chipcfg = dsXmlFindSubNode(cfgnode, "Configuration");
		if (chipcfg) {
		    dsXmlGetNodeDuration(chipcfg, "ReadoutInterval", dev->force_duration);
		    dev->channels = dsXmlGetNodeInt(chipcfg, "Channels", -1);
		    ioport = dsXmlGetNodeInt(chipcfg, "PortAddress", 0x378);
		}
		dsConfigReturnConfig(dev->device.cfg, cfgnode);
	    }
	    
	    if (dev->channels<0) {
		if (!informated) {
		    dsDebugServer(DS_DEBUG_ERROR, "NMV(%s). Error in configuration (Number of channels not specified)", dev->device.id);
		    informated = 1;
		}
		dsWait(slp);
		continue;
	    }
	    
	    if (dsDurationIsZero(dev->force_duration)) {
		if (!informated) {
		    dsDebugServer(DS_DEBUG_ERROR, "NMV(%s). Error in configuration (Duration is not specified)", dev->device.id);
		    informated = 1;
		}
		dsWait(slp);
		continue;
	    }

	    channels = dev->channels;
	    dsDurationCeilSeconds(dev->force_duration, 1);
	    
	    tint = dsDurationGetSeconds(dev->force_duration);
	    Gate = 128*tint - GateTimeOut;
	    kg = 128.*tint/Gate;

#ifndef _WIN32 
	    dsTimeCurrentTime(datatime);
	    dsTimeAddMilliseconds(datatime, 100);
	    dsTimeCeil(datatime, dev->force_duration);
	    
	    dsWaitRTUntil(datatime);

	    InitialisePorts(ioport);
	    InitCounter(0,1);					// Timer0 as One Shot
	    LoadCounter(0,Gate);				// Interval = Gate/128 s
	    for (i=0;i<channels;i++) InitCounter(i+1,0);	// Timers 1..nch as counters
	    ResetCounters(channels);

	    
	    
#else /* _WIN32 */
	    dsDebugServer(DS_DEBUG_ERROR, "NMV(%s). NMV device is not supported in Windows!", dev->device.id);
	    continue;
#endif /* _WIN32 */

	    dev->device.sf = 1;
	    while ((DeviceContinueLoop((Device)dev))&&(!err)&&(!dev->reset)) {
		channel = (uint16_t*)malloc(channels*sizeof(uint16_t));
		if (!channel) {
		    err = DS_ERR_MEM;
		    Abort();
		    break;
		}
		
		dsTimeAdd(datatime, dev->force_duration);
		
#ifndef _WIN32 
		err = ProcTIMERS(channels, channel, datatime);
#else /* _WIN32 */
		err = DS_ERR_NOTSUPPORTED;
#endif /* _WIN32 */
	

		if (err) free(channel);
		else {
		    dsDebugServer(DS_DEBUG_DEBUG, "NMV(%s). New iteration", dev->device.id);
		    
		    for (i=0;i<channels;i++)
			channel[i] = kg*channel[i];

		    dsLockWrite(dev->lock);
		    if (dev->channel) free(dev->channel);
		    dev->channels = channels;
		    dev->channel = channel;
		    dsLockReturn(dev->lock);

		    dev->device.status = DEVICE_STATUS_OK;
		    
			// submit
		    nmvBufferAdd(dev, datatime, channels, channel);
		} 
		    

		if (err) {
		    if (err == DS_ERR_MEM) {
			dsDebugServer(DS_DEBUG_FATAL, "NMV(%s). Fatal error: %X(%s)", dev->device.id, err, dsError(err, NULL));
			Abort();
		    } else {
			dsDebugServer(DS_DEBUG_ERROR, "NMV(%s). Error reading data: %X(%s)", dev->device.id, err, dsError(err, NULL));
		    }
		}
	    }
	    
	    dev->device.status = DEVICE_STATUS_NOT_CONNECTED;
	    dsWait(slp);
	}
	
	dsDurationDestroy(slp);
	dsDurationDestroy(duration);
	dsTimeDestroy(datatime);
	dsTimeDestroy(nextread);
	
	return 0;
}

static int nmvProcessor(NMVDevice dev, NMVReaderBuf buf) {
    int err;
    long duration;
    
    size_t pos;
    size_t idlen;
    char *stcBuffer;
    
    ds_assert_param(dev);
    ds_assert_param(buf);

    if (buf->buf.sf) {
	dsTimeCopy(dev->lasttime, buf->buf.tm);

	if (dev->dw) DeviceFreeWriters((Device)dev, dev->dw);
	dev->dw = DeviceInitWriters((Device)dev, NULL);

	return 0;
    } else {
	if (dsDurationIsZero(dev->force_duration))
	    duration = dsTimeDiffMilliseconds(dev->lasttime, buf->buf.tm);
	else
	    duration = dsDurationGetMilliseconds(dev->force_duration);
	    
	dsTimeCopy(dev->lasttime, buf->buf.tm);
    }

    if (dev->dw) err = DeviceOpenWriters((Device)dev, dev->dw, NULL, buf->buf.tm);
    else err = DS_ERR_MEM;

    if (!err) {
	DeviceWritersSetDurationMilliseconds((Device)dev, dev->dw, duration);
	err = DeviceWrite((Device)dev, dev->dw, DATA_CHANNEL|DATA_CHANNEL_UINT16, NULL, 0, buf->channels, NULL, buf->channel);
	DeviceCloseWriters((Device)dev, dev->dw);
    }

    if (err) {
	dsDebugServer(DS_DEBUG_ERROR, "NMV(%s). Data writeout failed: %lX(%s).", dev->device.id, err, dsError(err, NULL));
    }

    idlen = strlen(dev->device.id) + 1;
    stcBuffer = dsMalloca((2+idlen)*sizeof(uint8_t) + 3*sizeof(uint16_t) + sizeof(time_t) + (2)*sizeof(uint32_t) + buf->channels*sizeof(uint16_t) + sizeof(double));
    if (stcBuffer) {
        *(char*)stcBuffer = 0xE1;
	pos = 1;			*(uint16_t*)(stcBuffer+pos) = dev->device.type;
	pos += sizeof(uint16_t);	*(uint16_t*)(stcBuffer+pos) = dev->device.version;
	pos += sizeof(uint16_t);	memcpy(stcBuffer+pos, dev->device.id, idlen);
	pos += idlen*sizeof(uint8_t); 	*(time_t*)(stcBuffer+1) = dsTimeGetUnixTime(buf->buf.tm);
	pos+=sizeof(time_t);		*(uint32_t*)(stcBuffer+pos)=duration;
	pos+=sizeof(uint32_t);		*(uint32_t*)(stcBuffer+pos)=buf->channels;
	pos+=sizeof(uint32_t);  	*(uint8_t*)(stcBuffer+pos) = DATA_TYPE_UINT16;
	pos+=sizeof(uint8_t);		*(double*)(stcBuffer+pos) = 1;
	pos+=sizeof(double);		*(uint16_t*)(stcBuffer+pos) = 0;
	pos+=sizeof(uint16_t);
	
	memcpy(stcBuffer+pos, buf->channel, buf->channels*sizeof(uint16_t));
	pos+=buf->channels*sizeof(uint16_t);

	dsDatagramServerBroadcast(dev->device.dsc, NULL, dev->device.id, stcBuffer, pos, 0, NULL);
	dsFreea(stcBuffer);
    }

    return err;
}

static void nmvReset(NMVDevice dev) {
    ds_assert_return(dev);
    dev->reset = 1;
}

static int nmvDebug(NMVDevice dev) {
    size_t pos, idlen;
    char *stcBuffer;

    ds_assert_param(dev);

    if (dev->device.status != DEVICE_STATUS_OK) return 0;
    if ((!dev->device.debug_dsc)||(dsDatagramServerIsEmpty(dev->device.debug_dsc))) return 0;

    dsLockWrite(dev->lock);
    
    idlen = strlen(dev->device.id) + 1;
    stcBuffer = dsMalloca((2+idlen)*sizeof(uint8_t) + sizeof(time_t) + 3*sizeof(uint16_t) + sizeof(time_t) + (2)*sizeof(uint32_t) + dev->channels*sizeof(uint16_t) + sizeof(double));
    if (stcBuffer) {

	*(char*)stcBuffer = 0xC1;
	pos = 1; 			*(uint16_t*)(stcBuffer+pos) = dev->device.type;
	pos += sizeof(uint16_t);	*(uint16_t*)(stcBuffer+pos) = dev->device.version;
	pos += sizeof(uint16_t);	memcpy(stcBuffer+pos, dev->device.id, idlen);
	pos += idlen * sizeof(uint8_t);	*(time_t*)(stcBuffer+pos)=0;
	pos += sizeof(time_t);		*(uint32_t*)(stcBuffer+pos)=0;
	pos+=sizeof(uint32_t);		*(uint32_t*)(stcBuffer+pos)=dev->channels;
	pos+=sizeof(uint32_t);  	*(uint8_t*)(stcBuffer+pos) = DATA_TYPE_UINT16;
	pos+=sizeof(uint8_t);		*(double*)(stcBuffer+pos) = 1;
	pos+=sizeof(double);		*(uint16_t*)(stcBuffer+pos) = 0;
	pos+=sizeof(uint16_t);
	
	memcpy(stcBuffer+pos, dev->channel, dev->channels*sizeof(uint16_t));
	pos+=dev->channels*sizeof(uint16_t);
	
	dsLockReturn(dev->lock);
	
	dsDatagramServerBroadcast(dev->device.debug_dsc, NULL, dev->device.id, stcBuffer, pos, 0, NULL);
	dsFreea(stcBuffer);
	
    } else dsLockReturn(dev->lock);
    
    return 0;
}
