#define _DS_INCLUDE_XML_HEADERS

#include <stdio.h>
#include <string.h>

#include <dsdebug.h>
#include <dsassert.h>
#include <dserror.h>
#include <dsstring.h>
#include <dsxml.h>

#include "error.h"
#include "defaults.h"
#include "lowlevel.h"
#include "final.h"

#include "sevan.h"
#include "sevanbuffer.h"

static int sevanReader(SEVANDevice dev);
static int sevanProcessor(SEVANDevice dev, SEVANReaderBuf buf);
static void sevanDeviceDestroy(SEVANDevice dev);
static void sevanReset(SEVANDevice dev);
static int sevanDebug(SEVANDevice dev);
static char* sevanControl(SEVANDevice dev, const char *request);
static int sevanSetConfig(SEVANDevice dev, xmlDocPtr config);

void sevanDeviceSetDefaults(SEVANDevice dev) {
    DeviceSetDefaults((Device)dev);

    dev->dw = NULL;
    dev->channel = NULL;
    dev->control_data = NULL;
}

Device sevanDeviceCreate(DSConfigContext config, DSConfigReference ref) {
    int err = 0;
    SEVANDevice dev;
    
    dev = (SEVANDevice)malloc(sizeof(SEVANDeviceS));
    if (!dev) return (Device)dev;

    sevanDeviceSetDefaults(dev);

    dev->lasttime = dsTimeCreate(NULL);
    if (!dev->lasttime) err = DS_ERR_MEM;
    
    dev->force_duration = dsDurationCreate(NULL);
    if (!dev->force_duration) err = DS_ERR_MEM;

    dev->lock = dsLockCreateContext(NULL);
    if (dev->lock) err = dsLockInitContext(dev->lock);
    else err = DS_ERR_MEM;
    
    if (err) {
	if (dev->force_duration) dsDurationDestroy(dev->force_duration);
	if (dev->lasttime) dsTimeDestroy(dev->lasttime);
	if (dev->lock) dsLockDestroyContext(dev->lock);
	free(dev);
	return NULL;
    }

    dev->device.version = SEVAN_VERSION;
    
//    dev->device.configure = sevanConfigure;
    dev->device.destroy = (DeviceDestroyFunction)sevanDeviceDestroy;
    dev->device.reader = (DeviceReaderFunction)sevanReader;
    dev->device.processor = (DeviceProcessorFunction)sevanProcessor;
    dev->device.reset = (DeviceResetFunction)sevanReset;
    dev->device.debug = (DeviceDebugFunction)sevanDebug;
    dev->device.control = (DeviceControlFunction)sevanControl;
    dev->device.setconfig = (DeviceSetConfigurationFunction)sevanSetConfig;
    
    return (Device)dev;
}

static void sevanDeviceDestroy(SEVANDevice dev) {
    ds_assert_return(dev);
    
    if (dev->dw) DeviceFreeWriters((Device)dev, dev->dw);
    if (dev->channel) free(dev->channel);
    dsTimeDestroy(dev->lasttime);
    dsDurationDestroy(dev->force_duration);
    dsLockDestroyContext(dev->lock);
    free(dev);
}

static int sevanIsSpace(SEVANDevice dev, char ch) {
    if (((ch>='0')&&(ch<='9'))||(ch=='e')||(ch=='E')||(ch=='.')||(ch=='-')||(ch=='+')) return 0;
    return 1;
}

#define SEVAN_READ_WEATHER			1
#define SEVAN_READ_HV				2
#define SEVAN_READ_CONTROL_DATA_RETRY 		4
static int sevanReadControlData(SEVANDevice dev, Connection ccon, double *data, int flags) {
    int err = 0;
    int retries, i, j;
    char *buf, *stmp;
    char request[16];
    
    const char *pos;

    ds_assert_param(dev);
    ds_assert_param(ccon);
    ds_assert_param(data);

    if (flags&SEVAN_READ_WEATHER) {
	if (flags&SEVAN_READ_CONTROL_DATA_RETRY) retries = ccon->retries;
	else retries = 1;

	do {
	    ConnectionReadAll(ccon);
	    if (!ConnectionWrite(ccon, "ws ", 3)) {
    		buf = DeviceReadText((Device)dev, ccon, "");

		if (buf) dsDebugServer(DS_DEBUG_INFO, "SEVAN(%s). ReadControlData(ws): %s", dev->device.id, buf);
		else dsDebugServer(DS_DEBUG_INFO, "SEVAN(%s). ReadControlData(ws): No return", dev->device.id);
		
		if (buf) {
		    err = 0;
		
		    stmp = strstr(buf, "T=");
		    if (stmp) data[0] = atof(stmp+2);
		    else err = DS_ERR_INVALID_RESPONSE;

		    stmp = strstr(buf, "H=");
		    if (stmp) data[1] = atof(stmp+2);
		    else err = DS_ERR_INVALID_RESPONSE;

		    stmp = strstr(buf, "P=");
		    if (stmp) data[2] = atof(stmp+2);
		    else err = DS_ERR_INVALID_RESPONSE;
		
		    free(buf);
		} else err = DS_ERR_CONNECTION_READ;
	    } else err = DS_ERR_CONNECTION_WRITE;

	    if (err) retries--;
	} while ((err)&&(retries>0));

	if (err) {
	    for (i=0;i<SEVAN_WEATHER_DATA_SIZE;i++)
		data[i] = dev->control_data[i];
	}
    } 
    
    if (flags&SEVAN_READ_HV) {
	for (i=0;((!err)&&(i<dev->sensors));i++) {
	    sprintf(request,"h%u ",i+1);
	
	    if (flags&SEVAN_READ_CONTROL_DATA_RETRY) retries = ccon->retries;
	    else retries = 1;
	
	    do {	
	        ConnectionReadAll(ccon);
		if (!ConnectionWrite(ccon, request, strlen(request))) {
    		    buf = DeviceReadText((Device)dev, ccon, "");
		    if (buf) {
			if (buf) dsDebugServer(DS_DEBUG_INFO, "SEVAN(%s). ReadControlData(%s): %s", dev->device.id, request, buf);
			else dsDebugServer(DS_DEBUG_INFO, "SEVAN(%s). ReadControlData(%s): No return", dev->device.id, request);

			err = 0;
		    
			stmp = strstr(buf, "=");

			if (stmp) {
			    pos = stmp + 1;
			    for (j=0;((pos[j]<='9')&&(pos[j]>='0'));j++);
			    if ((pos[j])&&(pos[j]!=' ')&&(pos[j]!='\t')&&(pos[j]!='\r')&&(pos[j]!='\n')) {
				stmp = NULL;
			    }
			}
			
			if (stmp) data[SEVAN_WEATHER_DATA_SIZE+i] = atof(stmp+1);
			else err = DS_ERR_INVALID_RESPONSE;
		    
			free(buf);
		    } else err = DS_ERR_CONNECTION_READ;
    		} else err = DS_ERR_CONNECTION_WRITE;

		if (err) retries--;
	    } while ((err)&&(retries>0));
	}

	if (err) {
	    for (i=SEVAN_WEATHER_DATA_SIZE;i<SEVAN_WEATHER_DATA_SIZE+dev->sensors;i++)
		data[i] = dev->control_data[i];
	}
    }
    
    return err;
    
}

#define SEVAN_REQUEST_RETRY 	1
#define SEVAN_REQUEST_VOID	2
static char *sevanRequest(SEVANDevice dev, Connection ccon, const char *request, int flags) {
	int err = 0;
	int retries;

	char *req, *resp, *reqpos;
	char *buf;
	
	if (flags&SEVAN_REQUEST_RETRY) retries = ccon->retries;
	else retries = 1;

	do {
	    err = 0;
	    
	    ConnectionReadAll(ccon);
	    if (!ConnectionWrite(ccon, request, strlen(request))) {
		if (flags&SEVAN_REQUEST_VOID) buf = NULL;
    		else {
		    buf = DeviceReadText((Device)dev, ccon, "");
		    if (!buf) err = DS_ERR_CONNECTION_READ;
		}

		if (dsDebugCheckLevel(DS_DEBUG_INFO)) {
	    	    req = dsStringCopy(NULL, request);
		    if (req) {
			reqpos = strchr(req, '\r');
			if (reqpos) *reqpos = 0;
			reqpos = strchr(req, '\n');
			if (reqpos) *reqpos = 0;
		    }

		    if ((buf)&&(!dsDebugCheckLevel(DS_DEBUG_DEBUG))) {
			reqpos = strchr(buf, '=');
		        resp = dsStringCopy(NULL, reqpos?(reqpos+1):buf);
			if (resp) {
			    reqpos = strchr(resp, '\r');
			    if (reqpos) *reqpos = 0;
			    reqpos = strchr(resp, '\n');
			    if (reqpos) *reqpos = 0;
			}
		    } else resp = NULL;
			    
		    if (buf) dsDebugServer(DS_DEBUG_INFO, "SEVAN(%s). Request(%s): %s", dev->device.id, req?req:request, resp?resp:buf);
		    else dsDebugServer(DS_DEBUG_INFO, "SEVAN(%s). Request(%s): No return", dev->device.id, req?req:request);

		    if (resp) dsStringDestroy(NULL, resp);
		    if (req) dsStringDestroy(NULL, req);
		}		    
	    } else err = DS_ERR_CONNECTION_WRITE;

	    if (err) retries--;
	} while ((err)&&(retries>0));
	
	return err?NULL:buf;
}

static int sevanRequestInt(SEVANDevice dev, Connection ccon, const char *request, int default_value, int flags) {
    int res;
    char *resp;
    char *pos;
    size_t i;

    int retries = ccon->retries;

retry:    
    resp = sevanRequest(dev, ccon, request, flags);
    if (resp) {
	pos = strchr(resp, '=');

	    // Check for incorrect return	
	pos = pos?(pos+1):resp;
	for (i=0;((pos[i]<='9')&&(pos[i]>='0'));i++);
	if ((pos[i])&&(pos[i]!=' ')&&(pos[i]!='\t')&&(pos[i]!='\r')&&(pos[i]!='\n')) {
	    if (retries>0) {
		free(resp);
		goto retry;
	    }
	}
	
	res = atoi(pos);
//	res = atoi(pos?(pos+1):resp);
	free(resp);
	return res;
    }
    
    return default_value;
}

void sevanRequestVoid(SEVANDevice dev, Connection ccon, const char *request, int flags) {
    char *resp;
    
    resp = sevanRequest(dev, ccon, request, flags|SEVAN_REQUEST_VOID);
    if (resp) puts(resp);
    if (resp) free(resp);
}

static int sevanReader(SEVANDevice dev) {
	int control_err = 0, err = 0;
	int i, j, start, spl;
	unsigned char doread = 0;
	unsigned char informated = 0, cinformated = 0;
	uint16_t value;

	char *buf = NULL;
	unsigned int channels;
	double *channel = NULL;
	double *control_data = NULL;
	unsigned int max_control_errors, control_errors;
	
	unsigned char read_weather = 0, read_hv = 0;
	int rcd_flags = 0;
    
	Connection con, ccon;

	DSDuration slp;
	DSDuration duration;
	DSDuration timeout;
	
	DSTime datatime;
	
	xmlNodePtr cfgnode, chipcfg, thresholds;
	char request[16];
	double *threshold = NULL;

	slp = dsDurationCreate(NULL);
	datatime = dsTimeCreate(NULL);
	duration = dsDurationCreate(NULL);
	timeout = dsDurationCreate(NULL);
	
	if ((!slp)||(!datatime)||(!duration)||(!timeout)) {
	    if (slp) dsDurationDestroy(slp);
	    if (datatime) dsTimeDestroy(datatime);
	    if (duration) dsDurationDestroy(duration);
	    if (timeout) dsDurationDestroy(timeout);
	    
	    Abort();
	    return DS_ERR_MEM;
	}    

	dsDurationImportMicroseconds(slp, SEVAN_TIMEOUT_REINIT);
	dsDurationZeroDuration(timeout);
	dsDurationZeroDuration(dev->force_duration);
	

    	while (DeviceContinueLoop((Device)dev)) {
	    err = 0; control_err = 0;
	    dev->device.status = DEVICE_STATUS_NOT_CONNECTED;
	    dev->reset = 0;

	    dsDebugServer(DS_DEBUG_INFO, "SEVAN(%s). Reininitializing!", dev->device.id);
	    dev->sensors = 3;	/* Currently we have exactly 3 sensors (one in each layer) */
	    spl = dev->sensors / 3;
	    
	    if (threshold) free(threshold);
	    threshold = (double*)malloc(dev->sensors * sizeof(double));
	    if (!threshold) {
		err = DS_ERR_MEM;
		Abort();
		break;
	    }

	    cfgnode = dsConfigGetConfigByReference(dev->device.cfg, dev->device.cfgref);
	    if (cfgnode) {
		chipcfg = dsConfigGetSubNode(dev->device.cfg, cfgnode, "Configuration", 0);
		if (chipcfg) {
		    dsConfigGetNodeDuration(dev->device.cfg, chipcfg, "Timeout", timeout);
		    dsConfigGetNodeDuration(dev->device.cfg, chipcfg, "ForceDuration", dev->force_duration);
		    read_hv = dsConfigGetNodeBoolean(dev->device.cfg, chipcfg, "ReadHighVoltage", 0);
		    read_weather = dsConfigGetNodeBoolean(dev->device.cfg, chipcfg, "ReadWeather", 0);
		    max_control_errors = dsConfigGetNodeInt(dev->device.cfg, chipcfg, "MaxControlErrors", 0);
		    thresholds = dsConfigGetSubNode(dev->device.cfg, chipcfg, "Thresholds", 1);
		    if (thresholds) {
			for (i=0;i<3;i++) {
			    for (j=0;j<spl;j++) {
				threshold[i*spl+j] = dsConfigGetDouble(dev->device.cfg, -1., thresholds, "Threshold[@layer=%u and @sensor=%u]", i+1, j+1);
			    }
			}
		    }
		}
	
		dsConfigReturnConfig(dev->device.cfg, cfgnode);
	    }
	    
	    
	    rcd_flags = 0;
	    if (read_hv) rcd_flags |= SEVAN_READ_HV;
	    if (read_weather) rcd_flags |= SEVAN_READ_WEATHER;
	    dev->rcd_flags = rcd_flags;
	    
	    con = DeviceInitializeConnection((Device)dev);
	    if (con) informated = 0;
	    else {
		if (!informated) {
		    dsDebugServer(DS_DEBUG_CRITICAL, "SEVAN(%s). Error opening communication device!", dev->device.id);
		    informated = 1;
		}
		DeviceCloseConnection((Device)dev);
		dsWait(slp);
		continue;
	    } 

	    ConnectionReadAll(con);
	
	    cfgnode = dsConfigGetConfigByReference(dev->device.cfg, dev->device.cfgref);
	    if (cfgnode) {
		chipcfg = dsConfigGetSubNode(dev->device.cfg, cfgnode, "Configuration", 0);
		if (chipcfg) {
		    thresholds = dsConfigGetSubNode(dev->device.cfg, chipcfg, "Thresholds", 1);
		    if (thresholds) {
			for (i=0;i<3;i++) {
			    for (j=0;j<spl;j++) {
				if (threshold[i*spl+j] < 0) {
				    sprintf(request, "U%u\r\n",i+1);
				    threshold[i*spl+j] = 0.01 * sevanRequestInt(dev, con, request, -1, SEVAN_REQUEST_RETRY);
				    if (threshold[i*spl+j] != -1) {
					dsConfigSetDouble(dev->device.cfg, threshold[i*spl+j], thresholds, "Threshold[@layer=%u and @sensor=%u]", i+1, j+1);
				    }
				} else {
				    sprintf(request, "U%u %u\r\n",i+1, lrint(100*threshold[i*spl+j]));
				    sevanRequestVoid(dev, con, request, SEVAN_REQUEST_RETRY);
				}
			    }
			}
		    }
		}
		dsConfigReturnConfig(dev->device.cfg, cfgnode);
	    }
	

	    ccon = DeviceInitializeControlConnection((Device)dev);
	    if (ccon) {
		ConnectionReadAll(ccon);
		dev->control_data = (double*)dsMalloc(NULL,(dev->sensors+SEVAN_WEATHER_DATA_SIZE)*sizeof(double));
		control_data = (double*)dsMalloc(NULL,(dev->sensors+SEVAN_WEATHER_DATA_SIZE)*sizeof(double));
		if ((dev->control_data)&&(control_data))
		    err = sevanReadControlData(dev, ccon, dev->control_data, rcd_flags|SEVAN_READ_CONTROL_DATA_RETRY);
		else
		    err = DS_ERR_MEM;
		
		if (err) {
		    if (dev->control_data) {
			dsFree(NULL, dev->control_data);
			dev->control_data = NULL;
		    }
		    if (control_data) {
			dsFree(NULL, control_data);
			control_data = NULL;
		    }
		    dsDebugServer(DS_DEBUG_WARNING, "SEVAN(%s). Bypassing invalid control connection!", dev->device.id);
		    DeviceCloseControlConnection((Device)dev);
		    ccon = NULL;
		    err = 0;
		} else {
		    cinformated = 0;
		    control_errors = 0;
		    ConnectionUnlock(ccon);
		}
	    } else {
		if (!cinformated) {
		    dsDebugServer(DS_DEBUG_WARNING, "SEVAN(%s). Error opening control connection!", dev->device.id);
		    cinformated = 1;
		}
		DeviceCloseControlConnection((Device)dev);
	    }

	    
	    dev->device.sf = 1;
	    dsTimeCurrentTime(datatime);
	    while ((DeviceContinueLoop((Device)dev))&&(!err)&&(!dev->reset)) {
		buf = DeviceReadText((Device)dev, con, NULL);
		
		if (buf) {
		    dsTimeCurrentTime(datatime);
		    
		    dsDebugServer(DS_DEBUG_DEBUG, "SEVAN(%s). New iteration", dev->device.id);
		    
		    for (i=0, start=-1, channels=0;buf[i];i++) {
			if (sevanIsSpace(dev, buf[i])) {
			    if (i>(start+1)) channels++;
			    start = i;
			}
		    }
		    if (i>(start+1)) channels++;
		} else channels = 0;
		
		if (channels >= dev->sensors) {
		    channel = (double*)malloc(channels*sizeof(double));
		    if (!channel) {
			err = DS_ERR_MEM;
			Abort();
			break;
		    }

		    for (i=0, start = -1, channels = 0;buf[i];i++) {
			if (sevanIsSpace(dev, buf[i])) {
			    if (i>(start+1)) {
				buf[i] = 0;
				channel[channels++] = atof(buf + start + 1);
			    } 
			    start = i;
			}
		    }
		    if (i>(start+1)) channel[channels++] = atof(buf + start + 1);
		    free(buf); buf = NULL;
		    
		    dsLockWrite(dev->lock);
		    if (dev->channel) free(dev->channel);
		    dev->channels = channels;
		    dev->channel = channel;
		    dsLockReturn(dev->lock);

		    dev->device.status = DEVICE_STATUS_OK;
		    
		    if (ccon) {
			ConnectionLock(ccon);
			control_err = sevanReadControlData(dev, ccon, control_data, rcd_flags);
			ConnectionUnlock(ccon);

			if (control_err) {
			    if (control_errors > max_control_errors) {
				if (DeviceContinueLoop((Device)dev)) {
				    dsDebugServer(DS_DEBUG_ERROR, "SEVAN(%s). Error reading control information.", dev->device.id);
				    err = control_err;
				}
			    } else { 
				if (DeviceContinueLoop((Device)dev)) dsDebugServer(DS_DEBUG_INFO, "SEVAN(%s). Problem reading control information.", dev->device.id);
				control_errors++;
			    }
			} else {
			    control_errors = 0;
			    if (rcd_flags) {
				dsLockWrite(dev->lock);
				for (i=0;i<SEVAN_WEATHER_DATA_SIZE+dev->sensors;i++)
				    dev->control_data[i] = control_data[i];
				dsLockReturn(dev->lock);
			    }
			}
			
		    }
			// submit
		    sevanBufferAdd(dev, datatime, channels, channel, dev->sensors, rcd_flags, control_data);
		} else {
		    if (buf) {
			free(buf);
			buf = NULL;
		    }

			// checking for timeout
		    dsTimeCalcPassedDuration(duration, datatime);
		    if ((!dsDurationIsZero(timeout))&&(dsDurationCmp(duration, timeout)>0)) err = DS_ERR_TIMEOUT;
		}
		    

		if (err) {
		    if (err == DS_ERR_MEM) {
			dsDebugServer(DS_DEBUG_FATAL, "SEVAN(%s). Fatal error: %X(%s)", dev->device.id, err, dsError(err, NULL));
			Abort();
		    } else {
			if ((ccon)&&(control_err)) {
			    dsDebugServer(DS_DEBUG_ERROR, "SEVAN(%s). Error reading control data: %X(%s)", dev->device.id, control_err, dsError(control_err, NULL));
			    if (err!=control_err) dsDebugServer(DS_DEBUG_ERROR, "SEVAN(%s). Error reading data: %X(%s)", dev->device.id, err, dsError(err, NULL));
			} else dsDebugServer(DS_DEBUG_ERROR, "SEVAN(%s). Error reading data: %X(%s)", dev->device.id, err, dsError(err, NULL));
		    }
		}
	    }

	    dev->device.status = DEVICE_STATUS_NOT_CONNECTED;
	    if (dev->control_data) { dsFree(NULL, dev->control_data); dev->control_data = NULL; }
	    if (control_data) { dsFree(NULL, control_data); control_data = NULL; }
	    if (ccon) {
		    // Control Connection can be used at this point
		dsLockWrite(dev->lock);
		ConnectionLock(ccon);
		dsLockReturn(dev->lock);
		
		DeviceCloseControlConnection((Device)dev);
	    }
	    DeviceCloseConnection((Device)dev);
	    
	    dsWait(slp);
	}
	
	if (threshold) free(threshold);
	if (buf) free(buf);
	dsDurationDestroy(slp);
	dsDurationDestroy(duration);
	dsDurationDestroy(timeout);
	dsTimeDestroy(datatime);

	return err;
}

static int sevanProcessor(SEVANDevice dev, SEVANReaderBuf buf) {
    int i;
    int err;
    long duration;
    
    size_t pos;
    size_t idlen;
    char *stcBuffer;
    WriterDataID *ids;
    
    ds_assert_param(dev);
    ds_assert_param(buf);

    if (buf->buf.sf) {
	dsTimeCopy(dev->lasttime, buf->buf.tm);

	if (dev->dw) DeviceFreeWriters((Device)dev, dev->dw);
	dev->dw = DeviceInitWriters((Device)dev, NULL);

	return 0;
    } else {
	if (dsDurationIsZero(dev->force_duration))
	    duration = dsTimeDiffMilliseconds(dev->lasttime, buf->buf.tm);
	else
	    duration = dsDurationGetMilliseconds(dev->force_duration);
	    
	dsTimeCopy(dev->lasttime, buf->buf.tm);
    }

    if (dev->dw) err = DeviceOpenWriters((Device)dev, dev->dw, NULL, buf->buf.tm);
    else err = DS_ERR_MEM;

    if (!err) {
	ids = dsMalloca(buf->sensors*sizeof(uint32_t));
	if (ids) {
	    for (i=0;i<buf->sensors;i++)
		ids[i]=1+((i+1)<<16);
	} else err = DS_ERR_MEM;
    }

    if (!err) {
	DeviceWritersSetDurationMilliseconds((Device)dev, dev->dw, duration);
	err = DeviceWrite((Device)dev, dev->dw, DATA_CHANNEL|DATA_CHANNEL_INTENSITY, NULL, 0, buf->sensors, ids, buf->channel);
	if ((!err)&&((buf->channels-buf->sensors)>0)) err = DeviceWrite((Device)dev, dev->dw, DATA_CHANNEL, NULL, 0, buf->channels-buf->sensors, NULL, buf->channel+buf->sensors);
	if (buf->control_data) {
	    if (buf->rcd_flags&SEVAN_READ_WEATHER) {
		if (!err) err = DeviceWrite((Device)dev, dev->dw, DATA_CHANNEL|DATA_CHANNEL_TEMPERATURE, NULL, 0, 1, NULL, buf->control_data);
		if (!err) err = DeviceWrite((Device)dev, dev->dw, DATA_CHANNEL|DATA_CHANNEL_HUMIDITY, NULL, 0, 1, NULL, buf->control_data+1);
		if (!err) err = DeviceWrite((Device)dev, dev->dw, DATA_CHANNEL|DATA_CHANNEL_PRESSURE, NULL, 0, 1, NULL, buf->control_data+2);
	    }
	    if (buf->rcd_flags&SEVAN_READ_HV) {
		if (!err) err = DeviceWrite((Device)dev, dev->dw, DATA_CHANNEL|DATA_CHANNEL_VOLTAGE, NULL, 0, buf->sensors, ids, buf->control_data+SEVAN_WEATHER_DATA_SIZE);
	    }
	}
	DeviceCloseWriters((Device)dev, dev->dw);
    }
    
    if (ids) dsFreea(ids);

    if (err) {
	dsDebugServer(DS_DEBUG_ERROR, "SEVAN(%s). Data writeout failed: %lX(%s).", dev->device.id, err, dsError(err, NULL));
    }

    idlen = strlen(dev->device.id) + 1;
    stcBuffer = dsMalloca((2+idlen)*sizeof(uint8_t) + 3*sizeof(uint16_t) + sizeof(time_t) + (2)*sizeof(uint32_t) + (1+buf->channels)*sizeof(double));
    if (stcBuffer) {
        *(char*)stcBuffer = 0xE1;
	pos = 1;			*(uint16_t*)(stcBuffer+pos) = dev->device.type;
	pos += sizeof(uint16_t);	*(uint16_t*)(stcBuffer+pos) = dev->device.version;
	pos += sizeof(uint16_t);	memcpy(stcBuffer+pos, dev->device.id, idlen);
	pos += idlen*sizeof(uint8_t); 	*(time_t*)(stcBuffer+1) = dsTimeGetUnixTime(buf->buf.tm);
	pos+=sizeof(time_t);		*(uint32_t*)(stcBuffer+pos)=duration;
	pos+=sizeof(uint32_t);		*(uint32_t*)(stcBuffer+pos)=buf->channels;
	pos+=sizeof(uint32_t);  	*(uint8_t*)(stcBuffer+pos) = DATA_TYPE_MIXED;
	pos+=sizeof(uint8_t);		*(double*)(stcBuffer+pos) = 1;
	pos+=sizeof(double);		*(uint16_t*)(stcBuffer+pos) = 0;
	pos+=sizeof(uint16_t);
	
	memcpy(stcBuffer+pos, buf->channel, buf->channels*sizeof(double));
	pos+=buf->channels*sizeof(double);

	dsDatagramServerBroadcast(dev->device.dsc, NULL, dev->device.id, stcBuffer, pos, 0, NULL);
	dsFreea(stcBuffer);
    }

    return err;
}

static void sevanReset(SEVANDevice dev) {
    ds_assert_return(dev);
    dev->reset = 1;
}

static int sevanDebug(SEVANDevice dev) {
    size_t pos, idlen;
    char *stcBuffer;

    ds_assert_param(dev);

    if (dev->device.status != DEVICE_STATUS_OK) return 0;
    if ((!dev->device.debug_dsc)||(dsDatagramServerIsEmpty(dev->device.debug_dsc))) return 0;

    dsLockWrite(dev->lock);
    
    idlen = strlen(dev->device.id) + 1;
    stcBuffer = dsMalloca((2+idlen)*sizeof(uint8_t) + sizeof(time_t) + 3*sizeof(uint16_t) + sizeof(time_t) + (2)*sizeof(uint32_t) + (1+dev->channels)*sizeof(double));
    if (stcBuffer) {

	*(char*)stcBuffer = 0xC1;
	pos = 1; 			*(uint16_t*)(stcBuffer+pos) = dev->device.type;
	pos += sizeof(uint16_t);	*(uint16_t*)(stcBuffer+pos) = dev->device.version;
	pos += sizeof(uint16_t);	memcpy(stcBuffer+pos, dev->device.id, idlen);
	pos += idlen * sizeof(uint8_t);	*(time_t*)(stcBuffer+pos)=0;
	pos += sizeof(time_t);		*(uint32_t*)(stcBuffer+pos)=0;
	pos+=sizeof(uint32_t);		*(uint32_t*)(stcBuffer+pos)=dev->channels;
	pos+=sizeof(uint32_t);  	*(uint8_t*)(stcBuffer+pos) = DATA_TYPE_MIXED;
	pos+=sizeof(uint8_t);		*(double*)(stcBuffer+pos) = 1;
	pos+=sizeof(double);		*(uint16_t*)(stcBuffer+pos) = 0;
	pos+=sizeof(uint16_t);
	
	memcpy(stcBuffer+pos, dev->channel, dev->channels*sizeof(double));
	pos+=dev->channels*sizeof(double);
	
	dsLockReturn(dev->lock);
	
	dsDatagramServerBroadcast(dev->device.debug_dsc, NULL, dev->device.id, stcBuffer, pos, 0, NULL);
	dsFreea(stcBuffer);
	
    } else dsLockReturn(dev->lock);
    
    return 0;
}

static char* sevanControl(SEVANDevice dev, const char *request) {
    char *res;
    unsigned short int size = 0;

    Connection con;

    dsLockWrite(dev->lock);

    if (dev->device.status != DEVICE_STATUS_OK) {
	dsLockReturn(dev->lock);
	return NULL;
    }

    con = DeviceGetControlConnection((Device)dev);
    dsLockReturn(dev->lock);
    
    if (!con) return NULL;
    
    res = sevanRequest(dev, con, request, SEVAN_REQUEST_RETRY);

    DeviceReturnControlConnection((Device)dev);

    return res;
}


static int sevanSetConfig(SEVANDevice dev, xmlDocPtr config) {
    int err;
    xmlNodePtr rnode, vnode;
    xmlNodeSetPtr voltages;
    int i;
    unsigned int sensor, layer, spl;
    unsigned int voltage;
    char request[32];
    Connection con;
        
    spl = dev->sensors / 3;
    
    rnode = xmlDocGetRootElement(config);
    if (!rnode) return 0;
    
    voltages = dsXmlFindSubNodes(rnode, "Voltages", "Voltage");
    if (!voltages) return 0;

    dsLockWrite(dev->lock);
    if (dev->device.status != DEVICE_STATUS_OK) {
	dsXmlFreeSubNodes(voltages);
	dsLockReturn(dev->lock);
	return 0;
    }
    con = DeviceGetControlConnection((Device)dev);
    dsLockReturn(dev->lock);
    
    if (!con) {
	dsXmlFreeSubNodes(voltages);
	return 0;
    }

    for (i=0;i<voltages->nodeNr;i++) {
	vnode = voltages->nodeTab[i];
	
	layer = dsXmlGetPropInt(vnode, "layer", 0);
	sensor = dsXmlGetPropInt(vnode, "sensor", 0);
	voltage = dsXmlGetNodeInt(vnode, NULL, 0);
	if ((sensor<=0)||(sensor>spl)) continue;
	if ((layer<1)||(layer>3)) continue;
	if ((voltage<900)||(voltage>3000)) continue;
	
	sprintf(request, "V%u %u ",layer, voltage);
	sevanRequestVoid(dev, con, request, SEVAN_REQUEST_RETRY);
    }

    DeviceReturnControlConnection((Device)dev);
    
    dsXmlFreeSubNodes(voltages);
    return 0;
}

