#include <stdio.h>
#include <string.h>
#include <math.h>

#include <dsassert.h>
#include <dsdebug.h>
#include <dserror.h>
#include <dsdirectory.h>

#include "defaults.h"

#include "writer.h"
#include "filewriter.h"

static void FileWriterDestroy(FileWriter writer);
static WriterInstance FileWriterCreateInstance(FileWriter ctx, const char *instid);
static int FileWriterClose(FileWriter writer, FileWriterInstance instance);
static int FileWriterOpen(FileWriter writer, FileWriterInstance instance, const char *suffix, DSTime time);
static int FileWriterWrite(FileWriter ctx, FileWriterInstance instance, WriterDataType type, WriterDataOptions options, WriterLayerID layerid, WriterDataID n, WriterDataID *id, WriterData data);

Writer FileWriterCreate(xmlNodePtr cfg) {
    FileWriter writer;
    
    writer = (FileWriter)malloc(sizeof(FileWriterS));
    if (!writer) return (Writer)writer;

    WriterSetDefaults((Writer)writer);
    writer->path = NULL;
    writer->extension = NULL;

    dsExchangeString(NULL, &writer->path, dsXmlGetNodeString(cfg, "DataPath", DEFAULT_FILE_PATH));
    dsExchangeString(NULL, &writer->extension, dsXmlGetNodeString(cfg, "DataExtension", DEFAULT_FILE_EXTENSION));

    writer->writer.destroy = (WriterDestroyFunction)FileWriterDestroy;
    writer->writer.create_instance = (WriterCreateInstanceFunction)FileWriterCreateInstance;
    writer->writer.open = (WriterOpenFunction)FileWriterOpen;
    writer->writer.close = (WriterCloseFunction)FileWriterClose;
    writer->writer.write = (WriterWriteFunction)FileWriterWrite;

    return (Writer)writer;
}


static void FileWriterDestroy(FileWriter writer) {
    ds_assert_return(writer);
    dsExchangeString(NULL, &writer->path, NULL);
    dsExchangeString(NULL, &writer->extension, NULL);
    free(writer);
}

static WriterInstance FileWriterCreateInstance(FileWriter ctx, const char *instid) {
    return (WriterInstance)malloc(sizeof(FileWriterInstanceS));
}


static int FileWriterOpen(FileWriter writer, FileWriterInstance instance, const char *suffix, DSTime time) {
    int err;
    struct tm date;
    char ct[32];
    char fn[255];

    ds_assert_param(writer);
    ds_assert_param(instance);
    ds_assert_param(time);

    err = dsTimeExportGmtime(time, &date);
    if (err) return err;

    if ((strlen(writer->path)+strlen(instance->instance.id)+strlen(writer->extension))>220) return DS_ERR_TOOLONG;
    sprintf(fn,"%s%c%s%c%04u%02u%02u%s%s",writer->path,DS_DIRECTORY_SEPARATOR,instance->instance.id,DS_DIRECTORY_SEPARATOR,date.tm_year+1900,date.tm_mon+1,date.tm_mday,suffix?suffix:"",writer->extension);
    sprintf(ct,"%02u:%02u:%02u  ",date.tm_hour,date.tm_min,date.tm_sec);

    if (instance->instance.sf) {
	err = dsFileCreateDirectory(fn);
	if (err) return err;
	instance->instance.sf = 0;
    }

    instance->f = fopen(fn,"a+");
    if (!instance->f) return DS_ERR_FILE_OPEN;

    fprintf(instance->f,"%s",ct);

    return 0;
}

static int FileWriterClose(FileWriter writer, FileWriterInstance instance) {
    ds_assert_param(writer);
    ds_assert_param(instance);

    
    if (instance->f) {
	fprintf(instance->f, "\n");
	fclose(instance->f);
	instance->f = NULL;
    }
    
    return 0;
}


int FileWriterWrite(FileWriter ctx, FileWriterInstance instance, WriterDataType type, WriterDataOptions options, WriterLayerID layerid, WriterDataID n, WriterDataID *id, WriterData data) {
    int i, items = 0;
    double *dbldata;
    uint8_t *bytedata;
    uint16_t *shortdata;
    uint32_t *intdata;
    long int_value;
    
    ds_assert_param(ctx);
    ds_assert_param(instance);
    ds_assert_param((type&0xFFFF)<DATA_UNKNOWN);
    ds_assert_param(data);
    ds_assert_param(instance->f);
    
    switch(type&0xFFFF) {
	case DATA_CHANNEL:
	    items = n;

	    switch(type&0xFFFF0000) {
	        case DATA_CHANNEL_UINT8:
		    bytedata = (uint8_t*)data;
		    for (i=0; i<items;i++) {
			fprintf(instance->f, "% 5u ", bytedata[i]);
		    }
		break;
	        case DATA_CHANNEL_UINT16:
		    shortdata = (uint16_t*)data;
		    for (i=0; i<items;i++) {
			fprintf(instance->f, "% 9u ", shortdata[i]);
		    }
		break;
	        case DATA_CHANNEL_UINT32:
		    intdata = (uint32_t*)data;
		    for (i=0; i<items;i++) {
			fprintf(instance->f, "% 9lu ", intdata[i]);
		    }
		break;
    		default:
		    dbldata = (double*)data;

		    for (i=0; i<items;i++) {
			int_value = lround(dbldata[i]);
			if (fabs(dbldata[i]-int_value)<1E-4)
			    fprintf(instance->f, "% 9li ", int_value);
			else
			    fprintf(instance->f, "% 9.4lf ", dbldata[i]);
		
		    }
	    }
	break;
        case DATA_DETECTOR_COUNTER:
	    if (!items) items = n;
	case DATA_COINCIDENCE:
	    if (!items) {
		switch (type&0xFFFF0000) {
		    case DATA_COINCIDENCE_COINCIDENCE:
			items = n;
		    break;
		    case DATA_COINCIDENCE_PAIR:
			items = n * n;
		    break;
		    default:
			items = 1;
		}
	    }
	    
	    intdata = (uint32_t*)data;
	    for (i=0; i<items; i++)
		fprintf(instance->f, "%5u ", intdata[i]);

	break;

	case DATA_DISPERSION:
	    if (!items) items = n;
	case DATA_CORRELATION:
	    if (!items) items = n * (n - 1) / 2;

	    dbldata = (double*)data;
	    for (i=0; i<items;i++)
		fprintf(instance->f, "% 9.4lf ", dbldata[i]);
	break;

	default:
	    return DS_ERR_NOTSUPPORTED;
    }    
    
    return 0;
}

