#define _DS_INCLUDE_XML_HEADERS
#define _WRITER_C

#include <stdio.h>
#include <string.h>

#include <dsassert.h>
#include <dsdebug.h>

#include "writer.h"


Writer WriterCreate(DSConfigContext config, DSConfigReference ref) {
    int err;
    int i;
    const char *tmp;
    char *id = NULL;
    Writer ctx = NULL;
    DSLockContext lock;
    
    DSConfigReference cfgref;
    xmlNodePtr cfg;
    
    ds_assert_null(config);
    ds_assert_null(ref);
    
    cfg = dsConfigGetConfigByReference(config, ref);
    if (!cfg) return NULL;
    
    tmp = dsXmlGetPropString(cfg, "id", NULL);
    if (tmp) err = dsExchangeString(NULL, &id, tmp);
    else err = DS_ERR_CONFIG;

    for (i=0;writer_configuration[i].id;i++) {
	if (!xmlStrcmp(cfg->name, writer_configuration[i].id)) {
    	    ctx = writer_configuration[i].create(cfg);
	    break;
	}
    }
    if (!ctx) err = DS_ERR_CONFIG;

    dsConfigReturnConfig(config, cfg);
    
    lock = dsLockCreateContext(NULL);
    if (lock) {
	err = dsLockInitContext(lock);
	if (err) {
	    dsLockDestroyContext(lock);
	    lock = NULL;
	}
    }

    cfgref = dsConfigCopyReference(config, ref);
    
    if ((!cfgref)||(!ctx)||(!lock)||(err)) {
	if (ctx) {
	    if (ctx->destroy) ctx->destroy(ctx);
	    else free(ctx);
	}
	if (id) dsExchangeString(NULL, &id, NULL);
	if (lock) dsLockDestroyContext(lock);
	if (cfgref) dsConfigDestroyReference(config, cfgref);
	return NULL;
    }
    
    ctx->cfg = config;
    ctx->cfgref = cfgref;
    ctx->id = id;
    ctx->lock = lock;

    ctx->next = NULL;
    
    return ctx;
}

void WriterDestroy(Writer ctx) {
    ds_assert_return(ctx);
    
    dsLockDestroyContext(ctx->lock);
    dsConfigDestroyReference(ctx->cfg, ctx->cfgref);
    dsExchangeString(NULL, &ctx->id, NULL);
    
    if (ctx->destroy) ctx->destroy(ctx);
    else free(ctx);
}


void WriterSetDefaults(Writer ctx) {
    ds_assert_return(ctx);
    
    memset(ctx, 0, sizeof(WriterS));
}


Writer WriterFind(Writer ctx, const char *id) {
    if ((!ctx)||(!id)) return NULL;
    
    if (!strcmp(ctx->id, id)) return ctx;
    return WriterFind(ctx->next, id);
}

WriterInstance WriterCreateInstance(Writer ctx, Device dev, const char *instid) {
    WriterInstance instance;
    
    ds_assert_null(ctx);
    ds_assert_null(instid);
    
    
    if (ctx->create_instance) instance = ctx->create_instance(ctx, instid);
    else instance = (WriterInstance)malloc(sizeof(WriterInstanceS));
    
    if (!instance) return instance;
    
    instance->id = NULL;
    instance->sf = 1;
    instance->block = 0;
    instance->dev = dev;
    if (dsExchangeString(NULL, &instance->id, instid)) {
	WriterDestroyInstance(ctx, instance);
	return NULL;
    }

    return instance;
}

void WriterDestroyInstance(Writer ctx, WriterInstance instance) {
    ds_assert_return(ctx);
    ds_assert_return(instance);
    
    dsExchangeString(NULL, &instance->id, NULL);

    if (ctx->destroy_instance) ctx->destroy_instance(ctx, instance);
    else free(instance);
}

int WriterResetInstance(Writer ctx, WriterInstance instance) {
    ds_assert_param(instance);
    
    instance->sf = 1;
    return 0;
}

void WriterEnableFixedConfigOptimization(Writer ctx, WriterInstance instance) {
    ds_assert_return(ctx);
    ds_assert_return(instance);
    
    if (ctx->fixed_config) ctx->fixed_config(ctx, instance);
}

void WriterEnableFixedLayoutOptimization(Writer ctx, WriterInstance instance, const char *suffix) {
    ds_assert_return(ctx);
    ds_assert_return(instance);
    
    if (ctx->fixed_layout) ctx->fixed_layout(ctx, instance, suffix);
}

void WriterStartBlock(Writer ctx, WriterInstance instance) {
    ds_assert_return(instance);

    instance->block = 1;
}

void WriterFinishBlock(Writer ctx, WriterInstance instance) {
    ds_assert_return(instance);

    instance->block = 0;
}

int WriterOpen(Writer ctx, WriterInstance instance, const char *suffix, DSTime time) {
    int err = 0;
    
    ds_assert_param(ctx);
    ds_assert_param(instance);
    ds_assert_param(time);

    if (ctx->open) err = ctx->open(ctx, instance, suffix, time);
    if ((instance->block)&&(instance->block+1)) instance->block++;

    return err;
}

int WriterClose(Writer ctx, WriterInstance instance) {
    ds_assert_param(ctx);
    ds_assert_param(instance);
    
    if (ctx->close) return ctx->close(ctx, instance);
    return 0;
}

int WriterSetQuality(Writer ctx, WriterInstance instance, double quality) {
    ds_assert_param(ctx);
    ds_assert_param(instance);
    
    if (ctx->set_quality) return ctx->set_quality(ctx, instance, quality);
    return 0;
}

int WriterSetDuration(Writer ctx, WriterInstance instance, DSDuration duration) {
    ds_assert_param(ctx);
    ds_assert_param(instance);
    ds_assert_param(duration);
    
    if (ctx->set_duration) return ctx->set_duration(ctx, instance, duration);
    return 0;
}

int WriterSetDurationMilliseconds(Writer ctx, WriterInstance instance, unsigned long duration) {
    int err;
    DSDuration dur;
    
    ds_assert_param(ctx);
    ds_assert_param(instance);

    
    if (ctx->set_duration) {
	dur = dsDurationCreate(NULL);
	if (!dur) return DS_ERR_MEM;
	
	dsDurationImportMilliseconds(dur, duration);
	err = ctx->set_duration(ctx, instance, dur);
	dsDurationDestroy(dur);
	
	return err;
    }
    return 0;

}


int WriterWrite(Writer ctx, WriterInstance instance, WriterDataType type, WriterDataOptions options, WriterLayerID layerid, WriterDataID n, WriterDataID *id, WriterData data) {
    ds_assert_param(ctx);
    ds_assert_param(instance);
    ds_assert_param((type&0xFFFF)<DATA_UNKNOWN);
    ds_assert_param(data);
    
    if (ctx->write) return ctx->write(ctx, instance, type, options, layerid, n, id, data);
    return 0;
}

WriterDataOptions WriterDataOptionsCreate() {
    WriterDataOptions ctx;
    
    ctx = (WriterDataOptions)dsMalloc(NULL, sizeof(WriterDataOptionsS));
    if (ctx) memset(ctx, 0, sizeof(WriterDataOptionsS));
    return ctx;
}

int WriterDataOptionsSetType(WriterDataOptions ctx, const char *type) {
    ds_assert_param(ctx);
    
    return dsExchangeString(NULL, &ctx->type, type);
}

void WriterDataOptionsDestroy(WriterDataOptions ctx) {
    ds_assert_return(ctx);
    
    if (ctx->type) dsExchangeString(NULL, &ctx->type, NULL);
    dsFree(NULL, ctx);
}

