#ifndef _WRITER_H
#define _WRITER_H

#include <dslock.h>
#include <dsconfig.h>
#include <dstime.h>

typedef struct WriterT *Writer;
typedef struct WriterInstanceT *WriterInstance;
typedef struct WriterDataOptionsT *WriterDataOptions;

typedef void *WriterData;
typedef uint32_t WriterDataID;

#define did(ids, i) (ids?(ids[i]&0xFFFF):(i+1))
#define lid(layer, ids, i) (layer?layer:((ids?(ids[i]>>16):0)))

#define LAYER_1	0x10000
#define LAYER_2	0x20000

typedef unsigned char WriterLayerID;

enum WriterDataTypeT {
    DATA_CHANNEL,
    DATA_DETECTOR_COUNTER,
    DATA_COINCIDENCE,
    DATA_DISPERSION,
    DATA_CORRELATION,
    DATA_UNKNOWN
};

typedef enum WriterDataTypeT WriterDataType;

enum WriterChannelDataTypeT {
    DATA_CHANNEL_UNKNOWN = 0,
    DATA_CHANNEL_TEMPERATURE = 0x10000,
    DATA_CHANNEL_HUMIDITY = 0x20000,
    DATA_CHANNEL_PRESSURE = 0x30000,
    DATA_CHANNEL_VOLTAGE = 0x40000,
    DATA_CHANNEL_THRESHOLD = 0x50000,
    DATA_CHANNEL_INTENSITY = 0x60000,
    DATA_CHANNEL_UINT8 = 0x70000,
    DATA_CHANNEL_UINT16 = 0x80000,
    DATA_CHANNEL_UINT32 = 0x90000
};
typedef enum WriterChannelDataTypeT WriterChannelDataType;

enum WriterCoincidenceDataTypeT {
    DATA_COINCIDENCE_COINCIDENCE = 0,
    DATA_COINCIDENCE_PAIR = 0x10000,
    DATA_COINCIDENCE_MANY_MANY = 0x20000,
    DATA_COINCIDENCE_MANY_ZERO = 0x30000,
    DATA_COINCIDENCE_ZERO_MANY = 0x40000,
    DATA_COINCIDENCE_ZERO_ONE = 0x50000,
    DATA_COINCIDENCE_ONE_ZERO = 0x60000,
    DATA_COINCIDENCE_MANY_ONE = 0x70000,
    DATA_COINCIDENCE_ONE_MANY = 0x80000,
    DATA_COINCIDENCE_TOTAL = 0x90000
};
typedef enum WriterCoincidenceDataTypeT WriterCoincidenceDataType;


typedef Writer (*WriterCreateFunction)(xmlNodePtr cfg);
typedef void (*WriterDestroyFunction)(Writer ctx);

typedef WriterInstance (*WriterCreateInstanceFunction)(Writer ctx, const char *instid);
typedef void (*WriterDestroyInstanceFunction)(Writer ctx, WriterInstance instance);
typedef void (*WriterEnableFixedConfigOptimizationFunction)(Writer ctx, WriterInstance instance);
typedef void (*WriterEnableFixedLayoutOptimizationFunction)(Writer ctx, WriterInstance instance, const char *suffix);

typedef int (*WriterOpenFunction)(Writer ctx, WriterInstance instance, const char *suffix, DSTime time);
typedef int (*WriterCloseFunction)(Writer ctx, WriterInstance instance);
typedef int (*WriterSetQualityFunction)(Writer ctx, WriterInstance instance, double quality);
typedef int (*WriterSetDurationFunction)(Writer ctx, WriterInstance instance, DSDuration duration);

typedef int (*WriterWriteFunction)(Writer ctx, WriterInstance instance, WriterDataType type, WriterDataOptions options, WriterLayerID layerid, WriterDataID n, WriterDataID *id, WriterData data);

struct WriterDataOptionsT {
    char *type;
};
typedef struct WriterDataOptionsT WriterDataOptionsS;

struct WriterT {
    char *id;
    Writer next;
        
    DSLockContext lock;
    
    DSConfigContext cfg;
    DSConfigReference cfgref;
    
    WriterDestroyFunction destroy;
    
    WriterCreateInstanceFunction create_instance;
    WriterDestroyInstanceFunction destroy_instance;
    WriterEnableFixedConfigOptimizationFunction fixed_config;
    WriterEnableFixedLayoutOptimizationFunction fixed_layout;
    WriterOpenFunction open;
    WriterCloseFunction close;
    WriterSetQualityFunction set_quality;
    WriterSetDurationFunction set_duration;
    WriterWriteFunction write;
};
typedef struct WriterT WriterS;

#include <device.h>

struct WriterInstanceT {
    char *id;
    unsigned long int block;
    unsigned char sf;
    Device dev;
};
typedef struct WriterInstanceT WriterInstanceS;

Writer WriterCreate(DSConfigContext config, DSConfigReference ref);
void WriterDestroy(Writer ctx);
void WriterSetDefaults(Writer ctx);

Writer WriterFind(Writer ctx, const char *id);

/*
    Block Functions: WriterStartBlock/WriterFinishBlock
    Within a block only items with the same layout should be passed to the writer.
*/
WriterInstance WriterCreateInstance(Writer ctx, Device dev, const char *instid);
void WriterDestroyInstance(Writer ctx, WriterInstance instance);
int WriterResetInstance(Writer ctx, WriterInstance instance);
void WriterEnableFixedConfigOptimization(Writer ctx, WriterInstance instance);
void WriterEnableFixedLayoutOptimization(Writer ctx, WriterInstance instance, const char *suffix);
void WriterStartBlock(Writer ctx, WriterInstance instance);
void WriterFinishBlock(Writer ctx, WriterInstance instance);
int WriterOpen(Writer ctx, WriterInstance instance, const char *suffix, DSTime time);
int WriterClose(Writer ctx, WriterInstance instance);
int WriterSetQuality(Writer ctx, WriterInstance instance, double quality);
int WriterSetDuration(Writer ctx, WriterInstance instance, DSDuration duration);
int WriterSetDurationMilliseconds(Writer ctx, WriterInstance instance, unsigned long duration);
int WriterWrite(Writer ctx, WriterInstance instance, WriterDataType type, WriterDataOptions options, WriterLayerID layerid, WriterDataID n, WriterDataID *id, WriterData data);

WriterDataOptions WriterDataOptionsCreate();
int WriterDataOptionsSetType(WriterDataOptions ctx, const char *type);
void WriterDataOptionsDestroy(WriterDataOptions ctx);


#ifdef _WRITER_C
struct WriterConfigurationT {
    const char *id;
    WriterCreateFunction create;
};

typedef struct WriterConfigurationT *WriterConfiguration;
typedef struct WriterConfigurationT WriterConfigurationS;

#include "filewriter.h"
#include "xmlwriter.h"
WriterConfigurationS writer_configuration[] = {
    { "FileWriter", FileWriterCreate },
    { "XMLWriter", XMLWriterCreate },
    { NULL, NULL }
};
#endif /* _WRITER_C */

#endif /* _WRITER_C */
