#define _DS_INCLUDE_XML_HEADERS
#include <stdio.h>
#include <string.h>
#include <math.h>

#include <dsassert.h>
#include <dsdebug.h>
#include <dserror.h>
#include <dsdirectory.h>

#include <dsxml.h>
#include <libxml/c14n.h>

#define _XML_WRITER_C
#include "defaults.h"

#include "writer.h"
#include "xmlwriter.h"

#define MAX_SUFFIX_SIZE 64
#define INFO_SUFFIX "@suffix and starts-with('%s',@suffix)"

static void XMLWriterDestroy(XMLWriter writer);
static WriterInstance XMLWriterCreateInstance(XMLWriter ctx, const char *instid);
static void XMLWriterDestroyInstance(Writer ctx, XMLWriterInstance instance);
static void XMLWriterEnableFixedConfigOptimization(XMLWriter ctx, XMLWriterInstance instance);
static void XMLWriterEnableFixedLayoutOptimization(XMLWriter ctx, XMLWriterInstance instance, const char *suffix);
static int XMLWriterClose(XMLWriter writer, XMLWriterInstance instance);
static int XMLWriterOpen(XMLWriter writer, XMLWriterInstance instance, const char *suffix, DSTime time);
static int XMLWriterWrite(XMLWriter ctx, XMLWriterInstance instance, WriterDataType type, WriterDataOptions options, WriterLayerID layerid, WriterDataID n, WriterDataID *id, WriterData data);
static int XMLWriterSetQuality(XMLWriter ctx, XMLWriterInstance instance, double quality);
static int XMLWriterSetDuration(XMLWriter ctx, XMLWriterInstance instance, DSDuration duration);



Writer XMLWriterCreate(xmlNodePtr cfg) {
    XMLWriter writer;
    
    writer = (XMLWriter)malloc(sizeof(XMLWriterS));
    if (!writer) return (Writer)writer;

    WriterSetDefaults((Writer)writer);
    writer->path = NULL;
    writer->descpath = NULL;

    dsExchangeString(NULL, &writer->path, dsXmlGetNodeString(cfg, "DataPath", DEFAULT_FILE_PATH));
    dsExchangeString(NULL, &writer->descpath, dsXmlGetNodeString(cfg, "DescriptionPath", writer->path));

    writer->writer.destroy = (WriterDestroyFunction)XMLWriterDestroy;
    writer->writer.create_instance = (WriterCreateInstanceFunction)XMLWriterCreateInstance;
    writer->writer.destroy_instance = (WriterDestroyInstanceFunction)XMLWriterDestroyInstance;
    writer->writer.fixed_config = (WriterEnableFixedConfigOptimizationFunction)XMLWriterEnableFixedConfigOptimization;
    writer->writer.fixed_layout = (WriterEnableFixedLayoutOptimizationFunction)XMLWriterEnableFixedLayoutOptimization;
    writer->writer.open = (WriterOpenFunction)XMLWriterOpen;
    writer->writer.close = (WriterCloseFunction)XMLWriterClose;
    writer->writer.write = (WriterWriteFunction)XMLWriterWrite;
    writer->writer.set_quality = (WriterSetQualityFunction)XMLWriterSetQuality;
    writer->writer.set_duration = (WriterSetDurationFunction)XMLWriterSetDuration;

    return (Writer)writer;
}


static void XMLWriterDestroy(XMLWriter writer) {
    ds_assert_return(writer);
    dsExchangeString(NULL, &writer->path, NULL);
    dsExchangeString(NULL, &writer->descpath, NULL);
    free(writer);
}

static XMLWriterSuffixConfig xmlWriterSuffixConfigCreate(const char *suffix);
static void xmlWriterSuffixConfigDestroy(XMLWriterSuffixConfig sufcfg);
static int xmlWriterSuffixConfigReset(XMLWriterSuffixConfig sufcfg, void *attr);

static WriterInstance XMLWriterCreateInstance(XMLWriter ctx, const char *instid) {
    XMLWriterInstance instance;
    
    instance = malloc(sizeof(XMLWriterInstanceS));
    if (instance) {
	instance->data = xmlNewDoc("1.0");

	instance->sufconfigs = dsHashTableInit(NULL, XML_WRITER_SUFFIX_HASH_SIZE, DS_HASH_TYPE_POINTER, NULL);
	if (instance->sufconfigs) {
	    dsHashTableSetDestructor(instance->sufconfigs, (DSHashTableDestructorFunction)xmlWriterSuffixConfigDestroy);
	}

	if ((!instance->data)||(!instance->sufconfigs)) {
	    if (instance->data) xmlFreeDoc(instance->data);
	    if (instance->sufconfigs) dsHashTableDestroy(instance->sufconfigs);
	    free(instance);
	    return NULL;
	}
    } else return NULL;

    instance->doc = NULL;
    instance->xp = NULL;
    
    instance->suffix = NULL;
    instance->layoutid = NULL;
    instance->configid = NULL;
    instance->sufcfg = NULL;
    instance->fixed_config = 0;
    
    return (WriterInstance)instance;

}

static void XMLWriterDestroyInstance(Writer ctx, XMLWriterInstance instance) {
    ds_assert_return(instance);
    if (instance->sufconfigs) dsHashTableDestroy(instance->sufconfigs);
    if (instance->suffix) dsExchangeString(NULL,&instance->suffix, NULL);
    if (instance->layoutid) dsExchangeString(NULL,&instance->layoutid, NULL);
    if (instance->configid) dsExchangeString(NULL,&instance->configid, NULL);
    xmlFreeDoc(instance->data);
    free(instance);
}

static XMLWriterSuffixConfig xmlWriterSuffixConfigCreate(const char *suffix) {
    XMLWriterSuffixConfig sufcfg; 
    
    sufcfg = (XMLWriterSuffixConfig)dsMalloc(NULL, sizeof(XMLWriterSuffixConfigS));
    if (sufcfg) {
	sufcfg->suffix = NULL;
	sufcfg->layoutid = NULL;
	dsExchangeString(NULL, &sufcfg->suffix, suffix?suffix:"");
    }
    return sufcfg;
}

static void xmlWriterSuffixConfigDestroy(XMLWriterSuffixConfig sufcfg) {
    dsExchangeString(NULL, &sufcfg->layoutid, NULL);
    dsExchangeString(NULL, &sufcfg->suffix, NULL);
    dsFree(NULL, sufcfg);
}

static int xmlWriterSuffixConfigReset(XMLWriterSuffixConfig sufcfg, void *attr) {
    dsExchangeString(NULL, &sufcfg->layoutid, NULL);
    return 0;
}


static void XMLWriterEnableFixedConfigOptimization(XMLWriter ctx, XMLWriterInstance instance) {
    ds_assert_return(instance);
    instance->fixed_config = 1;
}

static void XMLWriterEnableFixedLayoutOptimization(XMLWriter ctx, XMLWriterInstance instance, const char *suffix) {
    XMLWriterSuffixConfig sufcfg;
    
    ds_assert_return(instance);

    sufcfg = xmlWriterSuffixConfigCreate(suffix);
    if ((sufcfg)&&(dsHashTableAdd(instance->sufconfigs, sufcfg, 1))) {
	xmlWriterSuffixConfigDestroy(sufcfg);
    }
}

static int XMLWriterOpen(XMLWriter writer, XMLWriterInstance instance, const char *suffix, DSTime time) {
    int err;
    struct tm date;
    char fn[512];
    const char *devname;
    const char *suffixstart;
    
    xmlNodePtr node;
    
    XMLWriterSuffixConfig sufcfg;
    unsigned char fixed_layout = 0, fixed_config = 0;

    ds_assert_param(writer);
    ds_assert_param(instance);
    ds_assert_param(time);

    err = dsTimeExportGmtime(time, &date);
    if (err) return err;

    if (suffix) {
	if (strlen(suffix)>MAX_SUFFIX_SIZE) return DS_ERR_TOOLONG;
	for (suffixstart = suffix; *suffixstart == '-'; suffixstart++);
	err = dsExchangeString(NULL, &instance->suffix, suffixstart);
    } else err = dsExchangeString(NULL, &instance->suffix, suffix);
    if (err) return err;

    
    if ((strlen(writer->path)+strlen(instance->instance.id))>220) return DS_ERR_TOOLONG;
    snprintf(fn,511,"%s%c%s%c%04u%02u%02u%s%s",writer->path,DS_DIRECTORY_SEPARATOR,instance->instance.id,DS_DIRECTORY_SEPARATOR,date.tm_year+1900,date.tm_mon+1,date.tm_mday,suffix?suffix:"",".xml");

    if (instance->instance.sf) {
	err = dsFileCreateDirectory(fn);
	if (err) return err;

	dsHashTableWalk(instance->sufconfigs, (DSHashTableWalkFunction)xmlWriterSuffixConfigReset, NULL);

	if ((instance->sufcfg)&&(!strcmp(instance->sufcfg->suffix, suffix?suffix:""))) sufcfg = instance->sufcfg;
	else sufcfg = dsHashTableFind(instance->sufconfigs, suffix?suffix:"");
//	if (sufcfg) instance->sufcfg = sufcfg;
    } else {
	if (instance->instance.block<2) {
	    if ((instance->sufcfg)&&(!strcmp(instance->sufcfg->suffix, suffix?suffix:""))) sufcfg = instance->sufcfg;
	    else sufcfg = dsHashTableFind(instance->sufconfigs, suffix?suffix:"");
	    
	    if (sufcfg) {
//		instance->sufcfg = sufcfg;
		if (sufcfg->layoutid) {
		    err = dsExchangeString(NULL, &instance->layoutid, sufcfg->layoutid);
		    if (err) return err;
		    fixed_layout = 1;
		}
	    }
	    
	    if ((instance->fixed_config)&&(instance->configid)) fixed_config = 1;
	} else sufcfg = NULL;
    }
    instance->sufcfg = sufcfg;
    
    instance->f = fopen(fn,"a+");
    if (!instance->f) return DS_ERR_FILE_OPEN;
    
    if ((instance->instance.sf)||((instance->instance.block<2)&&((!fixed_config)||(!fixed_layout)))) {
	snprintf(fn, 511, "%s/%s.xml", writer->descpath,instance->instance.id);
	instance->doc = xmlReadFile(fn, NULL, 0);
	if (!instance->doc) {
	    devname = DeviceGetDeviceName(instance->instance.dev);
	    if (devname) {
		snprintf(fn, 511, "%s/%s.xml",writer->descpath,devname);
		instance->doc = xmlReadFile(fn, NULL, 0);
	        if (instance->doc) {
		    snprintf(fn, 511, "%s Detector: %s", devname, instance->instance.id);
		    node = xmlDocGetRootElement(instance->doc);
		    if (node) dsXmlSetNodeString(node, "Title", fn);
	        }
	    }
	}
	
	if (!instance->doc) {
	    instance->doc = xmlNewDoc("1.0");
	    if (!instance->doc) {
		fclose(instance->f);
		instance->f = NULL;
		return DS_ERR_XML;
	    }
    	    node = xmlNewChild((xmlNodePtr)instance->doc, NULL, "Installation", NULL);
	    if (node) xmlSetProp(node, "id", instance->instance.id);
	    
	    fixed_layout = 0; fixed_config = 0;
	    dsExchangeString(NULL,&instance->layoutid, NULL);
	    dsExchangeString(NULL,&instance->configid, NULL);
	} else {
	    node = xmlDocGetRootElement(instance->doc);
	    if (!node) node = xmlNewChild((xmlNodePtr)instance->doc, NULL, "Installation", NULL);
	}
	
	if (fixed_layout) instance->layout = NULL;
	else {
	    if (node) instance->layout = xmlNewChild(node, NULL, "Layout", NULL);
	    else instance->layout = NULL;
	
	    if (!instance->layout) {
		xmlFreeDoc(instance->doc);
		instance->doc = NULL;
	        fclose(instance->f);
		instance->f = NULL;
	        return DS_ERR_XML;
	    }
	    
	    instance->pos = 1;
	}
	
	instance->xp = xmlXPathNewContext(instance->doc);
	if (!instance->xp) {
	    xmlFreeDoc(instance->doc);
	    instance->doc = NULL;
	    fclose(instance->f);
	    instance->f = NULL;
	    return DS_ERR_XML;
	}
    }
    
    instance->data_node = xmlNewDocNode(instance->data, NULL, "Data", NULL);
    if (!instance->data_node) {
	if (instance->doc) {
	    if (instance->xp) {
		xmlXPathFreeContext(instance->xp);
		instance->xp = NULL;
	    }
	    xmlFreeDoc(instance->doc);
	    instance->doc = NULL;
	}
	fclose(instance->f);
	instance->f = NULL;
	return DS_ERR_XML;
    }
    
    instance->data_buf = xmlBufferCreate();
    if (!instance->data_buf) {
	xmlFreeNode(instance->data_node);
	if (instance->doc) {
	    if (instance->xp) {
		xmlXPathFreeContext(instance->xp);
		instance->xp = NULL;
	    }
	    xmlFreeDoc(instance->doc);
	    instance->doc = NULL;
	}
	fclose(instance->f);
	instance->f = NULL;
	return DS_ERR_XML;
    }

    xmlSetProp(instance->data_node, "installation", instance->instance.id);
    
    dsTimeGetDateTime(time, fn);
    xmlNewChild(instance->data_node, NULL, "Time", fn);

    instance->channel_type_obj = NULL;
    instance->default_channel_type_obj = NULL;

    instance->instance.sf = 0;
    
    return 0;
}

static int XMLWriterClose(XMLWriter writer, XMLWriterInstance instance) {
    int i;
    
    xmlXPathObjectPtr obj, o;
    xmlNodeSetPtr node_set;
    
    char fn[255 + MAX_SUFFIX_SIZE];
    DSTime curtime;
    xmlNodePtr value;

    xmlNodePtr cfgnode, infonode;
    xmlNodePtr pnode, node, cur;
    xmlChar *current, *orig;
    const char *geometryid, *oldgeometryid;
    const char *currentid;

    unsigned char dowrite = 0;
        
    ds_assert_param(writer);
    ds_assert_param(instance);

    if (instance->f) {
	if (instance->doc) {
	    if (instance->layout) {
		pnode = xmlDocGetRootElement(instance->doc);
		for (node = NULL, cur = pnode->children; cur; cur = cur->next) {
    		    if ((cur->type == XML_ELEMENT_NODE)&&(xmlStrEqual(cur->name, "Geometry"))) node = cur;
		}

		if (node) geometryid = dsXmlGetPropString(node, "id", NULL);
		else geometryid = NULL;
	    
		obj = xmlXPathEvalExpression("//Installation/Layout[@id]", instance->xp);
		if (obj) {
		    node_set = obj->nodesetval;
	    	    if (node_set) {
		
			instance->xp->node = instance->layout;
			o = xmlXPathEvalExpression("descendant::node()|descendant::node()/@*|descendant::node()/namespace::*", instance->xp);
			if (o) {
			    xmlC14NDocDumpMemory(instance->doc, o->nodesetval, 1, NULL, 0, &current);
			    xmlXPathFreeObject(o);
			} else current = NULL;

			if (current) {
			    for (i=0;((i<node_set->nodeNr)&&(instance->layout));i++) {

				instance->xp->node = node_set->nodeTab[i];
				o = xmlXPathEvalExpression("descendant::node()|descendant::node()/@*|descendant::node()/namespace::*", instance->xp);
				if (o) {
				    xmlC14NDocDumpMemory(instance->doc, o->nodesetval, 1, NULL, 0, &orig);
				    xmlXPathFreeObject(o);
			
				    if (!xmlStrcmp(current, orig)) {
					oldgeometryid = dsXmlGetPropString(node_set->nodeTab[i], "geometry", NULL);
					if ((!geometryid)||((oldgeometryid)&&(!strcmp(geometryid, oldgeometryid)))) {
					    currentid = dsXmlGetPropString(node_set->nodeTab[i], "id", NULL);
					    if (currentid) {
						dsExchangeString(NULL, &instance->layoutid, currentid);
						if (instance->sufcfg) dsExchangeString(NULL, &instance->sufcfg->layoutid, currentid);
						xmlUnlinkNode(instance->layout);
						xmlFreeNode(instance->layout);
						instance->layout = NULL;
					    }
					}
				    }
			
				    xmlFree(orig);
				}
			    }

			}
			if (current) xmlFree(current);
		    }
		    xmlXPathFreeObject(obj);
		}
	    }
    
	    if (instance->layout) {
		if (geometryid) xmlSetProp(instance->layout, "geometry", geometryid);
	    
		curtime = dsTimeCreate(NULL);
		if (curtime) {
		    dsTimeCurrentTime(curtime);
		    
		    if (instance->suffix) { 
			sprintf(fn,"%s-",instance->suffix);
			dsTimeGetDateTime(curtime, fn + strlen(fn));
		    } else {
			dsTimeGetDateTime(curtime, fn);
		    }
		    dsTimeDestroy(curtime);

		    xmlSetProp(instance->layout, "id", fn);
		    dsExchangeString(NULL,&instance->layoutid,fn);
		    
		    if (instance->sufcfg) dsExchangeString(NULL, &instance->sufcfg->layoutid, fn);
		    
		    dowrite = 1;
		}
	    }

	    if ((instance->configid)&&(instance->fixed_config)) cfgnode = NULL;
	    else cfgnode = DeviceGetConfig(instance->instance.dev, DEVICE_CONFIGURATION_CONFIGURATION|DEVICE_CONFIGURATION_PROCESSOR|DEVICE_CONFIGURATION_LAYOUT);

	    if (cfgnode) {
		infonode = DeviceGetDriverInfo(instance->instance.dev);
		if (infonode) {
		    xmlUnsetProp(infonode, "id");
		    xmlAddChild(cfgnode, infonode);
		}
		xmlUnsetProp(cfgnode, "id");
		xmlAddChild(pnode, cfgnode);
		obj = xmlXPathEvalExpression("//Installation/DeviceConfiguration[@id]", instance->xp);
	    } else obj = NULL;
	    
	    if (obj) {
		node_set = obj->nodesetval;
	    	if (node_set) {
		
		    instance->xp->node = cfgnode;
		    o = xmlXPathEvalExpression("descendant::node()|descendant::node()/@*|descendant::node()/namespace::*", instance->xp);
		    if (o) {
			xmlC14NDocDumpMemory(instance->doc, o->nodesetval, 1, NULL, 0, &current);
			xmlXPathFreeObject(o);
		    } else current = NULL;

		    if (current) {
			for (i=0;((i<node_set->nodeNr)&&(cfgnode));i++) {

			    instance->xp->node = node_set->nodeTab[i];
			    o = xmlXPathEvalExpression("descendant::node()|descendant::node()/@*|descendant::node()/namespace::*", instance->xp);
			    if (o) {
				xmlC14NDocDumpMemory(instance->doc, o->nodesetval, 1, NULL, 0, &orig);
				xmlXPathFreeObject(o);
			
				if (!xmlStrcmp(current, orig)) {
				    currentid = dsXmlGetPropString(node_set->nodeTab[i], "id", NULL);
				    if (currentid) {
					dsExchangeString(NULL,&instance->configid,currentid);
					xmlUnlinkNode(cfgnode);
					xmlFreeNode(cfgnode);
					cfgnode = NULL;
				    }
				}
				xmlFree(orig);
			    }
			}
		    }
		    if (current) xmlFree(current);
		}
		xmlXPathFreeObject(obj);
	    }


	    if (cfgnode) {
		curtime = dsTimeCreate(NULL);
		if (curtime) {
		    dsTimeCurrentTime(curtime);
		    dsTimeGetDateTime(curtime, fn);
		    dsTimeDestroy(curtime);

		    xmlSetProp(cfgnode, "id", fn);
		    dsExchangeString(NULL,&instance->configid, fn);
		    
		    dowrite = 1;
		}
	    }
	    
	    if (dowrite) {
		sprintf(fn,"%s/%s.xml",writer->descpath,instance->instance.id);
		xmlSaveFormatFile(fn, instance->doc, 1);
	    }

	    xmlXPathFreeContext(instance->xp);
	    instance->xp = NULL;
	    xmlFreeDoc(instance->doc);
	    instance->doc = NULL;
	}

	if (instance->layoutid) 
	    xmlSetProp(instance->data_node, "layout", instance->layoutid);

	if (instance->configid) 
	    xmlSetProp(instance->data_node, "configuration", instance->configid);

	value = xmlNewChild(instance->data_node, NULL, "Value", NULL);
	if (value) xmlNodeSetContentLen(value, xmlBufferContent(instance->data_buf), xmlBufferLength(instance->data_buf));
	else dsDebugServer(DS_DEBUG_CRITICAL, "XMLWriter. XML data serialization failed");
	xmlBufferFree(instance->data_buf);

	xmlElemDump(instance->f, instance->data, instance->data_node);
	fprintf(instance->f, "\n");

/*	buf = xmlBufferCreate();
	if (buf) {
	    xmlNodeDump(buf, instance->data, instance->data_node, 0, 0);
	    xmlBufferDump(instance->f, buf);
	    fprintf(instance->f, "\n");
	    xmlBufferFree(buf);
	} else dsDebugServer(DS_DEBUG_CRITICAL, "XMLWriter. XML data serialization failed");*/
	xmlFreeNode(instance->data_node);

	fclose(instance->f);
	instance->f = NULL;
	
    }

    if (instance->channel_type_obj) {
	xmlXPathFreeObject(instance->channel_type_obj);
	instance->channel_type_obj = NULL;
    }
    
    if (instance->default_channel_type_obj) {
	xmlXPathFreeObject(instance->default_channel_type_obj);
	instance->default_channel_type_obj = NULL;
    }
    
    return 0;
}

static inline int FindID(const char *string, WriterDataID id) {
    char *interval;
    char buffer[32];
    const char *pos, *nextpos;
    WriterDataID min, max;
    if (!string) return 0;

    for(pos=string;pos;pos=nextpos?(nextpos+1):NULL) {
	nextpos = strchr(pos, ',');
	if (nextpos) {
	    if ((nextpos - pos) > 31) continue;
	    memcpy(buffer, pos, nextpos - pos);
	    buffer[nextpos - pos] = 0;
	} else {
	    strncpy(buffer, pos, 31);
	    buffer[31] = 0;
	}
	
	interval = strchr(buffer, '-');
	if (interval) {
	    *interval = 0;
	    min = atoi(buffer);
	    max = atoi(interval+1);
	    if ((id>=min)&&((!max)||(id<=max))) return 1;
	} else {
	    min = atoi(buffer);
	    if (id==min) return 1;
	}
    }

    return 0;
}

static xmlNodePtr XMLWriterGetSensorTypeNode(XMLWriter ctx, XMLWriterInstance instance, WriterLayerID layerid, WriterDataID detectorid) {
    int i;
    xmlNodePtr node;
    char request[128 + MAX_SUFFIX_SIZE];
    xmlXPathObjectPtr obj;
    xmlNodeSetPtr nset = NULL;

    const char *detectors;

    if (instance->suffix)    
	sprintf(request, "//Installation/Info[" INFO_SUFFIX "]/Layer[@id=%u]/SensorType", instance->suffix, layerid);
    else 
	sprintf(request, "//Installation/Info[not(@suffix)]/Layer[@id=%u]/SensorType", layerid);
    
    obj = xmlXPathEvalExpression(request, instance->xp);
    if (obj) {
	nset = obj->nodesetval;
	if (!nset) xmlXPathFreeObject(obj);
    }

    if ((!layerid)&&(!nset)) {
	if (instance->suffix)    
	    sprintf(request, "//Installation/Info[" INFO_SUFFIX "]/SensorType", instance->suffix);
	else
	    sprintf(request, "//Installation/Info[not(@suffix)]/SensorType");

	obj = xmlXPathEvalExpression(request, instance->xp);
	if (obj) {
	    nset = obj->nodesetval;
	    if (!nset) xmlXPathFreeObject(obj);
	}
    }
	
    if (nset) {
	for (i = 0; i < nset->nodeNr; i++) {
	    detectors = dsXmlGetNodeString(nset->nodeTab[i], NULL, NULL);
	    if (FindID(detectors, detectorid)) {
		node = nset->nodeTab[i];
		xmlXPathFreeObject(obj);
		return node;
	    }
	}
	xmlXPathFreeObject(obj);
	nset = NULL;
    }

    if (instance->suffix)
	sprintf(request, "//Installation/Info[" INFO_SUFFIX "]/Layer[@id=%u]/DefaultSensorType", instance->suffix, layerid);
    else
	sprintf(request, "//Installation/Info[not(@suffix)]/Layer[@id=%u]/DefaultSensorType", layerid);

    obj = xmlXPathEvalExpression(request, instance->xp);
    if (obj) {
	nset = obj->nodesetval;
	if ((!nset)||(!nset->nodeNr)) {
	    xmlXPathFreeObject(obj);
	    nset = NULL;
	}
    }	

    if (!nset) {
	if (instance->suffix)
	    sprintf(request, "//Installation/Info[" INFO_SUFFIX "]/DefaultSensorType", instance->suffix);
	else 
	    sprintf(request, "//Installation/Info[not(@suffix)]/DefaultSensorType");
	obj = xmlXPathEvalExpression(request, instance->xp);
	if (obj) {
	    nset = obj->nodesetval;
	    if ((!nset)||(!nset->nodeNr)) {
		xmlXPathFreeObject(obj);
		nset = NULL;
	    }
	}
    }

    if (nset) {	
	node = nset->nodeTab[0];
	xmlXPathFreeObject(obj);
	return node;
    }

    return NULL;
}

#define NOT_FOUND_PTR ((void*)-1)
static xmlNodePtr XMLWriterGetChannelTypeNode(XMLWriter ctx, XMLWriterInstance instance, WriterDataID channelid) {
    int i;
    xmlNodePtr node;
    char request[128 + MAX_SUFFIX_SIZE];
    xmlXPathObjectPtr obj;
    xmlNodeSetPtr nset = NULL;

    const char *detectors;

    if (instance->channel_type_obj) {
	if (instance->channel_type_obj == NOT_FOUND_PTR) nset = NULL;
	else nset = instance->channel_type_obj->nodesetval;
    } else {
	if (instance->suffix) 
	    sprintf(request, "//Installation/Info[" INFO_SUFFIX "]/ChannelType", instance->suffix);
	else
	    sprintf(request, "//Installation/Info[not(@suffix)]/ChannelType");
    
	obj = xmlXPathEvalExpression(request, instance->xp);
	if (obj) {
	    nset = obj->nodesetval;
	    if ((nset)&&(nset->nodeNr)) instance->channel_type_obj = obj;
	    else {
		xmlXPathFreeObject(obj);
		nset = NULL;
	    }
	}
    }

    if (nset) {
	for (i = 0; i < nset->nodeNr; i++) {
	    detectors = dsXmlGetNodeString(nset->nodeTab[i], NULL, NULL);
	    if (FindID(detectors, channelid)) {
		node = nset->nodeTab[i];
//		xmlXPathFreeObject(obj);
		return node;
	    }
	}
//	xmlXPathFreeObject(obj);
	nset = NULL;
    }

    if (instance->default_channel_type_obj) {
	if (instance->default_channel_type_obj == NOT_FOUND_PTR) nset = NULL;
	else nset = instance->default_channel_type_obj->nodesetval;
    } else {
	if (instance->suffix)
    	    sprintf(request, "//Installation/Info[" INFO_SUFFIX "]/DefaultChannelType", instance->suffix);
	else 
    	    sprintf(request, "//Installation/Info[not(@suffix)]/DefaultChannelType");

	obj = xmlXPathEvalExpression(request, instance->xp);
	if (obj) {
	    nset = obj->nodesetval;
	    if ((nset)&&(nset->nodeNr)) instance->default_channel_type_obj = obj;
	    else {
		xmlXPathFreeObject(obj);
		nset = NULL;
	    }
	}	
    }

    if (nset) {	
	node = nset->nodeTab[0];
//	xmlXPathFreeObject(obj);
	return node;
    }
    
    return NULL;
}

static WriterDataID XMLWriterGetChannelID(XMLWriter ctx, XMLWriterInstance instance, WriterLayerID layerid, WriterDataID detectorid) {
    WriterDataID res;
    xmlNodePtr node;
    char request[128];
    xmlXPathObjectPtr obj;
    xmlNodeSetPtr nset = NULL;

    sprintf(request, "/Installation/Layout[not(@id)]/Channel[@layer=%u and @sensor=%u]", layerid, detectorid);

    obj = xmlXPathEvalExpression(request, instance->xp);
    if (obj) {
	nset = obj->nodesetval;
	if ((!nset)||(!nset->nodeNr)) {
	    xmlXPathFreeObject(obj);
	    nset = NULL;
	}
    }

    if ((!layerid)&&(!nset)) {
	sprintf(request, "/Installation/Layout[not(@id)]/Channel[not(@layer) and @sensor=%u]", detectorid);

	obj = xmlXPathEvalExpression(request, instance->xp);
	if (obj) {
	    nset = obj->nodesetval;
	    if ((!nset)||(!nset->nodeNr)) {
		xmlXPathFreeObject(obj);
		nset = NULL;
	    }
	}
    }
	
    if (nset) {
	res = dsXmlGetNodeLong(nset->nodeTab[0], NULL, -1);
	xmlXPathFreeObject(obj);
    } else res = (WriterDataID)-1;

    return res;
}


int XMLWriterWrite(XMLWriter ctx, XMLWriterInstance instance, WriterDataType type, WriterDataOptions options, WriterLayerID layerid, WriterDataID n, WriterDataID *id, WriterData data) {
    WriterDataID targetid;
    WriterDataID *idcache;
    unsigned int i, j, items = 0;
    double *dbldata;
    uint8_t *bytedata;
    uint32_t *intdata;
    uint16_t *shortdata;
    long int_value;
    char stmp[32], spos[16];

    unsigned char done;    
    xmlNodePtr pnode, node;
    xmlChar *nodename;
    
    WriterDataID detector_id;
    WriterLayerID layer_id;
    
    
    ds_assert_param(ctx);
    ds_assert_param(instance);
    ds_assert_param((type&0xFFFF)<DATA_UNKNOWN);
    ds_assert_param(data);
    ds_assert_param(instance->f);
    
    if ((instance->doc)&&(instance->layout)) {
	switch(type&0xFFFF) {
	    case DATA_CHANNEL:
		pnode = instance->layout;
		
		for (i=0;i<n;i++) {
		    node = XMLWriterGetChannelTypeNode(ctx, instance, instance->pos);
		    sprintf(stmp, "%u", instance->pos++);
		    if (node) {
			node = xmlCopyNode(node, 1);
			if (node) {
			    nodename = xmlGetProp(node, "channel_type");
			    if (nodename) {
				xmlNodeSetName(node, nodename);
				xmlUnsetProp(node, "channel_type");
			    } else 
				xmlNodeSetName(node, "Channel");
			    xmlNodeSetContent(node, stmp);
			    xmlAddChild(pnode, node);
			}
		    } else node = xmlNewChild(pnode, NULL, "Channel", stmp);
		    
		    if (node) {
			xmlSetProp(node, "id", stmp);

			switch (type&0xFFFF0000) {
			    case DATA_CHANNEL_TEMPERATURE:
				xmlSetProp(node, "type", "temperature");
			    break;
			    case DATA_CHANNEL_HUMIDITY:
				xmlSetProp(node, "type", "humidity");
			    break;
			    case DATA_CHANNEL_PRESSURE:
				xmlSetProp(node, "type", "pressure");
			    break;
			    case DATA_CHANNEL_VOLTAGE:
				xmlSetProp(node, "type", "voltage");
				
				if (id) {
				    detector_id = did(id, i);
				    layer_id = lid(layerid, id, i);
		    		    
				    sprintf(stmp, "%u", detector_id);
				    xmlSetProp(node, "sensor", stmp);
				    if (layer_id) {
					sprintf(stmp, "%u", layer_id);
					xmlSetProp(node, "layer", stmp);
				    }
				}
			    break;
			    case DATA_CHANNEL_THRESHOLD:
				xmlSetProp(node, "type", "threshold");
				
				if (id) {
				    detector_id = did(id, i);
				    layer_id = lid(layerid, id, i);
		    		    
				    sprintf(stmp, "%u", detector_id);
				    xmlSetProp(node, "sensor", stmp);
				    if (layer_id) {
					sprintf(stmp, "%u", layer_id);
					xmlSetProp(node, "layer", stmp);
				    }
				}
			    break;
			    case DATA_CHANNEL_INTENSITY:
				nodename = xmlGetProp(node, "type");
				if ((!nodename)||(!strstr(nodename, "intensity")))
				    xmlSetProp(node, "type", "intensity");
				
				if (id) {
				    detector_id = did(id, i);
				    layer_id = lid(layerid, id, i);
		    		    
				    sprintf(stmp, "%u", detector_id);
				    xmlSetProp(node, "sensor", stmp);
				    if (layer_id) {
					sprintf(stmp, "%u", layer_id);
					xmlSetProp(node, "layer", stmp);
				    }
				}
			    break;
			    default:
				if (id) {
				    detector_id = did(id, i);
				    layer_id = lid(layerid, id, i);
		    		    
				    sprintf(stmp, "%u", detector_id);
				    xmlSetProp(node, "sensor", stmp);
				    if (layer_id) {
					sprintf(stmp, "%u", layer_id);
					xmlSetProp(node, "layer", stmp);
				    }
				}
				if ((options)&&(options->type)) {
				    xmlSetProp(node, "type", options->type);
				}
			}
		    }
		}
	    break;
    	    case DATA_DETECTOR_COUNTER:
		pnode = instance->layout;
		
		for (i=0;i<n;i++) {
		    detector_id = did(id, i);
		    layer_id = lid(layerid, id, i);

		    node = XMLWriterGetSensorTypeNode(ctx, instance, layer_id, detector_id);
		    if (!node)  node = XMLWriterGetChannelTypeNode(ctx, instance, instance->pos);

		    sprintf(stmp, "%u", instance->pos++);
	    
		    if (node) {
			node = xmlCopyNode(node, 1);
			if (node) {
			    xmlNodeSetName(node, "Channel");
			    xmlUnsetProp(node, "channel_type");
			    xmlNodeSetContent(node, stmp);
			    xmlAddChild(pnode, node);
			}
		    } else node = xmlNewChild(pnode, NULL, "Channel", stmp);
		    
		    if (node) {
			xmlSetProp(node, "id", stmp);
		    
		        sprintf(stmp, "%u", detector_id);
			xmlSetProp(node, "sensor", stmp);
			if (layer_id) {
			    sprintf(stmp, "%u", layer_id);
			    xmlSetProp(node, "layer", stmp);
			}
		    }
		}
	    break;
	    case DATA_COINCIDENCE:
		pnode = instance->layout;

		done = 0;
		switch (type&0xFFFF0000) {
			case DATA_COINCIDENCE_COINCIDENCE:
			    for (i = 0; i < n; i++) {
				node = XMLWriterGetChannelTypeNode(ctx, instance, instance->pos);

				sprintf(spos, "%u", instance->pos++);
				    
				if (node) {
				    node = xmlCopyNode(node, 1);
				    if (node) {
					xmlNodeSetName(node, "Coincidence");
					xmlUnsetProp(node, "channel_type");
					xmlNodeSetContent(node, spos);
					xmlAddChild(pnode, node);
				    } 
				} else node = xmlNewChild(pnode, NULL, "Coincidence", spos);
					
				if (node) {
				    xmlSetProp(node, "id", spos);
				    sprintf(stmp,"%u",i+1);
				    xmlSetProp(node, "direction", stmp);
				}
			    }
			    done = 1;
			break;
			case DATA_COINCIDENCE_PAIR:
			    for (i = 0; i < n; i++) {
				for (j = 0; j < n; j++) {
				    node = XMLWriterGetChannelTypeNode(ctx, instance, instance->pos);

				    sprintf(spos, "%u", instance->pos++);
				    
				    if (node) {
					node = xmlCopyNode(node, 1);
					if (node) {
					    xmlNodeSetName(node, "Coincidence");
					    xmlUnsetProp(node, "channel_type");
					    xmlNodeSetContent(node, spos);
					    xmlAddChild(pnode, node);
					} 
				    } else node = xmlNewChild(pnode, NULL, "Coincidence", spos);
					
				    if (node) {
					xmlSetProp(node, "id", spos);
					sprintf(stmp,"%u-%u",did(id,i),did(id,j));
					xmlSetProp(node, "direction", stmp);
				    }
				}
			    }
			    done = 1;
			break;
			case DATA_COINCIDENCE_MANY_MANY:
			    strcpy(stmp, "m-m");
			break;
			case DATA_COINCIDENCE_MANY_ZERO:
			    strcpy(stmp, "m-z");
			break;
			case DATA_COINCIDENCE_ZERO_MANY:
			    strcpy(stmp, "z-m");
			break;
			case DATA_COINCIDENCE_ZERO_ONE:
			    strcpy(stmp, "z-o");
			break;
			case DATA_COINCIDENCE_ONE_ZERO:
			    strcpy(stmp, "o-z");
			break;
			case DATA_COINCIDENCE_MANY_ONE:
			    strcpy(stmp, "m-o");
			break;
			case DATA_COINCIDENCE_ONE_MANY:
			    strcpy(stmp, "o-m");
			break;
			case DATA_COINCIDENCE_TOTAL:
			    instance->pos++;
			    done = 1;
			break;
			default:
			    return DS_ERR_NOTSUPPORTED;
		}
		
		if (!done) {
		    node = XMLWriterGetChannelTypeNode(ctx, instance, instance->pos);

		    sprintf(spos, "%u", instance->pos++);

		    if (node) {
			node = xmlCopyNode(node, 1);
			if (node) {
			    xmlNodeSetName(node, "Coincidence");
			    xmlUnsetProp(node, "channel_type");
			    xmlNodeSetContent(node, spos);
			    xmlAddChild(pnode, node);
			}
		    } else node = xmlNewChild(pnode, NULL, "Coincidence", spos);

		    if (node) {
			xmlSetProp(node, "id", spos);
			xmlSetProp(node, "direction", stmp);
		    }
		}
	    break;
	    case DATA_DISPERSION:
		pnode = instance->layout;

		for (i=0;i<n;i++) {
		    sprintf(stmp, "%u", instance->pos++);
		
		    node = xmlNewChild(pnode, NULL, "Variance", stmp);
		    if (node) {
			xmlSetProp(node, "id", stmp);
		    
		        targetid = XMLWriterGetChannelID(ctx, instance, lid(layerid, id, i), did(id, i));
			if (targetid != (WriterDataID)-1) {
			    sprintf(stmp, "%u", targetid);
			    xmlSetProp(node, "targetid", stmp);
			}
		    }
		}
	    break;
	    case DATA_CORRELATION:
		pnode = instance->layout;

		idcache = (WriterDataID*)dsMalloca(n*sizeof(WriterDataID));
		if (!idcache) return DS_ERR_MEM;
		
	    	for (i=0;i<n;i++) {
		    idcache[i] = XMLWriterGetChannelID(ctx, instance, lid(layerid, id, i), did(id, i));
		}
		
		for (i=0;i<n;i++) {
		    for (j=i+1;j<n;j++) {
			sprintf(stmp, "%u", instance->pos++);
			node = xmlNewChild(pnode, NULL, "Correlation", stmp);
			if (node) {
			    xmlSetProp(node, "id", stmp);
		
			    if (idcache[i] != (WriterDataID)-1) {
				sprintf(stmp, "%u", idcache[i]);
				xmlSetProp(node, "targetid1", stmp);
			    }
			    
			    if (idcache[j] != (WriterDataID)-1) {
				sprintf(stmp, "%u", idcache[j]);
				xmlSetProp(node, "targetid2", stmp);
			    }

			}
		    }
		}
		dsFreea(idcache);
	    break;
	    default:
		return DS_ERR_NOTSUPPORTED;
	}
    }
  
  
    switch(type&0xFFFF) {
	case DATA_CHANNEL:
	    items = n;

	    switch(type&0xFFFF0000) {
	        case DATA_CHANNEL_UINT8:
		    bytedata = (uint8_t*)data;
		    for (i=0; i<items;i++) {
	    		sprintf(stmp, "% 5u", bytedata[i]);
			xmlBufferCCat(instance->data_buf, stmp);
		    }
		break;
	        case DATA_CHANNEL_UINT16:
		    shortdata = (uint16_t*)data;
		    for (i=0; i<items;i++) {
	    		sprintf(stmp, "% 9u", shortdata[i]);
			xmlBufferCCat(instance->data_buf, stmp);
		    }
		break;
		case DATA_CHANNEL_UINT32:
		    intdata = (uint32_t*)data;
		    for (i=0; i<items;i++) {
			sprintf(stmp, "% 9lu", intdata[i]);
			xmlBufferCCat(instance->data_buf, stmp);
		    }
		break;
		default:   
		    dbldata = (double*)data;
		    for (i=0; i<items;i++) {
			int_value = lround(dbldata[i]);
			if (fabs(dbldata[i]-int_value)<1E-4)
			    sprintf(stmp, "% 9li ", int_value);
			else 
			    sprintf(stmp, "% 9.4lf ", dbldata[i]);
			
			xmlBufferCCat(instance->data_buf, stmp);
		    }
	    }
	break;
        case DATA_DETECTOR_COUNTER:
	    if (!items) items = n;
	case DATA_COINCIDENCE:
	    if (!items) {
		switch (type&0xFFFF0000) {
		    case DATA_COINCIDENCE_COINCIDENCE: 
			items = n;
		    break;
		    case DATA_COINCIDENCE_PAIR:
			items = n * n;
		    break;
		    default:
			items = 1;
		}
	    }
	    
	    intdata = (uint32_t*)data;
	    for (i=0; i<items; i++) {
		sprintf(stmp, "% 5u ", intdata[i]);
		xmlBufferCCat(instance->data_buf, stmp);
	    }

	break;

	case DATA_DISPERSION:
	    if (!items) items = n;
	case DATA_CORRELATION:
	    if (!items) items = n * (n - 1) / 2;

	    dbldata = (double*)data;
	    for (i=0; i<items;i++) {
		sprintf(stmp, "% 9.4lf ", dbldata[i]);
		xmlBufferCCat(instance->data_buf, stmp);
	    }
	break;

	default:
	    return DS_ERR_NOTSUPPORTED;
    }    

    
    return 0;
}

static int XMLWriterSetDuration(XMLWriter ctx, XMLWriterInstance instance, DSDuration duration) {
    char stmp[32];
    
    ds_assert_param(ctx);
    ds_assert_param(instance);
    ds_assert_param(duration);
    ds_assert_param(instance->f);

    dsDurationGet(duration, stmp);
    xmlNewChild(instance->data_node, NULL, "Duration", stmp);
    return 0;
}

static int XMLWriterSetQuality(XMLWriter ctx, XMLWriterInstance instance, double quality) {
    char stmp[16];
    
    ds_assert_param(ctx);
    ds_assert_param(instance);
    ds_assert_param(instance->f);
    
    sprintf(stmp, "%5.2lf", quality);
    xmlNewChild(instance->data_node, NULL, "Quality", stmp);
    return 0;
}
