#include "time.h"
#include "spi.h"
#include "ladc.h"


// Microcontroller in Master Mode

// FM25L256 support
// SPI0 Mode 0: CPOL=0, CPHA=0

// FM25L256 commands
#define WREN	0x06	/* Set Write Enable Latch 0000 0110b */
#define WRDI	0x04	/* Write Disable 0000 0100b */
#define RDSR	0x05	/* Read Status Register 0000 0101b */
#define WRSR	0x01	/* Write Status Register 0000 0001b */
#define READ	0x03	/* Read Memory Data 0000 0011b */
#define WRITE	0x02	/* Write Memory Data 0000 0010b */

void InitSPI0()
{
  S0SPCR = 0x0020;	// 8-bit, Master Mode
  S0SPCCR = 8;		// Divider for PCLK, 7.5 MHz for PCLK = 60 MHz

  IO0DIR |= Bit(12);
  IO0SET = RESET_LADC;
  TIMEdelay(10);
  PINSEL0 |= 0x1500;    // Set pins as SCK0, MISO0 and MOSI0
  TIMEdelay(10);
  IO0CLR = RESET_LADC;
}

U8 SPI0(U8 data)
{
  S0SPDR = data;

  while (!(S0SPSR & 0x80));
  return S0SPDR;
}

void InitFRAM()
{
  IO0DIR |= Bit(FRAM_Select) | (7 << FRAM_Chip);
  InitSPI0();
}



U8 FRAM_Command(U8 data)
{
  U8 b;

  IO0SET = Bit(FRAM_Select);
  b = SPI0(data);
  IO0CLR = Bit(FRAM_Select);
  return b;
}

U16 SelectFRAM(U32 addr)
{
  U8 selector;

  selector = addr >> FRAMchipSizeBits;
  IO0CLR = 0x7 << FRAM_Chip;
  IO0SET = selector << FRAM_Chip; 

  return (U16)(addr & MaxFRAMchip_addr);
}

void WriteFRAM_U8(U32 addr, U8 data)
{
  U16 faddr;

  faddr = SelectFRAM(addr);
  FRAM_Command(WREN);
  IO0SET = Bit(FRAM_Select);
  SPI0(WRITE);
  SPI0((U8)(faddr >> 8));
  SPI0((U8)faddr);
  SPI0(data);
  IO0CLR = Bit(FRAM_Select);   
}

void WriteFRAM_U16(U32 addr, U16 data)
{
  U16 faddr;

  faddr = SelectFRAM(addr);

  VICIntEnClr = BODbit;   // Disable BOD interruption
  FRAM_Command(WREN);
  IO0SET = Bit(FRAM_Select);
  SPI0(WRITE);
  SPI0((U8)(faddr >> 8));
  SPI0((U8)faddr);
  SPI0((U8)(data >> 8));
  SPI0((U8)data);
  IO0CLR = Bit(FRAM_Select);
  VICIntEnable   = BODbit;  // Enable BOD interruption    
}

void WriteFRAM_U32(U32 addr, U32 data)
{
  U16 faddr;

  faddr = SelectFRAM(addr);
  FRAM_Command(WREN);
  IO0SET = Bit(FRAM_Select);
  SPI0(WRITE);
  SPI0((U8)(faddr >> 8));
  SPI0((U8)faddr);
  SPI0((U8)(data >> 24));
  SPI0((U8)(data >> 16));
  SPI0((U8)(data >> 8));
  SPI0((U8)data);
  IO0CLR = Bit(FRAM_Select);   
}

int WriteFRAM(U32 addr, U8 *ptr, U32 counter)
{
  U16 faddr;
  U32 i;

  faddr = SelectFRAM(addr);
  FRAM_Command(WREN);
  IO0SET = Bit(FRAM_Select);
  SPI0(WRITE);
  SPI0((U8)(faddr >> 8));
  SPI0((U8)faddr);
  for (i = 0; i < counter; i++) {
    SPI0(*ptr++);
	addr++;
	if ((addr & MaxFRAMchip_addr) == 0) {
	  if (addr == (MaxFRAMaddr + 1))
	    return 0;
	  else {
        IO0CLR = Bit(FRAM_Select);	// finish the current write operation
        SelectFRAM(addr);			// select the next FRAM_Command chip
        FRAM_Command(WREN);			// start the next write operation
        IO0SET = Bit(FRAM_Select);
        SPI0(WRITE);
        SPI0(0);
        SPI0(0);
      }
	}
  }	    
  IO0CLR = Bit(FRAM_Select);   
  return 1;
}

U8 ReadFRAM_U8(U32 addr)
{
  U16 faddr;
  U8 data;

  faddr = SelectFRAM(addr);
  IO0SET = Bit(FRAM_Select);
  SPI0(READ);
  SPI0((U8)(faddr >> 8));
  SPI0((U8)faddr);
  data = SPI0(0);
  IO0CLR = Bit(FRAM_Select);
  return data;   
}

U16 ReadFRAM_U16(U32 addr)
{
  U16 faddr;
  U16 data;

  faddr = SelectFRAM(addr);
  IO0SET = Bit(FRAM_Select);
  SPI0(READ);
  SPI0((U8)(faddr >> 8));
  SPI0((U8)faddr);
  data = SPI0(0);
  data = (data << 8) | SPI0(0);
  IO0CLR = Bit(FRAM_Select);
  return data;   
}

U32 ReadFRAM_U32(U32 addr)
{
  U16 faddr;
  U32 data;

  faddr = SelectFRAM(addr);
  IO0SET = Bit(FRAM_Select);
  SPI0(READ);
  SPI0((U8)(faddr >> 8));
  SPI0((U8)faddr);
  data = SPI0(0);
  data = (data << 8) | SPI0(0);
  data = (data << 8) | SPI0(0);
  data = (data << 8) | SPI0(0);
  IO0CLR = Bit(FRAM_Select);
  return data;   
}

int ReadFRAM(U32 addr, U8 *ptr, U32 counter)
{
  U16 faddr;
  U32 i;
 
  faddr = SelectFRAM(addr);
  IO0SET = Bit(FRAM_Select);
  SPI0(READ);
  SPI0((U8)(faddr >> 8));
  SPI0((U8)faddr);
  for (i = 0; i < counter; i++) {
    *ptr++ = SPI0(0);
	addr++;
	if ((addr & MaxFRAMchip_addr) == 0) {
	  if (addr == (MaxFRAMaddr + 1))
	    return 0;
	  else {
        IO0CLR = Bit(FRAM_Select);	// finish the current read operation
        SelectFRAM(addr);			// select the next FRAM_Command chip
        IO0SET = Bit(FRAM_Select);  // start the next read operation
        SPI0(READ);
        SPI0(0);
        SPI0(0);
      }
	}
  }	    
  IO0CLR = Bit(FRAM_Select);   
  return 1;
}

